Package: h5lite
Title: Simplified 'HDF5' Interface
Version: 2.1.1.1
Authors@R: c(
    person(
      "Daniel P.", "Smith",
      email   = "dansmith01@gmail.com",
      role    = c("aut", "cre"),
      comment = c(ORCID = "0000-0002-2479-2044")),
    person(
      "Alkek Center for Metagenomics and Microbiome Research", 
      role = c("cph", "fnd") ))
Description: A user-friendly interface for the Hierarchical Data Format 5 
    ('HDF5') library designed to "just work." It bundles the necessary 
    system libraries to ensure easy installation on all platforms. Features 
    smart defaults that automatically map R objects (vectors, matrices, 
    data frames) to efficient 'HDF5' types, removing the need to manage 
    low-level details like dataspaces or property lists. Uses the 'HDF5' 
    library developed by The HDF Group <https://www.hdfgroup.org/>.
URL: https://github.com/cmmr/h5lite, https://cmmr.github.io/h5lite/
BugReports: https://github.com/cmmr/h5lite/issues
Depends: R (>= 4.2.0)
LinkingTo: hdf5lib (>= 2.1.1.0)
Suggests: bit64, knitr, rmarkdown, tinytest
NeedsCompilation: yes
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
Packaged: 2026-05-18 21:46:14 UTC; Daniel
Author: Daniel P. Smith [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-2479-2044>),
  Alkek Center for Metagenomics and Microbiome Research [cph, fnd]
Maintainer: Daniel P. Smith <dansmith01@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-18 22:00:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-18 23:33:22 UTC; unix
Archs: h5lite.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(as.character,h5)
S3method(print,compress)
S3method(print,h5)
S3method(print,inspect)
S3method(str,h5)
export(h5_attr_names)
export(h5_class)
export(h5_compression)
export(h5_create_file)
export(h5_create_group)
export(h5_delete)
export(h5_dim)
export(h5_exists)
export(h5_inspect)
export(h5_is_dataset)
export(h5_is_group)
export(h5_length)
export(h5_ls)
export(h5_move)
export(h5_names)
export(h5_open)
export(h5_read)
export(h5_str)
export(h5_typeof)
export(h5_write)
useDynLib(h5lite, .registration = TRUE)

# h5lite <img src="man/figures/logo.png" align="right" width="184" height="200" alt="h5lite logo" />

[![cran](https://img.shields.io/cran/v/h5lite?logo=r&label=CRAN)](https://CRAN.R-project.org/package=h5lite)
[![conda](https://img.shields.io/conda/v/conda-forge/r-h5lite?logo=anaconda&label=conda)](https://anaconda.org/conda-forge/r-h5lite)
[![covr](https://img.shields.io/codecov/c/gh/cmmr/h5lite?logo=codecov)](https://app.codecov.io/gh/cmmr/h5lite)

**h5lite** is the pain-free way to work with HDF5 files in R.

It is designed for data scientists who want to read/write objects and move on, and for package developers who need a reliable, dependency-free storage backend.


## Why h5lite?

If you've struggled with complex HDF5 bindings in the past, `h5lite` offers a fresh approach:

1.  **It Just Works:** No need to understand HDF5 dataspaces, hyperslabs, or property lists. `h5lite` maps R objects (numeric, character, factor, data.frame, and more) directly to their HDF5 equivalents.

2.  **Zero System Dependencies:** `h5lite` bundles the HDF5 library (via `hdf5lib`). Users do **not** need to install HDF5 system libraries manually.

3.  **Smart Defaults, Full Control:** It automatically selects the most efficient data types (e.g., saving space by storing small integers as `int8`), but gives you granular control when you need to conform to a strict spec.


## Installation

Install the released version from CRAN:

```r
install.packages("h5lite")
```

Or the development version from GitHub:

```r
# install.packages("pak")
pak::pak("cmmr/h5lite")
```


## Quick Start

The API consists primarily of two functions: `h5_write()` and `h5_read()`.

```r
library(h5lite)
file <- tempfile(fileext = ".h5")

# 1. Write simple objects
h5_write(1:10, file, "my_vector")
h5_write(I(42), file, "my_vector", attr = "my_id")
h5_write(matrix(rnorm(9), 3, 3), file, "my_matrix")

# 2. Write a list (creates a group hierarchy)
config <- list(version = 1.0, params = list(a = 1, b = 2))
h5_write(config, file, "simulation_config")

# 3. Read it back
my_vec <- h5_read(file, "my_vector")

# 4. Inspect the file
h5_ls(file)
#> [1] "my_vector"                  "my_matrix"
#> [3] "simulation_config"          "simulation_config/version"
#> [5] "simulation_config/params"   "simulation_config/params/a"
#> [7] "simulation_config/params/b"

h5_str(file)
#> /
#> ├── my_vector <uint8 × 10>
#> │   └── @my_id <uint8 scalar>
#> ├── my_matrix <float64 × 3 × 3>
#> └── simulation_config/
#>     ├── version <uint8 × 1>
#>     └── params/
#>         ├── a <uint8 × 1>
#>         └── b <uint8 × 1>
```


## Power User Features

### The `as` Argument: Precise Control

Need to conform to a specific file specification? The `as` argument allows you to override automatic behavior and explicitly define on-disk types.

```r
# Force specific numeric types
h5_write(1:10, file, "dataset_a", as = "int32")
h5_write(rnorm(10), file, "dataset_b", as = "float32")

# Control string lengths (e.g., fixed-length ASCII for compatibility)
h5_write(c("A", "B"), file, "fixed_strs", as = "ascii[10]")

h5_str(file)
#> ...
#> ├── dataset_a <int32 × 10>
#> ├── dataset_b <float32 × 10>
#> └── fixed_strs <ascii[10] × 2>
```

When writing Data Frames, you can map types for specific columns using a named vector.

```r
df <- data.frame(
  id    = 1:5, 
  score = c(1.1, 2.2, 3.3, 4.4, 5.5),
  note  = c("a", "b", "c", "d", "e")
)

# Store 'id' as 16-bit integer, 'score' as 32-bit float, and coerce 'note' to ascii
h5_write(df, file, "experiment_data", 
         as = c(id = "uint16", score = "float32", note = "ascii"))

h5_str(file)
#> ...
#> └── experiment_data <compound[3] × 5>
#>     ├── $id <uint16>
#>     ├── $score <float32>
#>     └── $note <ascii>
```

### Advanced Compression & Chunking

`h5lite` natively bundles an extensive suite of state-of-the-art compression filters (including **Blosc2**, **Zstandard**, **LZ4**, and lossy **ZFP**). For simple use cases, you can pass a string configuration to the `compress` argument. For precise control over the pipeline—including chunk sizing, bitshuffling, and Scale-Offset scaling—use the `h5_compression()` function.

```r
# Simple setup: High-performance Blosc2 with Zstandard
h5_write(rnorm(1000), file, "data_blosc", compress = "blosc2-zstd")

# Advanced setup: LZ4 compression, optimal integer packing, and custom chunk sizing
cmp <- h5_compression("lz4-9", int_packing = TRUE, chunk_size = 512 * 1024)
h5_write(1:1000, file, "data_custom", compress = cmp)
```

### Efficient Partial Reading
For large datasets that exceed system RAM, `h5lite` provides partial reading via `start` and `count` parameters. It automatically targets the most logical dimension (e.g., rows in a matrix or elements in a vector).

```r
# Read a 100-row slice starting from row 500
subset <- h5_read(file, "large_matrix", start = 500, count = 100)
```


## Comparison

| Feature               | h5lite                    | rhdf5 / hdf5r                                          |
| :-------------------- | :------------------------ | :----------------------------------------------------- |
| **Philosophy**        | "Opinionated" & Simple    | Comprehensive Wrapper                                  |
| **API Style**         | Native R (`read`/`write`) | Low-level (Files, Dataspaces, Memspaces)               |
| **HDF5 Installation** | **Bundled** (Zero-config) | **System Requirement** (Manual install often required) |
| **Data Typing**       | Automatic (safe defaults) | Manual (user specified)                                |
| **Partial I/O**       | **Supported** (Simplified)| Supported (Manual hyperslabs)                          |
| **Learning Curve**    | Low (Minutes)             | High (Days)                                            |

**Use `rhdf5` or `hdf5r` if you need to:**

-   Work with complex or custom HDF5 data types not supported by `h5lite` (e.g., bitfields, references, variable-length nested arrays).
-   Perform advanced, multi-dimensional hyperslab selections that cannot be expressed with simple `start`/`count` parameters.

**Use `h5lite` if you want to:**

-   Quickly and safely get data into or out of a file.
-   Take advantage of modern, zero-config compression pipelines (Blosc2, Zstd, ZFP) without building system libraries.
-   Perform efficient partial reads without the complexity of low-level hyperslab math.
-   Avoid thinking about low-level details.


## Documentation
 
-   **[Get Started](https://cmmr.github.io/h5lite/articles/h5lite.html)**: A general introduction.
-   **[Atomic Vectors](https://cmmr.github.io/h5lite/articles/atomic-vectors.html)**: Details on vectors and scalars.
-   **[Data Types](https://cmmr.github.io/h5lite/articles/data-types.html)**: Controlling storage types.
-   **[Compression](https://cmmr.github.io/h5lite/articles/compression.html)**: Configuring modern codecs and tuning chunk size.
-   **[Partial Reading](https://cmmr.github.io/h5lite/articles/partial-reading.html)**: Efficiently reading data subsets with `start` and `count`.
-   **[Matrices and Arrays](https://cmmr.github.io/h5lite/articles/matrices.html)**: Handling multi-dimensional data.
-   **[Data Frames](https://cmmr.github.io/h5lite/articles/data-frames.html)**: Using compound datasets.
-   **[Organization](https://cmmr.github.io/h5lite/articles/organization.html)**: Using groups and lists to structure files.
-   **[Attributes In-Depth](https://cmmr.github.io/h5lite/articles/attributes-in-depth.html)**: A deep dive into metadata handling.
-   **[Object-Oriented Interface](https://cmmr.github.io/h5lite/articles/oo-interface.html)**: A guide to the `h5_open()` handle for a streamlined workflow.
-   **[Parallel Processing](https://cmmr.github.io/h5lite/articles/parallel-io.html)**: Guide for multi-threaded and multi-process access.


![](https://cmmr-repos.goatcounter.com/count?p=/h5lite)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----setup--------------------------------------------------------------------
library(h5lite)
file <- tempfile(fileext = ".h5")

## -----------------------------------------------------------------------------
# Create a 3x4 matrix
mat <- matrix(1:12, nrow = 3, ncol = 4)

# Write to file
h5_write(mat, file, "linear_algebra/mat_a")

# Read back
mat_in <- h5_read(file, "linear_algebra/mat_a")

# Verify
all.equal(mat, mat_in)

## -----------------------------------------------------------------------------
# Create a 3D array (e.g., spatial data over time: x, y, time)
vol <- array(runif(24), dim = c(4, 3, 2))

h5_write(vol, file, "spatial/volume")

# Check dimensions without reading the full data
h5_dim(file, "spatial/volume")

## -----------------------------------------------------------------------------
# Create a matrix with row and column names
data <- matrix(rnorm(6), nrow = 2)
rownames(data) <- c("Sample_A", "Sample_B")
colnames(data) <- c("Gene_1", "Gene_2", "Gene_3")

h5_write(data, file, "genetics/expression")

# Read back
data_in <- h5_read(file, "genetics/expression")

print(data_in)

## -----------------------------------------------------------------------------
# Large matrix of zeros (highly compressible)
sparse_mat <- matrix(0, nrow = 1000, ncol = 1000)
sparse_mat[1:10, 1:10] <- 1

# Write with default compression (zlib level 5)
h5_write(sparse_mat, file, "compressed/matrix")

# Write with high compression (zlib level 9)
h5_write(sparse_mat, file, "compressed/matrix_max", compress = "gzip-9")

## ----include=FALSE------------------------------------------------------------
unlink(file)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----setup--------------------------------------------------------------------
library(h5lite)
file <- tempfile(fileext = ".h5")

## -----------------------------------------------------------------------------
# First, write a dataset
h5_write(1:10, file, "measurements/temperature")

# Now, attach attributes to it
h5_write(I("Celsius"),    file, "measurements/temperature", attr = "units")
h5_write(I("2023-10-27"), file, "measurements/temperature", attr = "date")
h5_write(I(0.1),          file, "measurements/temperature", attr = "precision")

## -----------------------------------------------------------------------------
# Create a vector with custom R attributes
data <- rnorm(5)
attr(data, "description") <- I("Randomized control group")
attr(data, "valid")       <- I(TRUE)

# Write the object
h5_write(data, file, "experiment/control")

# Check the file - the attributes are there
h5_attr_names(file, "experiment/control")

h5_str(file)

## -----------------------------------------------------------------------------
# Read just the 'units' attribute
units <- h5_read(file, "measurements/temperature", attr = "units")
print(units)

## -----------------------------------------------------------------------------
# Read the full dataset
temps <- h5_read(file, "measurements/temperature")

# The attributes are available in R
attributes(temps)

str(temps)

## -----------------------------------------------------------------------------
h5_attr_names(file, "measurements/temperature")

## -----------------------------------------------------------------------------
# Delete the 'precision' attribute
h5_delete(file, "measurements/temperature", attr = "precision")

# Verify removal
h5_attr_names(file, "measurements/temperature")

## -----------------------------------------------------------------------------
# A vector with names
named_vec <- c(a = 1, b = 2, c = 3)

# Write as a standard Dataset -> Names are preserved
h5_write(named_vec, file, "my_dataset")
h5_names(file, "my_dataset")

# Write as an Attribute -> Names are LOST
h5_write(named_vec, file, "measurements/temperature", attr = "meta_vec")
h5_names(file, "measurements/temperature", attr = "meta_vec")

## -----------------------------------------------------------------------------
# A data frame with metadata
df <- data.frame(
  id = 1:3, 
  status = c("ok", "fail", "ok")
)

# Write as attribute
h5_write(df, file, "measurements/temperature", attr = "log")

# Column names survive!
h5_names(file, "measurements/temperature", attr = "log")

## -----------------------------------------------------------------------------
# Write the temperature data again, but use a fixed length string for 'description'
h5_write(data, file, "experiment/control", as = c("@description" = "ascii[]"))

# Store an attribute as a `uint8` instead of the default `int32`
h5_write(I(42), file, "measurements/temperature", "sensor_id", as = "uint8")

## -----------------------------------------------------------------------------
# Force the 'valid' attribute to be read as logical, even if stored as integer
meta <- h5_read(file, "experiment/control", attr = "valid", as = "logical")

## ----include=FALSE------------------------------------------------------------
unlink(file)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----single-value-------------------------------------------------------------
library(h5lite)
file <- tempfile(fileext = ".h5")

# --- 1. Vectors (Element-level targeting) ---
h5_write(seq(10, 100, by = 10), file, "my_vector")

# Start at the 4th element, read 3 elements
h5_read(file, "my_vector", start = 4, count = 3)

# --- 2. Matrices (Row-level targeting) ---
mat <- matrix(1:50, nrow = 10, ncol = 5)
h5_write(mat, file, "my_matrix")

# Start at row 5, read 3 complete rows (automatically spans all columns)
h5_read(file, "my_matrix", start = 5, count = 3)

# --- 3. Data Frames (Row-level targeting) ---
h5_write(mtcars, file, "my_mtcars")

# Start at row 10, read 5 complete rows
h5_read(file, "my_mtcars", start = 10, count = 5)

# --- 4. 3D Arrays (Matrix-level targeting) ---
arr <- array(1:24, dim = c(2, 3, 4)) 
h5_write(arr, file, "my_array")

# Start at the 2nd matrix, read 2 complete matrices
h5_read(file, "my_array", start = 2, count = 2)

## ----exact-index--------------------------------------------------------------
# Read exactly row 5 of the matrix. 
# The row dimension is dropped, returning a 1D vector.
row_vec <- h5_read(file, "my_matrix", start = 5)
row_vec

class(row_vec)

## ----range-index--------------------------------------------------------------
# Read row 5, but signal a range request by setting count = 1.
# The original geometry is preserved, returning a 1x5 matrix.
row_mat <- h5_read(file, "my_matrix", start = 5, count = 1)
row_mat

class(row_mat)

## ----multi-value--------------------------------------------------------------
# Matrix: Start at row 5, column 2, and read 3 elements along that row.
# The row is an exact point index (dropped). The columns are a range (preserved).
# Returns a 1D vector of length 3.
h5_read(file, "my_matrix", start = c(5, 2), count = 3)

# Matrix: Extract exactly row 5, column 2. 
# Because count is omitted, the final dimension is also dropped.
# Returns an unnamed scalar value.
h5_read(file, "my_matrix", start = c(5, 2))

# 3D Array: Target matrix 2, row 1.
# The matrix and row are exact point indices (dropped). 
# Returns a 1D vector containing the columns of that specific row.
h5_read(file, "my_array", start = c(2, 1))

## ----cleanup, include=FALSE---------------------------------------------------
unlink(file)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----setup--------------------------------------------------------------------
library(h5lite)

# Create a temporary file for this example
file <- tempfile(fileext = ".h5")

# Open the handle
h5 <- h5_open(file)

# The print method shows the file path and current internal working directory
print(h5)

## ----write_read---------------------------------------------------------------
# Write data using the handle
h5$write(1:10, "dataset1")
h5$write(matrix(1:9, 3, 3), "matrix_data")

# List contents
h5$ls()

# Read data back
my_data <- h5$read("dataset1")
print(my_data)

## ----navigation---------------------------------------------------------------
# Create a group structure
h5$create_group("simulations")

# Navigate into the group
h5$cd("simulations")
h5$pwd()

# Write data using relative paths (writes to /simulations/run1)
h5$write(rnorm(10), "run1")

# Verify the location
h5$ls()

## ----absolute_paths-----------------------------------------------------------
# Writes to the root, ignoring the fact that we are in /simulations
h5$write(100, "/root_dataset")

## ----relative_nav-------------------------------------------------------------
h5$cd("..")
h5$pwd() # Now back at "/"

## ----ref_semantics------------------------------------------------------------
h5_alias <- h5

# Change directory in the alias
h5_alias$cd("simulations")

# The original handle is also updated
h5$pwd()

## ----close--------------------------------------------------------------------
h5$close()

# Further attempts to use the handle will result in an error:
# h5$ls() 
# Error: This h5 file handle has been closed.

#' Write an R Object to HDF5
#' 
#' Writes an R object to an HDF5 file, creating the file if it does not exist.
#' This function acts as a unified writer for datasets, groups (lists), and attributes.
#'
#' @param data The R object to write. Supported: `numeric`, `complex`, 
#'   `logical`, `character`, `factor`, `raw`, `matrix`, `data.frame`, `integer64`, 
#'   `POSIXt`, `NULL`, and nested `list`s.
#' @param file The path to the HDF5 file.
#' @param name The name of the dataset or group to write (e.g., "/data/matrix").
#' @param attr The name of an attribute to write.
#'   * If `NULL` (default), `data` is written as a dataset or group at the path `name`.
#'   * If provided (string), `data` is written as an attribute named `attr` attached to the object `name`.
#' @param as The target HDF5 data type. Defaults to `"auto"`.
#'   See the **Data Type Selection** section for a full list of valid options 
#'   (including `"int64"`, `"bfloat16"`, `"utf8[n]"`, etc.) and how to map 
#'   sub-components of `data`.
#' @param compress Compression configuration. Default is `"gzip"`. Pass a basic string 
#'   to specify the algorithm and level (e.g., `"none"`, `"gzip"`, `"zstd-7"`, `"lz4"`, 
#'   `"blosc1-lz4-9"`, `"blosc2-gzip-3"`, `"blosc2-zstd"`), or pass a `compress` object 
#'   created by [h5_compression()] for advanced pipeline control (including scale-offset 
#'   algorithms, Fletcher32 checksums, or Blosc2 pre-filters). See [h5_compression()] 
#'   for the complete list of available codecs and options.
#' 
#' @section Writing Scalars:
#' 
#' By default, `h5_write` saves single-element vectors as 1-dimensional arrays.
#' To write a true HDF5 scalar, wrap the value in `I()` to treat it "as-is."
#' 
#' #### Examples
#' ```
#' h5_write(I(5), file, "x") # Creates a scalar dataset
#' h5_write(5, file, "x")    # Creates a 1D array of length 1
#' ```
#' 
#' @section Data Type Selection (`as` Argument):
#' 
#' By default, `as = "auto"` will automatically select the most appropriate
#' data type for the given object. For numeric types, this will be the smallest
#' type that can represent all values in the vector. For character types, 
#' `h5lite` will use a ragged vs rectangular heuristic, favoring small file 
#' size over fast I/O. For R data types not mentioned below, see 
#' `vignette("data-types")` for information on their fixed mappings to HDF5 
#' data types.
#' 
#' ### Numeric and Logical Vectors
#' 
#' When writing a numeric or logical vector, you can specify one of the 
#' following storage types for it:
#' 
#' * **Floating Point:** `"float16"`, `"float32"`, `"float64"`, `"bfloat16"`
#' * **Signed Integer:** `"int8"`, `"int16"`, `"int32"`, `"int64"`
#' * **Unsigned Integer:** `"uint8"`, `"uint16"`, `"uint32"`, `"uint64"`
#' 
#' **NOTE:** `NA` values must be stored as `float64`. `NaN`, `Inf`, and `-Inf`
#' must be stored as a floating point type.
#' 
#' #### Examples
#' ```
#' h5_write(1:100, file, "big_ints", as = "int64")
#' h5_write(TRUE,  file, "my_bool",  as = "float32")
#' ```
#' 
#' ### Character Vectors
#' 
#' You can control whether character vectors are stored as variable or fixed
#' length strings, and whether to use UTF-8 or ASCII encoding.
#' 
#' * **Variable Length Strings:** `"utf8"`, `"ascii"`
#' * **Fixed Length Strings:**
#'     * `"utf8[]"` or `"ascii[]"` (length is set to the longest string)
#'     * `"utf8[n]"` or `"ascii[n]"` (where `n` is the length in bytes)
#' 
#' **NOTE:** Variable-length strings allow for `NA` values but cannot be 
#' compressed on disk. Fixed-length strings allow for compression but do not 
#' support `NA`.
#' 
#' #### Examples
#' ```
#' h5_write(letters[1:5],    file, "len10_strs", as = "utf8[10]")
#' h5_write(c('X', 'Y', NA), file, "var_chars",  as = "ascii")
#' ```
#' 
#' ### Lists, Data Frames, and Attributes
#' 
#' Provide a named vector to apply type mappings to sub-components of `data`. 
#' Set `"skip"` as the type to skip a specific component.
#' 
#' * **Specific Name:** `"col_name" = "type"` (e.g., `c(score = "float32")`)
#' * **Specific Attribute:** `"@@attr_name" = "type"`
#' * **Class-based:** `".integer" = "type"`, `".numeric" = "type"`
#' * **Class-based Attribute:** `"@@.character" = "type"`, `"@@.logical" = "type"`
#' * **Global Fallback:** `"." = "type"`
#' * **Global Attribute Fallback:** `"@@." = "type"`
#' 
#' #### Examples
#' ```
#' # To strip attributes when writing:
#' h5_write(data, file, 'no_attrs_obj', as = c('@@.' = "skip"))
#' 
#' # To only save the `hp` and `wt` columns:
#' h5_write(mtcars, file, 'my_df', as = c('hp' = "auto", 'wt' = "float32", '.' = "skip"))
#' ```
#' 
#' @section Dimension Scales:
#' `h5lite` automatically writes `names`, `row.names`, and `dimnames` as 
#' HDF5 dimension scales. Named vectors will generate an `<name>_names` 
#' dataset. A data.frame with row names will generate an `<name>_rownames` 
#' dataset (column names are saved internally in the original dataset). 
#' Matrices will generate `<name>_rownames` and `<name>_colnames` datasets. 
#' Arrays will generate `<name>_dimscale_1`, `<name>_dimscale_2`, etc. 
#' Special HDF5 metadata attributes link the dimension scales to the dataset. 
#' The dimension scales can be relocated with `h5_move()` without breaking the 
#' link.
#' 
#' @return Invisibly returns `file`. This function is called for its side effects.
#' @seealso [h5_read()], [h5_compression()], `vignette('compression')`
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' # 1. Writing Basic Datasets
#' h5_write(1:10, file, "data/integers")
#' h5_write(rnorm(10), file, "data/floats")
#' h5_write(letters[1:5], file, "data/chars")
#' 
#' # 2. Writing Attributes
#' # Write an object first
#' h5_write(1:10, file, "data/vector")
#' # Attach an attribute to it using the 'attr' parameter
#' h5_write(I("My Description"), file, "data/vector", attr = "description")
#' h5_write(I(100), file, "data/vector", attr = "scale_factor")
#' 
#' # 3. Controlling Data Types
#' # Store values as 32-bit signed integers
#' h5_write(1:5, file, "small_ints", as = "int32")
#' 
#' # 4. Writing Complex Structures (Lists/Groups)
#' my_list <- list(
#'   meta    = list(id = 1, name = "Experiment A"),
#'   results = matrix(runif(9), 3, 3),
#'   valid   = I(TRUE)
#' )
#' h5_write(my_list, file, "experiment_1", as = c(id = "uint16"))
#' 
#' # 5. Writing Data Frames (Compound Datasets)
#' df <- data.frame(
#'   id    = 1:5,
#'   score = c(10.5, 9.2, 8.4, 7.1, 6.0),
#'   grade = factor(c("A", "A", "B", "C", "D"))
#' )
#' h5_write(df, file, "records/scores", as = c(grade = "ascii[1]"))
#' 
#' # 6. Fixed-Length Strings
#' h5_write(c("A", "B"), file, "fixed_str", as = "ascii[10]")
#' 
#' # 7. Review the file structure
#' h5_str(file)
#' 
#' # 8. Clean up
#' unlink(file)
h5_write <- function(data, file, name, attr = NULL, as = "auto", compress = "gzip") {
  
  file <- validate_strings(file, name, attr)
  
  if (!is.null(attr) && !h5_exists(file, name))
    stop("Cannot write attribute '", attr, "' to non-existent object '", name, "'.", call. = FALSE)

  # Prepare the 'as' map for objects and attributes
  # Example: obj_as  = c("@ready" = "logical", ".uint" = "integer", "@." = "null")
  #          attr_as = c("ready" = "logical",  ".uint" = "integer", "."  = "null")
  obj_as  <- validate_as(as)
  attr_as <- obj_as
  if (!is.null(names(attr_as))) {
    attr_as <- attr_as[grepl("^[.@]", names(attr_as))] 
    if (length(attr_as) > 0) {
      attr_as <- attr_as[rev(order(names(attr_as)))]     
      names(attr_as) <- sub("^@", "", names(attr_as))
      attr_as <- attr_as[!duplicated(names(attr_as))]
    }
    if (is.null(attr_as) || length(attr_as) == 0) attr_as <- "auto"
  }

  compress <- h5_compression(compress)

  # Write the data
  h5_create_group(file, name = "/")
  if (is_list_group(data)) {
    write_group(data, file, name, obj_as, attr_as, compress, dry = TRUE)
    write_group(data, file, name, obj_as, attr_as, compress, dry = FALSE)
  } else {
    write_data(data, file, name, attr, obj_as, attr_as, compress, dry = TRUE)
    write_data(data, file, name, attr, obj_as, attr_as, compress, dry = FALSE)
  }
  
  invisible(file)
}



#' Recursively write a list as a group
#' @noRd
#' @keywords internal
write_group <- function(data, file, name, obj_as, attr_as, compress, dry = FALSE) {

  if (!dry) h5_delete(file, name, warn = FALSE)
  if (!dry) h5_create_group(file, name)
  
  write_attributes(data, file, name, attr_as, dry = dry)
  
  # Recursively write children
  for (child_name in names(data)) {
    child_path <- paste(name, child_name, sep = "/")
    child_data <- data[[child_name]]
    if (is_list_group(child_data)) {
      write_group(child_data, file, child_path, obj_as, attr_as, compress, dry = dry)
    } else {
      write_data(child_data, file, child_path, attr = NULL, obj_as, attr_as, compress, dry = dry)
    }
  }
}

#' Write a single dataset or attribute
#' @noRd
#' @keywords internal
write_data <- function(data, file, name, attr, obj_as, attr_as, compress, dry = FALSE) {
  
  # Convert POSIXt vectors/columns to ISO 8601 character strings.
  if (inherits(data, "POSIXt")) {
    data <- format(data, format = "%Y-%m-%dT%H:%M:%OSZ")
  }
  else if (is.data.frame(data)) {
    for (i in seq_along(data))
      if (inherits(data[[i]], "POSIXt"))
        data[[i]] <- format(data[[i]], format = "%Y-%m-%dT%H:%M:%OSZ")
  }

  map_key <- if (is.null(attr)) basename(name) else attr
  h5_type <- resolve_h5_type(data, map_key, obj_as)
  
  if (all(h5_type == "skip")) return (NULL)
  
  if (is.data.frame(data) && any(h5_type == "skip")) {
    data    <- data[, h5_type != "skip", drop = FALSE]
    h5_type <- h5_type[h5_type != "skip"]
  }
  
  for (i in which(startsWith(h5_type, "ascii"))) {
    
    # Converts from:      ÀÁÂÃÄÅÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖÙÚÛÜÝàáâãäåçèéêëìíîïðñòóôõöùúûüýÿ
    wanted   <- enc2utf8("AAAAAACEEEEIIIIDNOOOOOUUUUYaaaaaaceeeeiiiidnooooouuuuyy")
    unwanted <- enc2utf8(intToUtf8(c(
      192L, 193L, 194L, 195L, 196L, 197L, 199L, 200L, 201L, 202L, 203L, 
      204L, 205L, 206L, 207L, 208L, 209L, 210L, 211L, 212L, 213L, 214L, 
      217L, 218L, 219L, 220L, 221L, 224L, 225L, 226L, 227L, 228L, 229L, 
      231L, 232L, 233L, 234L, 235L, 236L, 237L, 238L, 239L, 240L, 241L, 
      242L, 243L, 244L, 245L, 246L, 249L, 250L, 251L, 252L, 253L, 255L )))
    
    data[[i]] <- enc2utf8(data[[i]])
    data[[i]] <- enc2utf8(chartr(unwanted, wanted, data[[i]]))
    data[[i]] <- iconv(data[[i]], "UTF-8", "ASCII//TRANSLIT", "?")
  }
  
  dims <- validate_dims(data)

  if (is.null(attr)) {
    if (!dry)
      .Call("C_h5_write_dataset", file, name, data, h5_type, dims, compress, PACKAGE = "h5lite")
    
    write_attributes(data, file, name, attr_as, dry = dry)
  }
  else {
    
    if (!dry)
      .Call("C_h5_write_attribute", file, name, attr, data, h5_type, dims, PACKAGE = "h5lite")
  }
}

#' Write R attributes to HDF5
#' @noRd
#' @keywords internal
write_attributes <- function(data, file, name, attr_as, dry = FALSE) {

  attr_names <- names(attributes(data))
  attr_names <- setdiff(attr_names, c("class", "dim", "dimnames", "names", "row.names", "levels"))

  for (attr in attr_names) {
    attr_data <- base::attr(data, attr, exact = TRUE)
    if (is_list_group(attr_data)) next
    write_data(attr_data, file, name, attr, attr_as, attr_as, list("none", NULL, NULL), dry = dry)
  }
}


# --- Type Resolution Logic ---

#' Resolves the HDF5 type based on the 'as' map and data properties
#' @noRd
#' @keywords internal
#' @param data The R object.
#' @param name The name of the object (column name, dataset name, or attribute name).
#' @param as_map The processed 'as' argument (named vector).
resolve_h5_type <- function(data, name, as_map) {
  
  # Resolve type for *each* column
  if (is.data.frame(data)) {

    if (ncol(data) == 0)
      stop("Cannot write a data.frame with zero columns: ", name, call. = FALSE)
    
    col_types <- character(ncol(data))
    col_names <- names(data)
    for (i in seq_along(data)) {
      col_types[i] <- resolve_h5_type(data[[i]], col_names[i], as_map)
    }
    return(col_types)
  }
  
  if (is.null(data))               return ("null")
  if (is.raw(data))                return ("raw")
  if (is.complex(data))            return ("complex")
  if (inherits(data, "integer64")) return ("bit64")
  if (inherits(data, "factor")) {
    
    if (!is.factor(data))
      stop("Non-factors with factor class cannot be written to HDF5.", call. = FALSE) # nocov
    if (!is.character(levels(data)))
      stop("Factors with non-character levels cannot be written to HDF5.", call. = FALSE) # nocov
    if (anyNA(data))
      stop("Factors with NA values cannot be written to HDF5 Enum types. Convert to character vector first.", call. = FALSE)
    if (typeof(data) != "integer")
      stop("Factors with non-integer values cannot be written to HDF5 Enum types.", call. = FALSE) # nocov
    
    return ("factor")
  }
  
  
  if (is.null(names(as_map))) {
    h5_type <- tolower(as_map)
    
  } else {
  
    # Generate type keys for lookup (e.g., .integer, .double)
    mode <- paste0(".", storage.mode(data))
    if (is.numeric(data)) { keys <- c(name, mode, ".numeric", ".") }
    else                  { keys <- c(name, mode,             ".") }
    
    h5_type <- "auto"
    for (key in keys) {
      if (key %in% names(as_map)) {
        h5_type <- tolower(as_map[[key]])
        break
      }
    }
    
  }
  
  
  if (is.character(data)) {

    arg_errmsg <- paste("Invalid `as` argument for character vector:", h5_type)
    na_errmsg  <- paste("`NA` cannot be encoded by fixed length strings.")
    
    h5_type <- tolower(h5_type)
    cset    <- tryCatch(
      expr  = match.arg(sub('\\[.+$', '', h5_type), c("auto", "ascii", "skip", "utf8")), 
      error = function (e) { stop(arg_errmsg, call. = FALSE) })
    
    if (cset == "skip") return ("skip")

    # Auto-select: Always use UTF-8.
    # Choose fixed length when strings are short and consistent.
    if (cset == "auto") {
      if (length(data) > 0 && !anyNA(data)) {
        str_lens <- nchar(data, type = "bytes")
        max_len  <- max(str_lens)
        mean_len <- mean(str_lens) + 16
        if (max_len < 4096 && max_len < mean_len * 4)
          return (paste0("utf8[", max_len, "]"))
      }
      return ("utf8")
    }
    
    # User-specified fixed length
    if (grepl("\\[\\d*\\]", h5_type)) {
      if (anyNA(data)) stop(na_errmsg, call. = FALSE)
      size <- as.integer(sub(".*\\[(\\d*)\\].*", "\\1", h5_type))
      if (is.na(size) || size < 1)
        size <- max(c(1L, nchar(data, type = "bytes")))
      return (paste0(cset, "[", size, "]"))
    }
    
    # User-specified variable length
    return (cset)
  }
  

  if (is.numeric(data) || is.logical(data)) {
    
    choices <- c(
      "auto", "skip", 
      "bfloat16", "float16", "float32", "float64", 
      "int8", "int16", "int32", "int64",
      "uint8", "uint16", "uint32", "uint64")
    
    h5_type <- match.arg(tolower(h5_type), choices)
    
    if (h5_type == "skip")    return ("skip")
    if (h5_type == "float64") return ("float64")
    
    if (h5_type == "auto") {
      
      if (anyNA(data))      return ("float64") # NA must use float64
      if (is.logical(data)) return ("uint8")

      if (is.double(data)) {

        if (any(data %% 1 > 0, na.rm = TRUE)) # Fractional values
          return ("float64")

        if (any(!is.finite(data))) { # NaN, Inf, or -Inf present
          rng <- range(c(1.0, data), na.rm = TRUE, finite = TRUE)
          if (rng[1] >= -2^24 && rng[2] <= 2^24) return ("float32")
          return ("float64")
        }
      }

      # Choose optimal data type for finite integers
      rng <- range(c(1, data), na.rm = TRUE, finite = TRUE)
      lo  <- rng[[1]]
      hi  <- rng[[2]]

      if (lo >= 0) { # Unsigned integer
        if (hi <= 2^8-1)  return ("uint8")
        if (hi <= 2^16-1) return ("uint16")
        if (hi <= 2^32-1) return ("uint32")
        if (hi <  2^53)   return ("uint64")
      }
      else { # Signed integer
        if (lo >= -2^7  && hi <= 2^7-1)  return ("int8")
        if (lo >= -2^15 && hi <= 2^15-1) return ("int16")
        if (lo >= -2^31 && hi <= 2^31-1) return ("int32")
        if (lo >  -2^53 && hi <  2^53)   return ("int64")
      }

      return ("float64")
    }
    
    # Sanity check user's requested HDF5 numeric type
    
    if (length(data) == 0) return(h5_type)
    if (any(!is.finite(data)) && !grepl("float", h5_type, fixed = TRUE))
      stop("Data contains NA/NaN/Inf; requires float type.", call. = FALSE)
    
    if (any(is.finite(data)) && !grepl("float", h5_type, fixed = TRUE)) {
      type_ranges <- list(
        'int8'    = c(-2^7,  2^7-1),   'uint8'    = c(0, 2^8-1),   
        'int16'   = c(-2^15, 2^15-1),  'uint16'   = c(0, 2^16-1),
        'int32'   = c(-2^31, 2^31-1),  'uint32'   = c(0, 2^32-1),  
        'int64'   = c(-2^63, 2^63-1),  'uint64'   = c(0, 2^64-1),
        'float16' = c(-65504,  65504), 'bfloat16' = c(-3.4e38, 3.4e38),
        'float32' = c(-3.4e38, 3.4e38) )
      val_range <- range(data, na.rm = TRUE, finite = TRUE)
      rng <- type_ranges[[h5_type]]
      if (!is.null(rng) && (val_range[1] < rng[1] || val_range[2] > rng[2])) {
        stop("Data range [", val_range[1], ", ", val_range[2], "] exceeds '", h5_type, "'", call. = FALSE)
      }
    }
    
    return(h5_type)
  }
  
  stop("Cannot write data of class ", paste(class(data), collapse = "/"), " to HDF5.", call. = FALSE) # nocov
}


#' @noRd
#' @keywords internal
validate_dims <- function (data) {
  
  # If data is wrapped in I(), treat as a scalar (NULL dims), but ONLY if it's length 1.
  # This prevents accidentally writing a multi-element 'AsIs' vector as a scalar, which would cause data loss.
  if (inherits(data, 'AsIs')) {
    if (length(data) == 1) {
      return(NULL)
    } else {
      warning("I() wrapper ignored for vector of length > 1. Writing as a 1D array.")
    }
  }
  
  # Otherwise, infer dimensions from the object. A vector will have length, a matrix/array will have dim().
  if (is.null(dim(data))) length(data) else dim(data)
}


#' @noRd
#' @keywords internal
is_list_group <- function (data) {
  # A "list group" is a list that is not a data.frame.
  return (is.list(data) && !is.data.frame(data))
}

local({
  #test_that("Double scalars and vectors work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  vec <- c(1.1, 2.2, 3.3)
  h5_write(vec, file, "vec")
  expect_equal(h5_read(file, "vec"), vec)
  
  h5_write(pi, file, "scalar")
  expect_equal(h5_read(file, "scalar"), pi)
})

local({
  #test_that("Special values (NA, NaN, Inf) work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  vec <- c(1.0, NA_real_, NaN, Inf, -Inf)
  h5_write(vec, file, "special")
  
  res <- h5_read(file, "special")
  expect_equal(res[1], 1.0)
  expect_true(is.na(res[2]))
  expect_true(is.nan(res[3]))
  expect_equal(res[4], Inf)
  expect_equal(res[5], -Inf)
})

local({
  #test_that("Double matrices work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  mat <- matrix(rnorm(9), 3, 3)
  h5_write(mat, file, "matrix")
  expect_equal(h5_read(file, "matrix"), mat)
})

local({
  #test_that("Double attributes work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  h5_write(1, file, "dset")
  h5_write(1.23, file, "dset", attr = "p_val")
  expect_equal(h5_read(file, "dset", attr = "p_val"), 1.23)
})

local({
  #test_that("Precision control (float32)", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  val <- 1.123456789
  # Write as 32-bit float
  h5_write(val, file, "f32", as = "float32")
  expect_equal(h5_typeof(file, "f32"), "float32")
  
  # Precision should be lost
  read_val <- h5_read(file, "f32")
  expect_false(read_val == val)
  expect_equal(read_val, val, tolerance = 1e-7)
})

local({
  #test_that("Factors work (ENUM)", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  f <- factor(c("a", "b", "a", "c"))
  m <- structure(f, dim = c(2, 2))
  h5_write(f, file, "f")
  h5_write(m, file, "m")
  
  expect_equal(h5_typeof(file, "f"), "enum")
  expect_equal(h5_typeof(file, "m"), "enum")
  expect_equal(h5_read(file, "f"), f)
  expect_equal(h5_read(file, "m"), m)
})

local({
  #test_that("Factors with NA throw error", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  f <- factor(c("a", NA), levels = c("a", "b"))
  expect_error(h5_write(f, file, "f"), "Factors with NA values cannot be written")
})

local({
  #test_that("Factor attributes work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  h5_write(1, file, "dset")
  f <- factor("low", levels = c("low", "high"))
  h5_write(f, file, "dset", attr = "quality")
  
  expect_equal(h5_read(file, "dset", attr = "quality"), f)
})

local({
  #test_that("Different magnitudes of levels", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  uids <- apply(expand.grid(letters, letters), 1L, paste0, collapse = '')
  
  h5_write(sm <- factor(uids[1:10], levels = uids[1:255]), file, 'enum_uint8')
  h5_write(md <- factor(uids[1:10], levels = uids[1:256]), file, 'enum_uint16')
  
  expect_equal(h5_read(file, "enum_uint8"),  sm)
  expect_equal(h5_read(file, "enum_uint16"), md)
})

local({
  #test_that("Coercion on read", {
  
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  h5_write(TRUE, file, "my_lgl")
  h5_write(2L,   file, "my_int")
  h5_write(3.1,  file, "my_dbl")
  h5_write(4L,   file, "my_big", as = "int64")
  h5_write(5L,   file, "obj")
  h5_write(6L,   file, "obj", "attr")
  
  expect_identical(h5_read(file, "my_lgl"), 1L)
  expect_identical(h5_read(file, "my_lgl", as = "integer"), 1L)
  expect_identical(h5_read(file, "my_lgl", as = "double"),  1.0)
  expect_identical(h5_read(file, "my_lgl", as = "logical"), TRUE)
  expect_identical(h5_read(file, "my_lgl", as = "null"),    NULL)
  expect_identical(h5_read(file, "my_lgl", as = c(my_lgl = "integer")), 1L)
  expect_identical(h5_read(file, "my_lgl", as = c(my_lgl = "auto")), 1L)
  
  expect_identical(h5_read(file, "my_int"), 2L)
  expect_identical(h5_read(file, "my_int", as = "integer"), 2L)
  expect_identical(h5_read(file, "my_int", as = "double"),  2.0)
  expect_identical(h5_read(file, "my_int", as = "logical"), TRUE)
  expect_identical(h5_read(file, "my_int", as = "null"),    NULL)
  expect_identical(h5_read(file, "my_int", as = c(my_int = "integer")), 2L)
  expect_identical(h5_read(file, "my_int", as = c(my_int = "auto")), 2L)
  
  expect_identical(h5_read(file, "my_dbl"), 3.1)
  expect_identical(h5_read(file, "my_dbl", as = "integer"), 3L)
  expect_identical(h5_read(file, "my_dbl", as = "double"),  3.1)
  expect_identical(h5_read(file, "my_dbl", as = "logical"), TRUE)
  expect_identical(h5_read(file, "my_dbl", as = "null"),    NULL)
  expect_identical(h5_read(file, "my_dbl", as = c(my_dbl = "integer")), 3L)
  expect_identical(h5_read(file, "my_dbl", as = c(my_dbl = "auto")), 3.1)
  
  expect_identical(h5_read(file, "my_big"), 4L)
  expect_identical(h5_read(file, "my_big", as = "integer"), 4L)
  expect_identical(h5_read(file, "my_big", as = "double"),  4.0)
  expect_identical(h5_read(file, "my_big", as = "logical"), TRUE)
  expect_identical(h5_read(file, "my_big", as = "null"),    NULL)
  expect_identical(h5_read(file, "my_big", as = c(my_big = "integer")), 4L)
  expect_identical(h5_read(file, "my_big", as = c(my_big = "auto")), 4L)
  
  expect_identical(h5_read(file, "my_lgl", as = c(my_lgl = "null")), NULL)
  expect_identical(h5_read(file, "my_lgl", as = c(.uint = "integer")), 1L)
  expect_identical(h5_read(file, "my_lgl", as = c(.     = "integer")), 1L)
  
  
  expect_identical(h5_read(file, "my_int", as = NULL), 2L)
  expect_error(h5_read(file, "my_int", as = 3))
  expect_error(h5_read(file, "my_int", as = structure("auto", names = '')))
  expect_error(h5_read(file, "my_int", as = "invalid"))
  expect_silent(h5_read(file, "obj", as = c('obj' = "integer")))
  
  
  if (requireNamespace("bit64", quietly=TRUE)) {
    #skip_if_not_installed('bit64')
    b64 <- bit64::as.integer64(4L)
    expect_identical(h5_read(file, "my_big", as = "bit64"), b64)
    expect_identical(h5_read(file, "my_big", as = c(my_big = "bit64")), b64)
  }
})


local({
  #test_that("Coercion on write", {
  
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  expect_error(h5_write(TRUE,        file, "my_lgl",   as = c("int", "uint")))
  expect_error(h5_write(NA_integer_, file, "NA_vals",  as = "int8"))
  expect_error(h5_write(Inf,         file, "Inf_vals", as = "int8"))
  suppressWarnings(expect_warning(h5_write(I(1:3), file, "non-scalar")))
  
  h5_write(Sys.time(), file, 'now')
  h5_write(data.frame(x=1:5, y=6:10), file, "df", as = c(x = "skip"))
  h5_write(integer(0), file, "zerolen", as = "float16")
  h5_write(integer(0), file, "zerolen", as = "bfloat16")
  
  lst <- list(vec = 1:5, nil = NULL)
  attr(lst$vec, 'attr_lst') <- list(z = 1)
  attr(lst$vec, 'a') <- as.raw(1)
  attr(lst$vec, 'b') <- 2.0
  attr(lst$vec, 'c') <- TRUE
  attr(lst$vec, 'd') <- data.frame(x=1:3)
  h5_write(lst, file, "lst", 
           as = c('@.' = "auto", '@.integer' = "int64", '@c' = "uint8", '@raw' = "skip"))
  
  h5_write(c(0, 0      ), file, "uint8_lo")
  h5_write(c(0, 2^8-1  ), file, "uint8_hi")
  h5_write(c(0, 2^8    ), file, "uint16_lo")
  h5_write(c(0, 2^16-1 ), file, "uint16_hi")
  h5_write(c(0, 2^16   ), file, "uint32_lo")
  h5_write(c(0, 2^32-1 ), file, "uint32_hi")
  h5_write(c(0, 2^32   ), file, "uint64_lo")
  h5_write(c(0, 2^53-1 ), file, "uint64_hi")
  
  expect_identical(h5_typeof(file, "uint8_lo"),  "uint8")
  expect_identical(h5_typeof(file, "uint8_hi"),  "uint8")
  expect_identical(h5_typeof(file, "uint16_lo"), "uint16")
  expect_identical(h5_typeof(file, "uint16_hi"), "uint16")
  expect_identical(h5_typeof(file, "uint32_lo"), "uint32")
  expect_identical(h5_typeof(file, "uint32_hi"), "uint32")
  expect_identical(h5_typeof(file, "uint64_lo"), "uint64")
  expect_identical(h5_typeof(file, "uint64_hi"), "uint64")
  
  h5_write(c(-1, -2^7    ), file, "int8_lo")
  h5_write(c(-1,  2^7-1  ), file, "int8_hi")
  h5_write(c(-1, -2^15   ), file, "int16_lo")
  h5_write(c(-1,  2^15-1 ), file, "int16_hi")
  h5_write(c(-1, -2^31   ), file, "int32_lo")
  h5_write(c(-1,  2^31-1 ), file, "int32_hi")
  h5_write(c(-1, -2^53+1 ), file, "int64_lo")
  h5_write(c(-1,  2^53-1 ), file, "int64_hi")
  
  expect_identical(h5_typeof(file, "int8_lo"),  "int8")
  expect_identical(h5_typeof(file, "int8_hi"),  "int8")
  expect_identical(h5_typeof(file, "int16_lo"), "int16")
  expect_identical(h5_typeof(file, "int16_hi"), "int16")
  expect_identical(h5_typeof(file, "int32_lo"), "int32")
  expect_identical(h5_typeof(file, "int32_hi"), "int32")
  expect_identical(h5_typeof(file, "int64_lo"), "int64")
  expect_identical(h5_typeof(file, "int64_hi"), "int64")
  
  h5_write(c(Inf, -2^24   ), file, "float32_lo")
  h5_write(c(Inf,  2^24   ), file, "float32_hi")
  h5_write(c(Inf, -2^24-1 ), file, "float64_a")
  h5_write(c(Inf,  2^24+1 ), file, "float64_b")
  h5_write(c(NA,   2^24+1 ), file, "float64_c")
  h5_write(        2^53,     file, "float64_d")
  
  expect_identical(h5_typeof(file, "float32_lo"), "float32")
  expect_identical(h5_typeof(file, "float32_hi"), "float32")
  expect_identical(h5_typeof(file, "float64_a"),  "float64")
  expect_identical(h5_typeof(file, "float64_b"),  "float64")
  expect_identical(h5_typeof(file, "float64_c"),  "float64")
  expect_identical(h5_typeof(file, "float64_d"),  "float64")
})

local({
  #test_that("Complex scalars and vectors work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  vec <- c(1+2i, 3+4i)
  h5_write(vec, file, "vec")
  expect_equal(h5_typeof(file, "vec"), "complex")
  expect_equal(h5_read(file, "vec"), vec)
  
  h5_write(I(1+1i), file, "scalar")
  expect_equal(h5_read(file, "scalar"), 1+1i)
})

local({
  #test_that("Complex matrices work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  mat <- matrix(c(1i, 2i, 3i, 4i), 2, 2)
  h5_write(mat, file, "mat")
  expect_equal(h5_read(file, "mat"), mat)
})

local({
  #test_that("Complex attributes work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  h5_write(1, file, "dset")
  h5_write(0+1i, file, "dset", attr = "imag")
  expect_equal(h5_read(file, "dset", attr = "imag"), 0+1i)
})

local({
  #test_that("Lists (Groups) work recursively", {
  
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  lst <- list(
    a = 1:5,
    b = list(sub = "hello"),
    c = matrix(1:4, 2, 2)
  )
  attr(lst, "meta") <- "top_level"
  
  h5_write(lst, file, "lst")
  expect_true(h5_is_group(file, "lst"))
  expect_true(h5_is_group(file, "lst/b"))
  
  res <- h5_read(file, "lst")
  expect_equal(res$a, lst$a)
  expect_equal(res$b$sub, lst$b$sub)
  expect_equal(res$c, lst$c)
  expect_equal(attr(res, "meta"), "top_level")
})


local({
  #test_that("Mappings apply per list element/attr", {
  
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  lst <- list(
    a = "alpha",
    b = bit64::as.integer64(10),
    c = data.frame(x = 1:5, y = letters[1:5]),
    d = NULL,
    e = 123,
    z = list(z1 = "omega")
  )
  attr(lst$a, "f") <- data.frame(i = 6:9, j = letters[6:9])
  attr(lst$a, "g") <- NA_integer_
  attr(lst$a, "h") <- 20L
  
  h5_write(lst, file, "lst", as = c('@integer' = "uint8", 'e' = "skip"))
  
  res <- h5_read(file, "lst", as = c('.int64' = "bit64"))
  lst$e <- NULL
  expect_equal(res, lst)
  
})

local({
  #test_that("Logical scalars and vectors work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))

  # 1. Vector
  vec <- c(TRUE, FALSE, TRUE)
  h5_write(vec, file, "vec")
  
  # Default read is integer (0/1)
  expect_equal(h5_read(file, "vec"), c(1, 0, 1))
  # Explicit read as logical
  expect_equal(h5_read(file, "vec", as = "logical"), vec)
  
  # 2. Scalar
  h5_write(TRUE, file, "scalar")
  expect_equal(h5_read(file, "scalar", as = "logical"), TRUE)
})

local({
  #test_that("Logical matrices work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  mat <- matrix(c(TRUE, FALSE, FALSE, TRUE), 2, 2)
  h5_write(mat, file, "matrix")
  expect_equal(h5_read(file, "matrix", as = "logical"), mat)
})

local({
  #test_that("Logical attributes work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  h5_write(1, file, "dset")
  h5_write(TRUE, file, "dset", attr = "is_valid")
  expect_equal(h5_read(file, "dset", attr = "is_valid", as = "logical"), TRUE)
})

local({
  #test_that("NAs in logicals are handled", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  # Logical NAs are integer NA_integer_ internally in R, so h5lite sees int NA
  # It promotes to float to store NaN unless we force integer/logical
  vec <- c(TRUE, NA, FALSE)
  
  # Default write (auto) -> Promotes to float because of NA
  h5_write(vec, file, "auto")
  expect_true(startsWith(h5_typeof(file, "auto"), "float"))
  expect_equal(h5_read(file, "auto", as = "logical"), vec)
})

local({
  #test_that("Character scalars work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  h5_write(I("hello"), file, "scalar")
  expect_equal(h5_read(file, "scalar"), "hello")
  
  h5_write(I("hello"), file, "scalar", as = "ascii")
  expect_equal(h5_read(file, "scalar"), "hello")
  
  h5_write(I("hello"), file, "scalar", as = "utf")
  expect_equal(h5_read(file, "scalar"), "hello")
  
  h5_write(I("hello"), file, "scalar", as = "ascii[]")
  expect_equal(h5_read(file, "scalar"), "hello")
  
  h5_write(I("hello"), file, "scalar", as = "utf[]")
  expect_equal(h5_read(file, "scalar"), "hello")
  
  h5_write(I("hello"), file, "scalar", as = "ascii[10]")
  expect_equal(h5_read(file, "scalar"), "hello")
  
  h5_write(I("hello"), file, "scalar", as = "utf[10]")
  expect_equal(h5_read(file, "scalar"), "hello")
})


local({
  #test_that("Character vectors work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  vec <- c("apple", "banana", "cherry")
  
  h5_write(vec, file, "vec")
  expect_equal(h5_read(file, "vec"), vec)
  
  h5_write(vec, file, "vec", as = "ascii")
  expect_equal(h5_read(file, "vec"), vec)
  
  h5_write(vec, file, "vec", as = "utf")
  expect_equal(h5_read(file, "vec"), vec)
  
  h5_write(vec, file, "vec", as = "ascii[]")
  expect_equal(h5_read(file, "vec"), vec)
  
  h5_write(vec, file, "vec", as = "utf[]")
  expect_equal(h5_read(file, "vec"), vec)
  
  h5_write(vec, file, "vec", as = "ascii[10]")
  expect_equal(h5_read(file, "vec"), vec)
  
  h5_write(vec, file, "vec", as = "utf[10]")
  expect_equal(h5_read(file, "vec"), vec)
})


local({
  #test_that("Character matrices work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  mat <- matrix(letters[1:4], 2, 2)
  
  h5_write(mat, file, "matrix")
  expect_equal(h5_read(file, "matrix"), mat)
  
  h5_write(mat, file, "matrix", as = "ascii")
  expect_equal(h5_read(file, "matrix"), mat)
  
  h5_write(mat, file, "matrix", as = "ascii[]")
  expect_equal(h5_read(file, "matrix"), mat)
  
  h5_write(mat, file, "matrix", as = "ascii[5]")
  expect_equal(h5_read(file, "matrix"), mat)
  
  h5_write(mat, file, "matrix", as = "utf8")
  expect_equal(h5_read(file, "matrix"), mat)
  
  h5_write(mat, file, "matrix", as = "utf8[]")
  expect_equal(h5_read(file, "matrix"), mat)
  
  h5_write(mat, file, "matrix", as = "utf8[5]")
  expect_equal(h5_read(file, "matrix"), mat)
})


local({
  #test_that("Character attributes work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  h5_write(1, file, "dset")
  
  h5_write("metadata", file, "dset", attr = "info")
  expect_equal(h5_read(file, "dset", attr = "info"), "metadata")
  
  h5_write("metadata", file, "dset", attr = "info", as = "utf8")
  expect_equal(h5_read(file, "dset", attr = "info"), "metadata")
  
  h5_write("metadata", file, "dset", attr = "info", as = "utf8[]")
  expect_equal(h5_read(file, "dset", attr = "info"), "metadata")
  
  h5_write("metadata", file, "dset", attr = "info", as = "utf8[10]")
  expect_equal(h5_read(file, "dset", attr = "info"), "metadata")
  
  h5_write("metadata", file, "dset", attr = "info", as = "ascii")
  expect_equal(h5_read(file, "dset", attr = "info"), "metadata")
  
  h5_write("metadata", file, "dset", attr = "info", as = "ascii[]")
  expect_equal(h5_read(file, "dset", attr = "info"), "metadata")
  
  h5_write("metadata", file, "dset", attr = "info", as = "ascii[10]")
  expect_equal(h5_read(file, "dset", attr = "info"), "metadata")
})

local({
  #test_that("UTF-8 characters are preserved", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  utf8_str  <- "Zürich"
  ascii_str <- "Zurich"
  
  h5_write(utf8_str, file, "city")
  expect_equal(h5_read(file, "city"), utf8_str)
  
  h5_write(utf8_str, file, "city", as = "utf8")
  expect_equal(h5_read(file, "city"), utf8_str)
  
  h5_write(utf8_str, file, "city", as = "utf8[]")
  expect_equal(h5_read(file, "city"), utf8_str)
  
  h5_write(utf8_str, file, "city", as = "utf8[20]")
  expect_equal(h5_read(file, "city"), utf8_str)
  
  h5_write(utf8_str, file, "city", as = 'ascii')
  expect_equal(h5_read(file, "city"), ascii_str)
  
  h5_write(utf8_str, file, "city", as = 'ascii[]')
  expect_equal(h5_read(file, "city"), ascii_str)
  
  h5_write(utf8_str, file, "city", as = 'ascii[20]')
  expect_equal(h5_read(file, "city"), ascii_str)
})

local({
  #test_that("Character errors work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  expect_error(h5_write('A',        file, "x", as = 'invalid'))
  expect_error(h5_write(c('A', NA), file, "x", as = 'utf[]'))
  expect_error(h5_write(c('A', NA), file, "x", as = 'utf[10]'))
  expect_error(h5_write(c('A', NA), file, "x", as = 'ascii[]'))
  expect_error(h5_write(c('A', NA), file, "x", as = 'ascii[10]'))
})

local({
  #test_that("Raw scalars and vectors work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  vec <- as.raw(c(0x00, 0xFF, 0x1A))
  
  h5_write(vec, file, "vec")
  expect_equal(h5_read(file, "vec"), vec)
  expect_equal(h5_typeof(file, "vec"), "opaque")
  
  h5_write(vec, file, "vec", "attr")
  expect_equal(h5_typeof(file, "vec", "attr"), "opaque")
  expect_equal(h5_read(file, "vec", "attr"), vec)
})

local({
  #test_that("Raw matrices work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  mat <- matrix(as.raw(1:4), 2, 2)
  h5_write(mat, file, "mat")
  expect_equal(h5_read(file, "mat"), mat)
})

local({
  #test_that("h5_inspect validates inputs and handles errors", {
  file <- tempfile(fileext = ".h5")
  
  # Create a group (not a dataset)
  h5_create_group(file, "my_group")
  
  # Ensure it throws the expected error when inspecting a group
  expect_error(
    h5_inspect(file, "my_group"), 
    "Target must be a dataset"
  )
  
  unlink(file)
})

local({
  #test_that("h5_inspect handles uncompressed scalars and print formatting", {
  file <- tempfile(fileext = ".h5")
  
  # Rank-0 Scalar (Bypasses chunking and compression entirely)
  h5_write(I(42L), file, "scalar_val")
  res <- h5_inspect(file, "scalar_val")
  
  expect_inherits(res, "inspect")
  expect_inherits(res, "list")
  expect_equal(res$layout, "contiguous") # Rank 0 defaults to contiguous
  expect_null(res$chunk_dims)
  expect_equal(length(res$filters), 0)
  
  out <- capture.output(print(res))
  expect_true(any(grepl("None \\(Uncompressed\\)", out)))
  expect_true(any(grepl("Chunks:\\s+N/A", out)))
  
  # --- Mock Size Formatting to Hit 100% Coverage on fmt_bytes ---
  res$uncompressed_size <- 0
  res$storage_size      <- 500
  out_b <- capture.output(print(res))
  expect_true(any(grepl("0 B", out_b)))
  expect_true(any(grepl("500.00 B", out_b)))
  
  res$uncompressed_size <- 1024 * 15.5      # 15.5 KB
  res$storage_size      <- 1024^2 * 4.2     # 4.2 MB
  out_mb <- capture.output(print(res))
  expect_true(any(grepl("15.50 KB", out_mb)))
  expect_true(any(grepl("4.20 MB", out_mb)))
  
  res$uncompressed_size <- 1024^3 * 1.8     # 1.8 GB
  res$storage_size      <- 1024^4 * 2.1     # 2.1 TB
  out_tb <- capture.output(print(res))
  expect_true(any(grepl("1.80 GB", out_tb)))
  expect_true(any(grepl("2.10 TB", out_tb)))
  
  unlink(file)
})

local({
  #test_that("h5_inspect detects standard compressors and fletcher32", {
  file <- tempfile(fileext = ".h5")
  dat <- matrix(rnorm(400), 20, 20)
  
  # Standard GZIP with Checksum
  opts <- h5_compression("gzip-6", checksum = TRUE)
  h5_write(dat, file, "gzip_chk", compress = opts)
  
  res <- h5_inspect(file, "gzip_chk")
  expect_equal(res$layout, "chunked")
  
  out <- capture.output(print(res))
  expect_true(any(grepl("gzip -> fletcher32", out)))
  
  unlink(file)
})

local({
  #test_that("h5_inspect decodes SZIP variations", {
  file <- tempfile(fileext = ".h5")
  dat <- matrix(1L:2500L, 50, 50)
  
  h5_write(dat, file, "szip_ec", compress = "szip-ec")
  out_ec <- capture.output(print(h5_inspect(file, "szip_ec")))
  expect_true(any(grepl("szip-ec", out_ec)))
  
  h5_write(dat, file, "szip_nn", compress = "szip-nn")
  out_nn <- capture.output(print(h5_inspect(file, "szip_nn")))
  expect_true(any(grepl("szip-nn", out_nn)))
  
  unlink(file)
})

local({
  #test_that("h5_inspect decodes Bitshuffle internal pipelines", {
  file <- tempfile(fileext = ".h5")
  dat <- matrix(1L:2500L, 50, 50)
  
  h5_write(dat, file, "bshuf_lz4", compress = "bshuf-lz4")
  out_lz4 <- capture.output(print(h5_inspect(file, "bshuf_lz4")))
  expect_true(any(grepl("bitshuffle -> lz4", out_lz4)))
  
  h5_write(dat, file, "bshuf_zstd", compress = "bshuf-zstd-3")
  out_zstd <- capture.output(print(h5_inspect(file, "bshuf_zstd")))
  expect_true(any(grepl("bitshuffle -> zstd", out_zstd)))
  
  unlink(file)
})

local({
  #test_that("h5_inspect decodes Blosc1 and Blosc2 pre-filters", {
  file <- tempfile(fileext = ".h5")
  dat <- matrix(rnorm(2500), 50, 50)
  
  # Blosc1 (Legacy)
  h5_write(dat, file, "blosc1", compress = "blosc1-lz4-5")
  out_b1 <- capture.output(print(h5_inspect(file, "blosc1")))
  expect_true(any(grepl("blosc \\[.*shuffle -> lz4hc\\]", out_b1)))
  
  # Blosc2 with Delta filter
  opts_delta <- h5_compression("blosc2-zstd-5", blosc2_delta = TRUE)
  h5_write(dat, file, "b2_delta", compress = opts_delta)
  out_b2d <- capture.output(print(h5_inspect(file, "b2_delta")))
  expect_true(any(grepl("blosc2 \\[bitshuffle -> delta -> zstd\\]", out_b2d)))
  
  # Blosc2 with Truncate filter
  opts_trunc <- h5_compression("blosc2-ndlz", blosc2_truncate = 10)
  h5_write(dat, file, "b2_trunc", compress = opts_trunc)
  out_b2t <- capture.output(print(h5_inspect(file, "b2_trunc")))
  expect_true(any(grepl("trunc\\(bits=10\\)", out_b2t)))
  expect_true(any(grepl("ndlz", out_b2t)))
  
  unlink(file)
})

local({
  #test_that("h5_inspect decodes ZFP standalone and Blosc2 modes", {
  file <- tempfile(fileext = ".h5")
  dat <- matrix(rnorm(2500), 50, 50)
  
  # Standalone ZFP Rate
  h5_write(dat, file, "zfp_rate", compress = "zfp-rate-16")
  out_rate <- capture.output(print(h5_inspect(file, "zfp_rate")))
  # Standalone ZFP does not get the bracket expansion
  expect_true(any(grepl("Pipeline: zfp", out_rate))) 
  
  # ZFP embedded inside Blosc2
  h5_write(dat, file, "b2_zfp", compress = "blosc2-zfp-prec-12")
  out_b2zfp <- capture.output(print(h5_inspect(file, "b2_zfp")))
  expect_true(any(grepl("blosc2 \\[zfp-prec\\]", out_b2zfp)))
  
  unlink(file)
})

local({
  #test_that("h5_inspect detects Scale-Offset integer and float packing", {
  file <- tempfile(fileext = ".h5")
  
  # Integer Packing
  dat_int <- matrix(1L:1000L, 50, 20)
  opts_int <- h5_compression("gzip-5", int_packing = 16)
  h5_write(dat_int, file, "so_int", compress = opts_int)
  out_int <- capture.output(print(h5_inspect(file, "so_int")))
  expect_true(any(grepl("scaleoffset -> gzip", out_int)))
  
  # Float Rounding
  dat_flt <- matrix(rnorm(1000), 50, 20)
  opts_flt <- h5_compression("zstd-3", float_rounding = 3)
  h5_write(dat_flt, file, "so_flt", compress = opts_flt)
  out_flt <- capture.output(print(h5_inspect(file, "so_flt")))
  expect_true(any(grepl("scaleoffset -> zstd", out_flt)))
  
  unlink(file)
})

local({
  #test_that("h5_inspect decodes legacy and unconfigurable compressors", {
  file <- tempfile(fileext = ".h5")
  dat <- matrix(1L:2500L, 50, 50)
  
  h5_write(dat, file, "lzf", compress = "lzf")
  expect_true(any(grepl("lzf", capture.output(print(h5_inspect(file, "lzf"))))))
  
  h5_write(dat, file, "snappy", compress = "snappy")
  expect_true(any(grepl("snappy", capture.output(print(h5_inspect(file, "snappy"))))))
  
  h5_write(dat, file, "bzip2", compress = "bzip2-4")
  expect_true(any(grepl("bzip2", capture.output(print(h5_inspect(file, "bzip2"))))))
  
  unlink(file)
})

local({
  #test_that("validate_start_count handles missing count correctly", {
  file <- tempfile(fileext = ".h5")
  v <- 1:10
  h5_write(v, file, "v")
  h5_write(1L, file, "v", attr = "my_attr")
  
  # 1. Start is provided, count is NULL (Valid exact index)
  expect_silent(h5_read(file, "v", start = 1))
  
  # 2. Count is provided, start is NULL (Invalid)
  expect_error(h5_read(file, "v", count = 1), "must be provided if `count` is specified")
  
  # 3. Attribute check
  expect_error(h5_read(file, "v", attr = "my_attr", start = 1), 
               "cannot be used on attributes")
  
  # 4. Numeric checks
  expect_error(h5_read(file, "v", start = "a"), "must be numeric")
  expect_error(h5_read(file, "v", start = 1, count = "a"), "must be numeric")
  
  # 5. Positivity checks
  expect_error(h5_read(file, "v", start = 0), "must be positive")
  expect_error(h5_read(file, "v", start = c(1, -1)), "must be positive")
  expect_error(h5_read(file, "v", start = 1, count = 0), "must be positive")
  
  # 6. Bounds checks
  expect_error(h5_read(file, "v", start = 11), "out of bounds")
  expect_error(h5_read(file, "v", start = 10, count = 2), "out of bounds")
  
  # 7. Other errors
  expect_error(h5_read(file, "v", start = numeric(0)), "cannot be an empty vector")
  expect_error(h5_read(file, "v", start = 1, count = 1:2), "must be a single numeric value")
  expect_error(h5_read(file, "v", start = 1, count = numeric(0)), "must be a single numeric value")
  expect_error(h5_read(file, "v", start = 1:2), "more dimensions than the dataset")
  
  unlink(file)
})

local({
  #test_that("partial reading works for 1D atomic vectors (with naming logic)", {
  file <- tempfile(fileext = ".h5")
  
  # Scalar
  h5_write(I(42), file, "scl")
  
  res0 <- h5_read(file, "scl", start = 1)
  expect_equal(res0, 42)
  
  res0 <- h5_read(file, "scl", start = 1, count = 1)
  expect_equal(res0, 42)
  
  # Setup Named Vector
  v_int <- c(a=1L, b=2L, c=3L, d=4L, e=5L)
  h5_write(v_int, file, "int")
  
  # 1. Range Indexing (count > 1)
  res1 <- h5_read(file, "int", start = 2, count = 3)
  expect_equal(res1, c(b=2L, c=3L, d=4L))
  
  # 2. Range Indexing of 1 (count = 1) -> retains name
  res2 <- h5_read(file, "int", start = 3, count = 1)
  expect_equal(res2, c(c=3L))
  
  # 3. Exact Point Indexing (count = NULL) -> strips name
  res3 <- h5_read(file, "int", start = 3)
  expect_equal(res3, 3L)
  expect_null(names(res3))
  
  unlink(file)
})

local({
  #test_that("partial reading works for matrices (with dimension dropping)", {
  file <- tempfile(fileext = ".h5")
  
  # 5x4 matrix
  m <- matrix(1:20, nrow = 5, ncol = 4, 
              dimnames = list(paste0("r", 1:5), paste0("c", 1:4)))
  h5_write(m, file, "m")
  
  # 1. Range Indexing (rows)
  res1 <- h5_read(file, "m", start = 3, count = 2)
  expect_equal(dim(res1), c(2, 4))
  expect_equal(rownames(res1), c("r3", "r4"))
  expect_equal(colnames(res1), c("c1", "c2", "c3", "c4"))
  
  # 2. Exact Indexing (row) -> drops matrix to vector, inherits column names
  res2 <- h5_read(file, "m", start = 3)
  expect_null(dim(res2))
  expect_equal(names(res2), c("c1", "c2", "c3", "c4"))
  expect_equal(res2, m[3, ])
  
  # 3. Range Indexing of 1 (row) -> preserves 1xN matrix, inherits row name
  res3 <- h5_read(file, "m", start = 3, count = 1)
  expect_equal(dim(res3), c(1, 4))
  expect_equal(rownames(res3), "r3")
  expect_equal(colnames(res3), c("c1", "c2", "c3", "c4"))
  
  # 4. Multi-value Exact+Range (row exactly, cols range) -> drops row, named vector
  res4 <- h5_read(file, "m", start = c(3, 2), count = 2)
  expect_null(dim(res4))
  expect_equal(names(res4), c("c2", "c3"))
  expect_equal(res4, m[3, 2:3])
  
  # 5. Multi-value Exact (row exactly, col exactly) -> fully unnamed scalar
  res5 <- h5_read(file, "m", start = c(3, 2))
  expect_null(dim(res5))
  expect_null(names(res5))
  expect_equal(res5, m[3, 2][[1]])
  
  unlink(file)
})

local({
  #test_that("partial reading works for data.frames (no dropping allowed)", {
  file <- tempfile(fileext = ".h5")
  
  df <- data.frame(
    id = 1:5,
    val = c(1.1, 2.2, 3.3, 4.4, 5.5),
    name = c("a", "b", "c", "d", "e"),
    row.names = c("r1", "r2", "r3", "r4", "r5"),
    stringsAsFactors = FALSE
  )
  h5_write(df, file, "df")
  
  # 1. Range Indexing
  res1 <- h5_read(file, "df", start = 2, count = 3)
  expect_inherits(res1, "data.frame")
  expect_equal(nrow(res1), 3)
  expect_equal(rownames(res1), c("r2", "r3", "r4"))
  
  # 2. Exact Indexing (count = NULL) -> Data Frames should never drop!
  res2 <- h5_read(file, "df", start = 2)
  expect_inherits(res2, "data.frame")
  expect_equal(nrow(res2), 1)
  expect_equal(rownames(res2), "r2")
  expect_equal(res2$val, 2.2)
  
  unlink(file)
})

local({
  #test_that("partial reading works for named N-dimensional arrays", {
  file <- tempfile(fileext = ".h5")
  
  # Create a 2x3x4x5 array (Rows=2, Cols=3, 3rd=4, 4th=5)
  arr4 <- array(
    data     = 1:120,
    dim      = c(2, 3, 4, 5), 
    dimnames = list(LETTERS[1:2], LETTERS[3:5], LETTERS[6:9], LETTERS[10:14]) )
  h5_write(arr4, file, "arr4")
  
  # 1. Target 4th=2, 3rd=3, Row=1. Read 2 rows (count=2).
  # Exact point indices: 4th, 3rd. (Dropped)
  # Range target: Rows. (Preserved)
  # Untargeted: Cols. (Preserved)
  # Result: 2x3 matrix
  res1 <- h5_read(file, "arr4", start = c(2, 3, 1), count = 2)
  expect_equal(dim(res1), c(2, 3))
  expect_equal(res1, arr4[1:2,, 3, 2, drop=TRUE])
  
  # 2. Exact index deep down: Target 4th=2, 3rd=3, Row=2, Col=1. (count=NULL)
  # Drops all dimensions -> scalar
  res2 <- h5_read(file, "arr4", start = c(2, 3, 2, 1))
  expect_null(dim(res2))
  expect_equal(res2, arr4[2, 1, 3, 2])
  
  # 3. Same target as above, but Range index (count=1).
  # Drops 4th, 3rd, Rows. Preserves Cols.
  # Result: 1D vector of length 1
  res3 <- h5_read(file, "arr4", start = c(2, 3, 2, 1), count = 1)
  expect_equal(length(res3), 1)
  expect_equal(names(res3), 'C')
  expect_equal(unname(res3), arr4[2, 1, 3, 2])
  
  unlink(file)
})

local({
  #test_that("partial reading works for unnamed N-dimensional arrays", {
  file <- tempfile(fileext = ".h5")
  
  # Create a 2x3x4x5 array (Rows=2, Cols=3, 3rd=4, 4th=5)
  arr4 <- array(data = 1:120, dim = c(2, 3, 4, 5))
  h5_write(arr4, file, "arr4")
  
  # 1. Target 4th=2, 3rd=3, Row=1. Read 2 rows (count=2).
  # Exact point indices: 4th, 3rd. (Dropped)
  # Range target: Rows. (Preserved)
  # Untargeted: Cols. (Preserved)
  # Result: 2x3 matrix
  res1 <- h5_read(file, "arr4", start = c(2, 3, 1), count = 2)
  expect_equal(dim(res1), c(2, 3))
  expect_equal(res1, arr4[1:2,, 3, 2, drop=TRUE])
  
  # 2. Exact index deep down: Target 4th=2, 3rd=3, Row=2, Col=1. (count=NULL)
  # Drops all dimensions -> scalar
  res2 <- h5_read(file, "arr4", start = c(2, 3, 2, 1))
  expect_null(dim(res2))
  expect_equal(res2, arr4[2, 1, 3, 2])
  
  # 3. Same target as above, but Range index (count=1).
  # Drops 4th, 3rd, Rows. Preserves Cols.
  # Result: 1D vector of length 1
  res3 <- h5_read(file, "arr4", start = c(2, 3, 2, 1), count = 1)
  expect_equal(length(res3), 1)
  expect_equal(names(res3), NULL)
  expect_equal(res3, arr4[2, 1, 3, 2])
  
  unlink(file)
})

#' Read an HDF5 Object or Attribute
#'
#' Reads a dataset, a group, or a specific attribute from an HDF5 file into an R object.
#' Supports partial reading (hyperslabs) to load specific subsets of data without 
#' loading the entire object into memory.
#'
#' @param file The path to the HDF5 file.
#' @param name The full path of the dataset or group to read (e.g., `"/data/matrix"`).
#' @param attr The name of an attribute to read.
#'   * If `NULL` (default), the function reads the object specified by `name` (and attaches its attributes to the result).
#'   * If provided (string), the function reads *only* the specified attribute from `name`.
#' @param as The target R data type.
#'   * **Global:** `"auto"` (default), `"integer"`, `"double"`, `"logical"`, `"bit64"`, `"null"`.
#'   * **Specific:** A named vector mapping names or type classes to R types (see Section "Type Conversion").
#' @param start A numeric vector specifying the 1-based coordinate(s) for a partial read.
#'   Most often, this is a **single value** targeting the most logical structural unit 
#'   (e.g., the row of a matrix, or the 2D matrix of a 3D array). 
#'   If `NULL` (default), the entire dataset is read.
#' @param count A single numeric value specifying the number of elements or units to read.
#'   If `NULL` (default) and `start` is provided, `h5lite` reads exactly 1 unit and 
#'   simplifies the resulting dimensions (see Section "Dimension Simplification").
#'
#' @section Partial Reading (Hyperslabs):
#' You can read specific subsets of an n-dimensional dataset by utilizing the `start` 
#' and `count` arguments.
#' 
#' **The "Smart" `start` Parameter**
#' 
#' `start` is designed to be intuitive. Most of the time, you only need to provide a single value. 
#' This single value automatically targets the most meaningful dimension of the dataset:
#' 
#' * **1D Vector:** `start` specifies the **element**.
#' * **2D Matrix / Data Frame:** `start` specifies the **row**.
#' * **3D Array:** `start` specifies the **2D matrix**.
#' 
#' The `count` parameter is a **single value** that determines how many of those units 
#' to read sequentially. For example, `start = 5` and `count = 3` on a matrix will read 3 complete 
#' rows starting at row 5 (automatically spanning all columns).
#' 
#' **Multi-Value `start` and N-Dimensional Arrays**
#' 
#' If you need to extract a specific block *inside* a structural unit, you can provide a vector of 
#' values to `start`. To make indexing intuitive across higher-order arrays, `start` maps 
#' its values to dimensions in the following priority order, targeting the outermost blocks first 
#' and specific rows/columns last:
#' 
#' * `N, N-1, ..., 3, 1 (Rows), 2 (Cols)`
#' 
#' For example, on a 3D array, `start = c(2, 5)` targets the 2nd matrix, and the 5th row. 
#' The `count` argument always applies to the **last** dimension specified in `start`.
#' 
#' **Dimension Simplification (Dropping)**
#' 
#' `h5lite` mimics R's native subsetting behavior regarding dimension preservation:
#' 
#' * **Exact Indexing (`count = NULL`):** If you provide `start` but omit `count`, `h5lite` 
#'   assumes you are targeting an exact point index. It will read 1 unit and **drop** the 
#'   targeted dimension. (e.g., reading a specific row of a matrix will return a 1D vector).
#' * **Range Indexing (`count` provided):** If you explicitly provide `count` (even `count = 1`), 
#'   `h5lite` assumes you are reading a range. The dataset's original structural geometry is 
#'   **preserved**. (e.g., reading `start = 5, count = 1` on a matrix will return a 1xN matrix).
#'
#' @section Type Conversion (`as`):
#' You can control how HDF5 data is converted to R types using the `as` argument.
#'
#' **1. Mapping by Name:**
#' \itemize{
#'   \item `as = c("data_col" = "integer")`: Reads the dataset/column named "data_col" as an integer.
#'   \item `as = c("@validated" = "logical")`: When reading a dataset, this forces the attached attribute "validated" to be read as logical.
#' }
#'
#' **2. Mapping by HDF5 Type Class:**
#' You can target specific HDF5 data types using keys prefixed with a dot (`.`).
#' Supported classes include:
#' \itemize{
#'   \item **Integer:** `.int`, `.int8`, `.int16`, `.int32`, `.int64`
#'   \item **Unsigned:** `.uint`, `.uint8`, `.uint16`, `.uint32`, `.uint64`
#'   \item **Floating Point:** `.float`, `.float16`, `.float32`, `.float64`
#' }
#' Example: `as = c(.uint8 = "logical", .int = "bit64")`
#'
#' **3. Precedence & Attribute Config:**
#' 
#' * **Attributes vs Datasets:** Attribute type mappings take precedence over dataset mappings.
#'   If you specify `as = c(.uint = "logical", "@.uint" = "integer")`, unsigned integer datasets
#'   will be read as `logical`, but unsigned integer *attributes* will be read as `integer`.
#' * **Specific vs Generic:** Specific keys (e.g., `.uint32`) take precedence over generic keys (e.g., `.uint`),
#'   which take precedence over the global default (`.`).
#'
#' @note
#' The `@` prefix is **only** used to configure attached attributes when reading a dataset (`attr = NULL`).
#' If you are reading a specific attribute directly (e.g., `h5_read(..., attr = "id")`), do **not** use
#' the `@` prefix in the `as` argument.
#' 
#' Partial reading (`start`/`count`) is currently only supported for datasets, not attributes.
#'
#' @return An R object corresponding to the HDF5 object or attribute.
#'   Returns `NULL` if the object is skipped via `as = "null"`.
#' 
#' @seealso [h5_write()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' # --- Setup: Write Test Data ---
#' h5_write(c(10L, 20L, 30L, 40L, 50L), file, "ints")
#' 
#' m <- matrix(1:50, nrow = 10, ncol = 5, dimnames = list(paste0("r", 1:10), paste0("c", 1:5)))
#' h5_write(m, file, "matrix_data")
#' 
#' arr <- array(1:24, dim = c(2, 3, 4))
#' h5_write(arr, file, "array_data")
#' 
#' # --- Standard Reading ---
#' # Read the entire dataset
#' x <- h5_read(file, "ints")
#' 
#' # --- Type Conversion ---
#' # Force integer dataset to be read as numeric (double)
#' x_dbl <- h5_read(file, "ints", as = "double")
#' class(x_dbl)
#' 
#' # --- Partial Reading: Single-Value 'start' ---
#' # Vector: Start at 2nd element, read 3 elements
#' h5_read(file, "ints", start = 2, count = 3)
#' 
#' # Matrix: Start at row 5, read 3 complete rows (returns 3x5 matrix)
#' h5_read(file, "matrix_data", start = 5, count = 3)
#' 
#' # 3D Array: Start at 2nd matrix, read 2 complete matrices (returns 2x3x2 array)
#' h5_read(file, "array_data", start = 2, count = 2)
#' 
#' # --- Partial Reading: Dimension Simplification ---
#' # Omit 'count' to extract an exact point index and drop the targeted dimension
#' 
#' # Matrix: Extract exactly row 5 (drops row dimension, returns a 1D vector)
#' h5_read(file, "matrix_data", start = 5)
#' 
#' # Matrix: Extract row 5, but preserve matrix structure (returns 1x5 matrix)
#' h5_read(file, "matrix_data", start = 5, count = 1)
#' 
#' # --- Partial Reading: Multi-Value 'start' ---
#' # Matrix: Extract exactly row 5, column 2 (drops both dims, returns a scalar)
#' h5_read(file, "matrix_data", start = c(5, 2))
#' 
#' # 3D Array: Target matrix 2, row 1. (drops matrix and row dims, returns 1D vector of cols)
#' h5_read(file, "array_data", start = c(2, 1))
#' 
#' unlink(file)
h5_read <- function(file, name = "/", attr = NULL, as = "auto", start = NULL, count = NULL) {
  
  file   <- validate_strings(file, name, attr, must_exist = TRUE)
  obj_as <- validate_as(as)
  
  validate_start_count(file, name, attr, start, count)
  
  # Validate choices
  choices <- c("auto", "integer", "double", "logical", "bit64", "null")
  if (!missing(choices))
    for (i in seq_along(obj_as))
      obj_as[i] <- tryCatch(
        expr  = match.arg(tolower(obj_as[[i]]), choices),
        error = function (e) { 
          stop(
            call. = FALSE,
            "Invalid `as` argument: '", obj_as[[i]], "'\n", 
            "Valid options are: '", paste(collapse = "', '", choices), "'.") })
  
  # Prepare the 'as' map for attributes
  # Example: obj_as  = c("@ready" = "logical", ".uint" = "integer", "@." = "null")
  #          attr_as = c("ready" = "logical",  ".uint" = "integer", "."  = "null")
  attr_as <- obj_as
  if (!is.null(names(attr_as))) {
    attr_as <- attr_as[grepl("^[.@]", names(attr_as))] 
    if (length(attr_as) > 0) {
      attr_as <- attr_as[rev(order(names(attr_as)))]     
      names(attr_as) <- sub("^@", "", names(attr_as))
      attr_as <- attr_as[!duplicated(names(attr_as))]
    }
    if (is.null(attr_as) || length(attr_as) == 0) attr_as <- "auto"
  }
  
  # --- Perform Read Operation ---
  read_data(file, name, attr, obj_as, attr_as, start, count)
}


read_data <- function (file, name, attr = NULL, obj_as, attr_as, start = NULL, count = NULL) {
  
  is_auto_count <- !is.null(start) && is.null(count)
  c_count       <- if (is_auto_count) 1 else count
  
  # Case 1: Read Specific Attribute directly
  if (!is.null(attr)) {
    return(.Call("C_h5_read_attribute", file, name, attr, attr_as, PACKAGE = "h5lite"))
  }
  
  # Case 2: Read Group (Recursive)
  if (h5_is_group(file, name)) {
    children <- sort(h5_ls(file, name, recursive = FALSE, full.names = TRUE))
    res      <- lapply(children, read_data, file = file, obj_as = obj_as, attr_as = attr_as, start = start, count = count)
    names(res) <- if (length(res) == 0) NULL else basename(children)
  }
  
  # Case 3: Read Dataset
  else {
    res <- .Call("C_h5_read_dataset", file, name, obj_as, basename(name), start, c_count, PACKAGE = "h5lite")
    
    # Surgically drop fully specified dimensions
    if (!is.null(start) && !inherits(res, "data.frame")) {
      dims <- dim(res)
      
      if (!is.null(dims)) {
        N <- length(dims)
        n_start <- length(start)
        
        if (N >= 3) {
          full_map <- c(seq(N, 3L, by = -1L), 1L, 2L)
        } else {
          full_map <- seq_len(N)
        }
        dim_map <- full_map[seq_len(n_start)]
        
        dims_to_drop <- integer(0)
        
        # 1. Any targeted dimension BEFORE the last one is a point index and must be dropped
        if (n_start > 1) {
          dims_to_drop <- dim_map[1:(n_start - 1)]
        }
        # 2. The final targeted dimension is also a point index if count was automatically inferred
        if (is_auto_count) {
          dims_to_drop <- c(dims_to_drop, dim_map[n_start])
        }
        
        if (length(dims_to_drop) > 0) {
          new_dims <- dims[-dims_to_drop]
          dnames <- dimnames(res)
          
          if (!is.null(dnames)) {
            new_dnames <- dnames[-dims_to_drop]
            if (all(sapply(new_dnames, is.null))) new_dnames <- NULL
          } else {
            new_dnames <- NULL
          }
          
          # Apply the new shape
          if (length(new_dims) <= 1) {
            dim(res) <- NULL # Simplify to vector or scalar
            
            if (length(new_dims) == 1) {
              # Retain names for 1D vectors, pulling from the surviving dimnames
              if (!is.null(new_dnames)) names(res) <- new_dnames[[1]]
            } else if (length(new_dims) == 0) {
              # Explicitly unnamed for fully simplified scalar (count = NULL)
              names(res) <- NULL
            }
          } else {
            dim(res) <- new_dims
            if (!is.null(new_dnames)) dimnames(res) <- new_dnames
          }
        }
      } else if (is_auto_count && length(res) == 1) {
        # Catch 1D vectors that were natively read as a point index (count = NULL)
        names(res) <- NULL
      }
    }
  }
  
  # --- Attach Attributes ---
  obj_attr_names <- h5_attr_names(file, name)
  obj_attr_names <- setdiff(obj_attr_names, c("DIMENSION_LIST", "REFERENCE_LIST"))
  for (attr in obj_attr_names)
    if (!is.na(h5_class(file, name, attr)))
      base::attr(res, attr) <- .Call("C_h5_read_attribute", file, name, attr, attr_as, PACKAGE = "h5lite")
  
  return(res)
}

.onLoad <- function(libname, pkgname) {
    # Register plugins and spin up Blosc thread pools once per session
    .Call("C_register_hdf5_filters", PACKAGE = "h5lite") # nocov
}

.onUnload <- function(libpath) {
    # Cleanly tear down threads and free memory to prevent Valgrind warnings
    .Call("C_destroy_hdf5_filters", PACKAGE = "h5lite") # nocov
}

assert_scalar_logical <- function(...) {
  dots <- list(...)
  for (i in seq_along(dots)) {
    x <- dots[[i]]
    if (!(is.logical(x) && length(x) == 1 && !is.na(x)))
      stop("Argument `", match.call()[[i + 1]], "` must be a scalar logical.", call. = FALSE)
  }
}

assert_scalar_integer <- function(..., .null_ok = FALSE) {
  dots <- list(...)
  for (i in seq_along(dots)) {
    x <- dots[[i]]
    if (.null_ok && is.null(x)) next
    if (!(is.numeric(x) && length(x) == 1 && !is.na(x) && isTRUE(x %% 1 == 0)))
      stop("Argument `", match.call()[[i + 1]], "` must be a scalar integer.", call. = FALSE)
  }
}

assert_scalar_character <- function(...) {
  dots <- list(...)
  for (i in seq_along(dots)) {
    x <- dots[[i]]
    if (!(is.character(x) && length(x) == 1 && !is.na(x) && nzchar(trimws(x))))
      stop("Argument `", match.call()[[i + 1]], "` must be a scalar character.", call. = FALSE)
  }
}



#' Validates string inputs
#' @noRd
#' @keywords internal
#' @param file Path to the HDF5 file.
#' @param name The full path of the object (group or dataset).
#' @param attr The name of an attribute to check. If provided, the length of the attribute is returned.
#' @param must_exist Logical. If `TRUE`, the function will stop if the object does not exist.
#' @return The expanded path to the file.
validate_strings <- function (file, name = "/", attr = NULL, must_exist = FALSE) {

  assert_scalar_character(file, name)
  if (!is.null(attr)) assert_scalar_character(attr)

  file <- path.expand(file)

  if (must_exist) {

    # Check File Existence
    if (!file.exists(file)) stop("File '", file, "' does not exist.", call. = FALSE)
    
    # Check Object/Attribute Existence
    if (!.Call("C_h5_exists", file, name, attr, PACKAGE = "h5lite")) {
      if (is.null(attr)) { stop("Object '", name, "' does not exist in file '", file, "'.", call. = FALSE) } 
      else               { stop("Attribute '", attr, "' does not exist on object '", name, "'.", call. = FALSE) }
    }
  }

  return (file)
}


#' Sanity check integer arguments for start and count
#' @noRd
#' @keywords internal
validate_start_count <- function (file, name, attr, start, count) {
  
  if (is.null(start) && is.null(count))
    return (invisible())
  
  if (is.null(start) && !is.null(count))
    stop('`start` must be provided if `count` is specified.', call. = FALSE)
  
  if (!is.null(attr))
    stop('`start` and `count` cannot be used on attributes.', call. = FALSE)
  
  if (!is.numeric(start)) stop ('`start` must be numeric', call. = FALSE)
  
  start <- as.numeric(start)
  
  if (length(start) <  1) stop ('`start` cannot be an empty vector', call. = FALSE)
  if (!isTRUE(all(start > 0))) stop ('`start` must be positive', call. = FALSE)
  
  # Allow count to be NULL in the parent frame, but enforce it as 1 for bounds checking here
  check_count <- count
  if (is.null(count)) {
    check_count <- 1
  } else {
    if (!is.numeric(count)) stop ('`count` must be numeric', call. = FALSE)
    check_count <- as.numeric(count)
    if (length(check_count) != 1) stop ('`count` must be a single numeric value', call. = FALSE)
    if (!isTRUE(all(check_count > 0))) stop ('`count` must be positive', call. = FALSE)
  }
  
  shape <- h5_dim(file, name, attr)
  if (length(shape) == 0) shape <- 1L # scalar
  
  N <- length(shape)
  n <- length(start)
  
  if (n > N) stop('`start` has more dimensions than the dataset', call. = FALSE)
  
  # Determine alignment: which dimensions in `shape` does `start` apply to?
  if (N >= 3) { full_map <- c(seq(N, 3L, by = -1L), 1L, 2L) }
  else        { full_map <- seq_len(N) }
  
  # Slice the map to match the number of values provided in `start`
  dim_map <- full_map[seq_len(n)]
  
  # Extract the specific dimension sizes that 'start' is targeting
  target_shape <- shape[dim_map]
  
  if (!isTRUE(all(target_shape >= start)))              stop('`start` is out of bounds', call. = FALSE)
  if (start[[n]] + check_count - 1 > target_shape[[n]]) stop('`count` is out of bounds', call. = FALSE)
  
  assign('start', start, pos = parent.frame())
  
  # Only overwrite count in the parent if it was actively provided
  if (!is.null(count)) {
    assign('count', check_count, pos = parent.frame())
  }
  
  return (invisible())
}


#' Sanity check the 'as' argument
#' Ensures a character vector and that multiple values are named.
#' @noRd
#' @keywords internal
validate_as <- function (as) {
  
  if (is.null(as))       return ("auto")
  if (!is.character(as)) stop('`as` must be a character vector.')
  
  if (length(as) > 1 && is.null(names(as)))
    stop("When `as` has multiple values, they must be named.")
  
  if (!is.null(names(as)) && (any(is.na(names(as))) || any(names(as) == "")))
    stop("The `as` argument's names cannot be NA or an empty string.")

  return (as)
}

#' h5lite: A Simple and Lightweight HDF5 Interface
#'
#' @description
#' The `h5lite` package provides a simple, lightweight, and user-friendly 
#' interface for reading and writing HDF5 files. It is designed for R users 
#' who want to save and load common R objects (vectors, matrices, arrays, 
#' factors, and data.frames) to an HDF5 file without needing to understand 
#' the low-level details of the HDF5 C API.
#'
#' @section Key Features:
#' \itemize{
#'   \item **Simple API:** Use familiar functions like [h5_read()] and [h5_write()].
#'   \item **Automatic Handling:** Dimensions, data types, and group creation are handled automatically.
#'   \item **Safe by Default:** Auto-selects a safe R data type for numeric data to prevent overflow.
#'   \item **Easy Installation:** The required HDF5 library is bundled with the package.
#' }
#'
#' @seealso
#' Useful links:
#' * <https://cmmr.github.io/h5lite/>
#' * Report bugs at <https://github.com/cmmr/h5lite/issues>
#'
#' Key functions: [h5_read()], [h5_write()], [h5_ls()], [h5_str()]
#'
#' @keywords internal
#' @aliases h5lite-package
"_PACKAGE"

## usethis namespace: start
#' @useDynLib h5lite, .registration = TRUE
## usethis namespace: end
NULL

#' Create an HDF5 Group
#' 
#' Explicitly creates a new group (or nested groups) in an HDF5 file.
#' This is useful for creating an empty group structure.
#'
#' @param file The path to the HDF5 file.
#' @param name The full path of the group to create (e.g., "/g1/g2").
#' @return Invisibly returns `NULL`. This function is called for its side effects.
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' h5_create_file(file)
#' 
#' # Create a nested group structure
#' h5_create_group(file, "/data/experiment/run1")
#' h5_ls(file)
#' 
#' unlink(file)
h5_create_group <- function (file, name) {
  file <- validate_strings(file, name)
  # Call the C function, which handles creating parent groups automatically.
  .Call("C_h5_create_group", file, name, PACKAGE = "h5lite")
  invisible(NULL)
}


#' Create an HDF5 File
#'
#' Explicitly creates a new, empty HDF5 file.
#'
#' @details
#' This function is a simple wrapper around `h5_create_group(file, "/")`.
#' Its main purpose is to allow for explicit file creation in code.
#'
#' Note that calling this function is almost always **unnecessary**, as all
#' `h5lite` writing functions (like [h5_write()] or
#' [h5_create_group()]) will automatically create
#' the file if it does not exist.
#'
#' It is provided as a convenience for users who prefer to explicitly create
#' a file before writing data to it.
#'
#' @param file Path to the HDF5 file to be created.
#'
#' @section File Handling:
#' - If `file` does not exist, it will be created as a new, empty HDF5 file.
#' - If `file` already exists and is a valid HDF5 file, this function does
#'   nothing and returns successfully.
#' - If `file` exists but is **not** a valid HDF5 file (e.g., a text file),
#'   an error will be thrown and the file will not be modified.
#'
#' @return Invisibly returns `NULL`. This function is called for its side
#'   effects.
#' @seealso [h5_create_group()], [h5_write()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' # Explicitly create the file
#' h5_create_file(file)
#' 
#' if (file.exists(file)) {
#'   message("File created successfully.")
#' }
#' 
#' unlink(file)
h5_create_file <- function (file) {
  # Creating a file is equivalent to creating the root group '/'.
  h5_create_group(file = file, name = "/")
}


#' Move or Rename an HDF5 Object
#'
#' Moves or renames an object (dataset, group, etc.) within an HDF5 file.
#'
#' @details
#' This function provides an efficient, low-level wrapper for the HDF5
#' library's `H5Lmove` function. It is a metadata-only operation, meaning the
#' data itself is not read or rewritten. This makes it extremely fast, even
#' for very large datasets.
#'
#' You can use this function to either rename an object within the same group
#' (e.g., `"data/old"` to `"data/new"`) or to move an object to a
#' different group (e.g., `"data/old"` to `"archive/old"`). The destination
#' parent group will be automatically created if it does not exist.
#'
#' @param file The path to the HDF5 file.
#' @param from The current (source) path of the object (e.g., `"/group/data"`).
#' @param to The new (destination) path for the object (e.g., `"/group/data_new"`).
#'
#' @return This function is called for its side-effect and returns `NULL`
#'   invisibly.
#'
#' @seealso [h5_create_group()], [h5_delete()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' h5_write(1:10, file, "group/dataset")
#'
#' # Review the file structure
#' h5_str(file)
#' 
#' # Rename within the same group
#' h5_move(file, "group/dataset", "group/renamed")
#'
#' # Review the file structure
#' h5_str(file)
#' 
#' # Move to a new group (creates parent automatically)
#' h5_move(file, "group/renamed", "archive/dataset")
#'
#' # Review the file structure
#' h5_str(file)
#' 
#' unlink(file)
h5_move <- function (file, from, to) {

  file <- validate_strings(file, from, must_exist = TRUE)
  assert_scalar_character(to)
  
  # Call the C function that wraps H5Lmove.
  .Call("C_h5_move", file, from, to, PACKAGE = "h5lite")
  invisible(NULL)
}


#' Delete an HDF5 Object or Attribute
#'
#' Deletes an object (dataset or group) or an attribute from an HDF5 file.
#' If the object or attribute does not exist, a warning is issued and the function returns
#' successfully (no error is raised).
#'
#' @param file The path to the HDF5 file.
#' @param name The full path of the object to delete (e.g., `"/data/dset"` or `"/groups/g1"`).
#' @param attr The name of the attribute to delete.
#'   * If `NULL` (the default), the object specified by `name` is deleted.
#'   * If a string is provided, the attribute named `attr` is removed from the object `name`.
#' @param warn Emit a warning if the name/attr does not exist. Default: `TRUE`
#'
#' @return Invisibly returns `NULL`. This function is called for its side effects.
#' @seealso [h5_create_group()], [h5_move()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' h5_create_file(file)
#'
#' # Create some data and attributes
#' h5_write(matrix(1:10, 2, 5), file, "matrix")
#' h5_write("A note", file, "matrix", attr = "note")
#'
#' # Review the file structure
#' h5_str(file)
#' 
#' # Delete the attribute
#' h5_delete(file, "matrix", attr = "note")
#'
#' # Review the file structure
#' h5_str(file)
#'
#' # Delete the dataset
#' h5_delete(file, "matrix")
#'
#' # Review the file structure
#' h5_str(file)
#'
#' # Cleaning up
#' unlink(file)
h5_delete <- function(file, name, attr = NULL, warn = TRUE) {

  file <- validate_strings(file, name, attr)
  assert_scalar_logical(warn)
  
  # Error if the file doesn't exist.
  if (!file.exists(file)) stop("File '", file, "' does not exist.")

  # Warn but do not error if the target doesn't exist.
  if (h5_exists(file, name, attr)) {
    if (is.null(attr)) { .Call("C_h5_delete",      file, name,       PACKAGE = "h5lite") }
    else               { .Call("C_h5_delete_attr", file, name, attr, PACKAGE = "h5lite") }
  }
  else if (isTRUE(warn)) {
    if (is.null(attr)) { warning("Object '", name, "' not found in file '", file, "'. Nothing to delete.") }
    else               { warning("Attribute '", attr, "' not found on object '", name, "'. Nothing to delete.")}
  }

  invisible(NULL)
}

local({
  #test_that("File creation and existence", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  expect_false(h5_exists(file))
  h5_create_file(file)
  expect_true(h5_exists(file))
  expect_true(h5_exists(file, "/"))
})

local({
  #test_that("Group creation", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  expect_null(h5_create_group(file, "foo/bar"))
  expect_null(h5_create_group(file, "foo/bar"))
  expect_true(h5_is_group(file, "foo"))
  expect_true(h5_is_group(file, "foo/bar"))
  
  
  h5_write(1, file, "a")
  expect_error(h5_create_group(file, "a"))
})

local({
  #test_that("Move and Delete", {
  
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  expect_error(h5_delete(file, "a"))
  
  h5_write(1, file, "a")
  
  expect_error(h5_delete(file, "b", warn = "invalid"))
  expect_error(h5_delete(file, name = 123))
  suppressWarnings({
    expect_warning(h5_delete(file, "b",      warn = TRUE))
    expect_warning(h5_delete(file, "a", "b", warn = TRUE))
  })
  
  # Move
  h5_move(file, "a", "b")
  expect_false(h5_exists(file, "a"))
  expect_true(h5_exists(file, "b"))
  
  # Delete
  h5_delete(file, "b")
  expect_false(h5_exists(file, "b"))
})

local({
  #test_that("ls, str, and names", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  h5_write(1, file, "g1/d1")
  h5_write(2, file, "g1/d2")
  
  lst <- list(
    a = 1:5,
    b = list(),
    c = matrix(1:4, 2, 2),
    d = data.frame(x=1:4, y=LETTERS[1:4]),
    e = logical(0),
    f = I('a_scalar') )
  attr(lst,   'attr1') <- lst$d
  attr(lst$a, 'attr2') <- lst$e
  attr(lst$a, 'attr3') <- lst$f
  h5_write(lst, file, "lst")
  
  expect_equal(sort(h5_names(file, "g1")), c("d1", "d2"))
  expect_equal(h5_attr_names(file, "lst"), c("attr1"))
  expect_equal(length(h5_ls(file, recursive = TRUE)), 10)
  expect_equal(length(h5_ls(file, recursive = TRUE, full.names = TRUE)), 10)
  expect_equal(length(h5_ls(file, recursive = FALSE)), 2)
  expect_equal(length(h5_ls(file, recursive = FALSE, full.names = TRUE)), 2)
  expect_equal(length(h5_ls(file, "lst/b")), 0)
  expect_equal(length(h5_ls(file, "lst", full.names = TRUE)), 6)
  expect_stdout(h5_str(file))
  expect_stdout(h5_str(file, "lst"))
  expect_stdout(h5_str(file, "lst", attrs = FALSE))
  
  expect_error(h5_ls(tempfile())) # file doesn't exist
})

#' Get R Class of an HDF5 Object or Attribute
#'
#' Inspects an HDF5 object (or an attribute attached to it) and returns the R class
#' that `h5_read()` would produce.
#'
#' @details
#' This function determines the resulting R class by inspecting the storage metadata.
#'
#' \itemize{
#'   \item **Group** \eqn{\rightarrow} `"list"`
#'   \item **Integer** \eqn{\rightarrow} `"numeric"`
#'   \item **Floating Point** \eqn{\rightarrow} `"numeric"`
#'   \item **String** \eqn{\rightarrow} `"character"`
#'   \item **Complex** \eqn{\rightarrow} `"complex"`
#'   \item **Enum** \eqn{\rightarrow} `"factor"`
#'   \item **Opaque** \eqn{\rightarrow} `"raw"`
#'   \item **Compound** \eqn{\rightarrow} `"data.frame"`
#'   \item **Null** \eqn{\rightarrow} `"NULL"`
#' }
#'
#' @param file The path to the HDF5 file.
#' @param name The full path of the object (group or dataset) to check.
#' @param attr The name of an attribute to check. If `NULL` (default), the function
#'   checks the class of the object itself.
#' @return A character string representing the R class (e.g., `"integer"`, `"numeric"`, 
#'   `"complex"`, `"character"`, `"factor"`, `"raw"`, `"list"`, `"NULL"`).
#'   Returns `NA_character_` for HDF5 types that `h5lite` cannot read.
#'
#' @seealso [h5_typeof()], [h5_read()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' h5_write(1:10, file, "my_ints", as = "int32")
#' h5_class(file, "my_ints") # "numeric"
#' 
#' h5_write(mtcars, file, "mtcars")
#' h5_class(file, "mtcars") # "data.frame"
#' 
#' h5_write(c("a", "b", "c"), file, "strings")
#' h5_class(file, "strings") # "character"
#' 
#' h5_write(c(1, 2, 3), file, "my_floats", as = "float64")
#' h5_class(file, "my_floats") # "numeric"
#' 
#' unlink(file)
h5_class <- function (file, name, attr = NULL) {

  if (h5_is_group(file, name, attr))
    return("list")

  hdf5_type <- h5_typeof(file, name, attr)

  if (!is.character(hdf5_type)) return(NA_character_) # nocov
  if (startsWith(hdf5_type, "int"))      return("numeric")
  if (startsWith(hdf5_type, "uint"))     return("numeric")
  if (startsWith(hdf5_type, "float"))    return("numeric")
  if (startsWith(hdf5_type, "compound")) return("data.frame")
  if (startsWith(hdf5_type, "utf8"))     return("character")
  if (startsWith(hdf5_type, "ascii"))    return("character")
  if (startsWith(hdf5_type, "string"))   return("character") # nocov
  if (identical(hdf5_type, "enum"))      return("factor")
  if (identical(hdf5_type, "complex"))   return("complex")
  if (identical(hdf5_type, "opaque"))    return("raw")
  if (identical(hdf5_type, "null"))      return("NULL")

  NA_character_
}


#' Get HDF5 Storage Type of an Object or Attribute
#'
#' Returns the low-level HDF5 storage type of a dataset or an attribute
#' (e.g., "int8", "float64", "utf8", "ascii\[10\]"). This allows inspecting the 
#' file storage type before reading the data into R.
#'
#' @param file The path to the HDF5 file.
#' @param name Name of the dataset or object.
#' @param attr The name of an attribute to check. If `NULL` (default), the function
#'   returns the type of the object itself.
#' @return A character string representing the HDF5 storage type 
#'   (e.g., "float32", "uint32", "ascii\[10\]", "compound\[2\]").
#'
#' @seealso [h5_class()], [h5_exists()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' h5_write(1L, file, "int32_val", as = "int32")
#' h5_typeof(file, "int32_val") # "int32"
#' 
#' h5_write(mtcars, file, "mtcars")
#' h5_typeof(file, "mtcars") # "compound[11]"
#' 
#' h5_write(c("a", "b", "c"), file, "strings")
#' h5_typeof(file, "strings") # "utf8[1]"
#' 
#' unlink(file)
h5_typeof <- function (file, name, attr = NULL) {

  file <- validate_strings(file, name, attr, must_exist = TRUE)

  if (is.null(attr)) { .Call("C_h5_typeof",      file, name,       PACKAGE = "h5lite") } 
  else               { .Call("C_h5_typeof_attr", file, name, attr, PACKAGE = "h5lite") }
}


#' Get Dimensions of an HDF5 Object or Attribute
#'
#' Returns the dimensions of a dataset or an attribute as an integer vector.
#' These dimensions match the R-style (column-major) interpretation.
#'
#' @param file The path to the HDF5 file.
#' @param name Name of the dataset or object.
#' @param attr The name of an attribute to check. If `NULL` (default), the function
#'   returns the dimensions of the object itself.
#' @return An numeric vector of dimensions, or `numeric(0)` for scalars.
#'
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' h5_write(matrix(1:10, 2, 5), file, "matrix")
#' h5_dim(file, "matrix") # 2 5
#' 
#' h5_write(mtcars, file, "mtcars")
#' h5_dim(file, "mtcars") # 32 11
#' 
#' h5_write(I(TRUE), file, "my_bool")
#' h5_dim(file, "my_bool") # numeric(0)
#' 
#' h5_write(1:10, file, "my_ints")
#' h5_dim(file, "my_ints") # 10
#' 
#' unlink(file)
h5_dim <- function (file, name, attr = NULL) {

  file <- validate_strings(file, name, attr, must_exist = TRUE)

  if (is.null(attr)) { .Call("C_h5_dim",      file, name,       PACKAGE = "h5lite") } 
  else               { .Call("C_h5_dim_attr", file, name, attr, PACKAGE = "h5lite") }
}


#' Check if an HDF5 File, Object, or Attribute Exists
#'
#' Safely checks if a file, an object within a file, or an attribute on an object exists.
#'
#' @details
#' This function provides a robust, error-free way to test for existence.
#'
#' \itemize{
#'   \item **Testing for a File:** If `name` is `/` and `attr` is `NULL`,
#'     the function checks if `file` is a valid, readable HDF5 file.
#'
#'   \item **Testing for an Object:** If `name` is a path (e.g., `/data/matrix`)
#'     and `attr` is `NULL`, the function checks if the specific object exists.
#'
#'   \item **Testing for an Attribute:** If `attr` is provided, the function checks
#'     if that attribute exists on the specified object `name`.
#' }
#'
#' @param file Path to the file.
#' @param name The full path of the object to check (e.g., `"/data/matrix"`).
#'   Defaults to `"/"` to test file validity.
#' @param attr The name of an attribute to check. If provided, the function tests
#'   for the existence of this attribute on `name`.
#' @param assert Logical. If `TRUE` and the target does not exist, the function
#'   will stop with an informative error message instead of returning `FALSE`.
#'   Defaults to `FALSE`.
#' @return A logical value: `TRUE` if the target exists and is valid, `FALSE` otherwise.
#' @seealso [h5_is_group()], [h5_is_dataset()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' h5_exists(file) # FALSE
#' 
#' h5_create_file(file)
#' h5_exists(file) # TRUE
#' 
#' h5_exists(file, "missing_object") # FALSE
#' 
#' h5_write(1:10, file, "my_ints")
#' h5_exists(file, "my_ints") # TRUE
#' 
#' h5_exists(file, "my_ints", "missing_attr") # FALSE
#' 
#' h5_write(1:10, file, "my_ints", attr = "my_attr")
#' h5_exists(file, "my_ints", "my_attr") # TRUE
#' 
#' unlink(file)
h5_exists <- function (file, name = "/", attr = NULL, assert = FALSE) {
  
  if (assert) {
    file <- validate_strings(file, name, attr, must_exist = TRUE)
  }
  else {
    file <- tryCatch(
      validate_strings(file, name, attr, must_exist = TRUE),
      error = function(e) { return(NULL) }
    )
  }
  
  return(isTRUE(nzchar(file)))
}

#' Check if an HDF5 Object is a Group
#'
#' Checks if the object at a given path is a group.
#'
#' @param file The path to the HDF5 file.
#' @param name The full path of the object to check.
#' @param attr The name of an attribute. This parameter is included for consistency with other functions.
#'   Since attributes cannot be groups, providing this will always return `FALSE`. (Default: `NULL`)
#' @return A logical value: `TRUE` if the object exists and is a group,
#'   `FALSE` otherwise (if it is a dataset, or does not exist).
#' @seealso [h5_is_dataset()], [h5_exists()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' h5_create_group(file, "grp")
#' h5_is_group(file, "grp") # TRUE
#' 
#' h5_write(1:10, file, "my_ints")
#' h5_is_group(file, "my_ints") # FALSE
#' 
#' unlink(file)
h5_is_group <- function (file, name, attr = NULL) {

  file <- validate_strings(file, name, attr)
  
  if (!is.null(attr))         return(FALSE)
  if (!h5_exists(file, name)) return(FALSE)

  # Call the C function to check if the object's type is H5O_TYPE_GROUP.
  .Call("C_h5_is_group", file, name, PACKAGE = "h5lite")
}

#' Check if an HDF5 Object is a Dataset
#'
#' Checks if the object at a given path is a dataset.
#'
#' @param file The path to the HDF5 file.
#' @param name The full path of the object to check.
#' @param attr The name of an attribute. If provided, the function returns `TRUE` if the attribute exists, 
#'   as all attributes are considered datasets in HDF5 context. (Default: `NULL`)
#' @return A logical value: `TRUE` if the object exists and is a dataset,
#'   `FALSE` otherwise (if it is a group, or does not exist).
#' @seealso [h5_is_group()], [h5_exists()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' h5_write(1, file, "dset")
#' h5_is_dataset(file, "dset") # TRUE
#' 
#' h5_create_group(file, "grp")
#' h5_is_dataset(file, "grp") # FALSE
#' 
#' h5_write(1, file, "grp", attr = "my_attr")
#' h5_is_dataset(file, "grp", "my_attr") # TRUE
#' 
#' unlink(file)
h5_is_dataset <- function (file, name, attr = NULL) {

  file <- validate_strings(file, name, attr)
  
  if (!h5_exists(file, name, attr)) return(FALSE)
  if (!is.null(attr))               return(TRUE)
  
  # Call the C function to check if the object's type is H5O_TYPE_DATASET.
  .Call("C_h5_is_dataset", file, name, PACKAGE = "h5lite")
}

#' Get Names of an HDF5 Object
#'
#' Returns the names of the object.
#' * For **Groups**, it returns the names of the objects contained in the group (similar to `ls()`).
#' * For **Compound Datasets** (data.frames), it returns the column names.
#' * For other **Datasets**, it looks for a dimension scale and returns it if found.
#'
#' @param file The path to the HDF5 file.
#' @param name The full path of the object.
#' @param attr The name of an attribute. If provided, returns the names associated with the attribute
#'   (e.g., field names if the attribute is a compound type). (Default: `NULL`)
#' @return A character vector of names, or `NULL` if the object has no names.
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' h5_write(data.frame(x=1, y=2), file, "df")
#' h5_names(file, "df") # "x" "y"
#' 
#' x <- 1:5
#' names(x) <- letters[1:5]
#' h5_write(x, file, "x")
#' h5_names(file, "x") # "a" "b" "c" "d" "e"
#' 
#' h5_write(mtcars[,c("mpg", "hp")], file, "dset")
#' h5_names(file, "dset") # "mpg" "hp"
#' 
#' unlink(file)
h5_names <- function (file, name = "/", attr = NULL) {

  file <- validate_strings(file, name, attr, must_exist = TRUE)

  if (h5_is_dataset(file, name, attr)) {
    res <- .Call("C_h5_names", file, name, attr, PACKAGE = "h5lite")
  }
  else {
    res <- h5_ls(file, name, recursive = FALSE, full.names = FALSE)
  }
  
  if (is.null(res)) res <- character(0)
  return (res)
}

#' List HDF5 Attributes
#' 
#' Lists the names of attributes attached to a specific HDF5 object.
#'
#' @param file The path to the HDF5 file.
#' @param name The path to the object (dataset or group) to query. 
#'   Use `/` for the file's root attributes.
#' @return A character vector of attribute names.
#' 
#' @seealso [h5_ls()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' h5_write(1:10,          file, "data")
#' h5_write(I("meters"),   file, "data", attr = "unit")
#' h5_write(I(Sys.time()), file, "data", attr = "timestamp")
#' 
#' h5_attr_names(file, "data") # "unit" "timestamp"
#' 
#' unlink(file)
h5_attr_names <- function (file, name = "/") {
  file <- validate_strings(file, name, must_exist = TRUE)
  
  # Call the C function that iterates over attributes and returns their names.
  .Call("C_h5_attr_names", file, name, PACKAGE = "h5lite")
}


#' Get the Total Length of an HDF5 Object or Attribute
#'
#' Behaves like `length()` for R objects.
#' * For **Compound Datasets** (data.frames), this is the number of columns.
#' * For **Datasets** and **Attributes**, this is the product of all dimensions (total number of elements).
#' * For **Groups**, this is the number of objects directly contained in the group.
#' * Scalar datasets or attributes return 1.
#'
#' @param file The path to the HDF5 file.
#' @param name The full path of the object (group or dataset).
#' @param attr The name of an attribute to check. If provided, the length of the attribute is returned.
#' @return An numeric scalar representing the total length (number of elements).
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' h5_write(1:100, file, "my_vec")
#' h5_length(file, "my_vec") # 100
#' 
#' h5_write(mtcars, file, "my_df")
#' h5_length(file, "my_df") # 11 (ncol(mtcars))
#' 
#' h5_write(as.matrix(mtcars), file, "my_mtx")
#' h5_length(file, "my_mtx") # 352 (prod(dim(mtcars)))
#' 
#' h5_length(file, "/") # 3
#' 
#' unlink(file)
h5_length <- function (file, name, attr = NULL) {

  if (h5_is_dataset(file, name, attr)) {
    dims <- h5_dim(file, name, attr)
    res <- if (identical(dims, numeric(0)))                     { 1          }
           else if (h5_class(file, name, attr) == "data.frame") { dims[2]    }
           else                                                 { prod(dims) }
    return(as.double(res))
  }
  else if (h5_is_group(file, name, attr)) {
    return(length(h5_ls(file, name, recursive = FALSE)))
  }

  return(NA_integer_)
}

local({
  #test_that("h5 wrapper object works", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  h5 <- h5_open(file)
  
  # Write via method
  h5$write(1:10, "dset")
  expect_true(h5$exists("dset"))
  
  # Read via method
  expect_equal(h5$read("dset"), 1:10)
  
  
  # Misc functions
  expect_stdout(str(h5))
  expect_stdout(print(h5))
  expect_equal(as.character(h5), file)
  expect_equal(h5$class("dset"), "numeric")
  expect_equal(h5$dim("dset"), 10)
  expect_inherits(h5$inspect("dset"), "inspect")
  expect_equal(h5$length("dset"), 10)
  expect_equal(h5$typeof("dset"), "uint8")
  expect_true(h5$is_dataset("dset"))
  expect_false(h5$is_group("dset"))
  expect_equal(h5$names("dset"), character(0))
  expect_equal(h5$attr_names("dset"), character(0))
  expect_null(h5$move("dset", "dset2"))
  
  
  # Navigation
  h5$create_group("sub")
  expect_equal(h5$class("sub"), "list")
  h5$cd("sub")
  expect_equal(h5$pwd(), "/sub")
  
  h5$write("hidden", "deep") # Should be at /sub/deep
  expect_true(h5_exists(file, "/sub/deep"))
  
  # Relative read/delete
  expect_equal(h5$read("deep"), "hidden")
  expect_null(h5$delete("../dset2"))
  
  
  # Pass-by-reference check
  h5_alias <- h5
  h5_alias$cd("/")
  expect_equal(h5$pwd(), "/")
  
  
  # Closing
  h5$close()
  
  expect_stdout(print(h5))
  expect_error(h5$ls(), "handle has been closed")
})

local({
  #test_that("Integer scalars and vectors work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  # 1. Simple Integer Vector
  vec <- 1:10
  h5_write(vec, file, "vec")
  expect_equal(h5_read(file, "vec"), vec)
  expect_equal(h5_class(file, "vec"), "numeric")
  expect_equal(h5_read(file, "vec", as = "integer"), vec)

  # 2. Integer Scalar (1D array of length 1)
  h5_write(42L, file, "scalar")
  expect_equal(h5_read(file, "scalar"), 42L)
  
  # 3. Explicit Scalar (I(x))
  h5_write(I(42L), file, "scalar_I")
  expect_equal(h5_dim(file, "scalar_I"), numeric(0)) # Rank 0
  expect_equal(h5_read(file, "scalar_I"), 42L)

  # 4. Integers with NA (Auto-promoted to float to store NA)
  vec_na <- c(1L, NA_integer_, 3L)
  h5_write(vec_na, file, "vec_na")
  expect_equal(h5_typeof(file, "vec_na"), "float64")
  read_back <- h5_read(file, "vec_na")
  expect_true(is.double(read_back))
  expect_true(is.na(read_back[2]))
  
  # 5. Named vector
  vec_named <- structure(c(1:5), names = letters[1:5])
  h5_write(vec_named, file, "vec_named")
  expect_equal(h5_read(file, "vec_named"), vec_named)
  expect_equal(h5_names(file, "vec_named"), letters[1:5])
  
  # Force read back as integer (NaN becomes NA_integer_)
  expect_equal(h5_read(file, "vec_na", as = "integer"), vec_na)
})

local({
  #test_that("Integer matrices and arrays work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  mat <- matrix(1:9, nrow = 3)
  h5_write(mat, file, "mat")
  expect_equal(h5_read(file, "mat"), mat)
  expect_equal(h5_dim(file, "mat"), c(3, 3))
  expect_equal(h5_names(file, "mat"), character(0))
  expect_equal(h5_length(file, "mat"), 9L)
  
  mat_named <- matrix(1:9, nrow = 3, dimnames = list(letters[1:3], LETTERS[1:3]))
  h5_write(mat_named, file, "mat_named")
  expect_equal(h5_read(file, "mat_named"), mat_named)
  expect_equal(h5_names(file, "mat_named"), LETTERS[1:3])
  
  mat_rownm <- matrix(1:9, nrow = 3, dimnames = list(letters[1:3], NULL))
  h5_write(mat_rownm, file, "mat_rownm")
  expect_equal(h5_read(file, "mat_rownm"), mat_rownm)
  expect_equal(h5_names(file, "mat_rownm"), character(0))
  
  mat_colnm <- matrix(1:9, nrow = 3, dimnames = list(NULL, LETTERS[1:3]))
  h5_write(mat_colnm, file, "mat_colnm")
  expect_equal(h5_read(file, "mat_colnm"), mat_colnm)
  expect_equal(h5_names(file, "mat_colnm"), LETTERS[1:3])
  
  arr3d <- array(
    data = 1:24, 
    dim = c(2,3,4), 
    dimnames = list(LETTERS[1:2], LETTERS[3:5], LETTERS[7:10]) )
  
  h5_write(arr3d, file, "arr3d")
  expect_equal(h5_read(file, "arr3d"), arr3d)
  expect_equal(h5_dim(file, "arr3d"), c(2,3,4) )
  expect_equal(h5_names(file, "arr3d"), LETTERS[3:5])
  expect_equal(h5_length(file, "arr3d"), prod(c(2,3,4)))
})

local({
  #test_that("Integer attributes work", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  h5_write(1:5, file, "data")
  h5_write(100L, file, "data", attr = "meta")
  
  expect_equal(h5_read(file, "data", attr = "meta", as = "integer"), 100L)
  
  expect_null(h5_delete(file, "data", attr = "meta"))
  expect_null(h5_delete(file, "data"))
  expect_error(h5_write(1:5, file, "data", attr = "meta"))
})

local({
  #test_that("Specific integer types (int8-int64) can be forced", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  val <- c(-100L, 100L)
  
  h5_write(val, file, "i8", as = "int8")
  expect_equal(h5_typeof(file, "i8"), "int8")
  expect_equal(h5_read(file, "i8", as = "integer"), val)

  # Can't write -100 to a uint
  expect_error(h5_write(val, file, "u8", as = "uint8"))
})

local({
  #test_that("Compression", {
  
  file_compressed   <- tempfile(fileext = ".h5")
  file_uncompressed <- tempfile(fileext = ".h5")
  on.exit(unlink(c(file_compressed, file_uncompressed)))
  
  
  vec <- 1:13
  
  h5_write(vec, file_compressed, "vec", compress = "none")
  expect_equal(h5_read(file_compressed, "vec"), vec)
  
  h5_write(vec, file_compressed, "vec", compress = "gzip-5")
  expect_equal(h5_read(file_compressed, "vec"), vec)
  
  h5_write(vec, file_compressed, "vec", compress = "szip-nn")
  expect_equal(h5_read(file_compressed, "vec"), vec)
  
  h5_write(vec, file_compressed, "vec", compress = "szip-ec")
  expect_equal(h5_read(file_compressed, "vec"), vec)
  
  
  vec <- rep(1L, 10000000)
  h5_write(vec, file_compressed,   "vec", compress = "gzip-5")
  h5_write(vec, file_uncompressed, "vec", compress = "none")
  expect_true(file.size(file_compressed) < file.size(file_uncompressed))
  
  
  # fallback to gzip for strings
  h5_write(letters, file_uncompressed, "letters", compress = "szip-nn")
  
  
  exit_if_not(at_home())
  #skip_on_cran()
  
  mtx <- matrix(vec, nrow = 100)
  h5_write(mtx, file_compressed, "mtx", compress = "gzip-5")
  expect_equal(h5_read(file_compressed, "mtx"), mtx)
  
  h5_write(mtx, file_compressed, "mtx", compress = "szip-nn")
  expect_equal(h5_read(file_compressed, "mtx"), mtx)
  
  h5_write(mtx, file_compressed, "mtx", compress = "szip-ec")
  expect_equal(h5_read(file_compressed, "mtx"), mtx)
  
  
  for (i in c(1, 2, 3, 7, 11, 19, 61, 1000)) {
    vec <- 1:i
    
    h5_write(vec, file_compressed, "vec", compress = "szip-nn")
    expect_equal(h5_read(file_compressed, "vec"), vec)
    
    h5_write(vec, file_compressed, "vec", compress = "szip-ec")
    expect_equal(h5_read(file_compressed, "vec"), vec)
  }
  
  
  # deprecated `compress` args
  vec <- 1:100
  h5_write(vec, file_compressed, "vec", compress = TRUE)
  h5_write(vec, file_compressed, "vec", compress = FALSE)
  h5_write(vec, file_compressed, "vec", compress = 5)
  
  
})

local({
  #test_that("Data frames work (Compound)", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  df <- data.frame(
    int = 1:3,
    dbl = c(1.1, 2.2, 3.3),
    chr0 = c("a", "b", NA),
    chr1 = c("a", "b", NA),
    chr2 = c("a", "b", NA),
    chr3 = c("a", "b", "c"),
    chr4 = c("a", "b", "c"),
    chr5 = c("a", "b", "c"),
    chr6 = c("a", "b", "c"),
    fac = factor(c("a", "b", "c")),
    cpx = c(1+1i, 1+2i, 3+4i),
    raw = as.raw(1:3),
    cnv = c(1:2, NA_integer_),
    lgl = c(TRUE, FALSE, TRUE),
    int_skp = 1:3,
    chr_skp = c("a", "b", "c"),
    stringsAsFactors = FALSE
  )
  
  rownames(df) <- LETTERS[1:3]
  
  col_map <- c(
    int_skp = "skip",
    chr_skp = "skip",
    chr1    = "utf",
    chr2    = "ascii",
    chr3    = "utf[]",
    chr4    = "utf[3]",
    chr5    = "ascii[]",
    chr6    = "ascii[3]" )
  
  
  # As a dataset ---------------------------------
  
  h5_write(df, file, "df", as = col_map)
  expect_equal(h5_class(file, "df"), "data.frame")
  expect_true(grepl("compound\\[\\d+\\]", h5_typeof(file, "df")))
  
  res <- h5_read(file, "df")
  expect_null(res$skp)
  expect_equal(res$int,  df$int)
  expect_equal(res$dbl,  df$dbl)
  expect_equal(res$chr0, df$chr0)
  expect_equal(res$chr1, df$chr1)
  expect_equal(res$chr2, df$chr2)
  expect_equal(res$chr3, df$chr3)
  expect_equal(res$chr4, df$chr4)
  expect_equal(res$chr5, df$chr5)
  expect_equal(res$chr6, df$chr6)
  expect_equal(res$fac,  df$fac)
  expect_equal(res$cpx,  df$cpx)
  expect_equal(res$raw,  df$raw)
  expect_equal(res$lgl,  c(1, 0, 1))       # Default int
  expect_equal(res$cnv,  c(1:2, NA_real_)) # Default float
  expect_equal(rownames(res), rownames(df))
  
  # Read with mapping
  res2 <- h5_read(file, "df", as = c("lgl" = "logical"))
  expect_equal(res2$lgl, df$lgl)
  
  
  # As an attribute (drops rownames) -------------
  
  h5_write(df, file, "df", "attr")
  expect_equal(h5_class(file, "df", "attr"), "data.frame")
  expect_true(grepl("compound\\[\\d+\\]", h5_typeof(file, "df", "attr")))
  
  res <- h5_read(file, "df", "attr")
  expect_equal(res$int, df$int)
  expect_equal(res$dbl, df$dbl)
  expect_equal(res$chr, df$chr)
  expect_equal(res$fac, df$fac)
  expect_equal(res$cpx, df$cpx)
  expect_equal(res$raw, df$raw)
  expect_equal(res$lgl, c(1, 0, 1))       # Default int
  expect_equal(res$cnv, c(1:2, NA_real_)) # Default float
  
  # Read with mapping
  res2 <- h5_read(file, "df", as = c("lgl" = "logical"))
  expect_equal(res2$lgl, df$lgl)
  
  # Zero columns = error
  no_cols <- mtcars[,integer(0),drop = FALSE]
  expect_error(h5_write(no_cols, file, "no_cols"))
})

local({
  #test_that("Data frame columns with POSIXt are converted", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  now <- Sys.time()
  df <- data.frame(id = 1, time = now)
  h5_write(df, file, "df")
  
  res <- h5_read(file, "df")
  expect_true(is.character(res$time))
  expect_equal(res$time, format(now, format = "%Y-%m-%dT%H:%M:%OSZ"))
})

local({
  #test_that("h5_typeof and h5_class", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  h5_create_file(file)
  
  h5_write(1, file, 'ds_uint8',    as = "uint8")
  h5_write(1, file, 'ds_uint16',   as = "uint16")
  h5_write(1, file, 'ds_uint32',   as = "uint32")
  h5_write(1, file, 'ds_uint64',   as = "uint64")
  h5_write(1, file, 'ds_int8',     as = "int8")
  h5_write(1, file, 'ds_int16',    as = "int16")
  h5_write(1, file, 'ds_int32',    as = "int32")
  h5_write(1, file, 'ds_int64',    as = "int64")
  h5_write(1, file, 'ds_bfloat16', as = "bfloat16")
  h5_write(1, file, 'ds_float16',  as = "float16")
  h5_write(1, file, 'ds_float32',  as = "float32")
  h5_write(1, file, 'ds_float64',  as = "float64")

  h5_write(NULL,      file, 'ds_null')
  h5_write('a',       file, 'ds_utf8',  as = 'utf8')
  h5_write('a',       file, 'ds_ascii', as = 'ascii')
  h5_write(1+1i,      file, 'ds_complex')
  h5_write(mtcars,    file, 'ds_compound')
  h5_write(as.raw(1), file, 'ds_opaque')
  h5_write(as.factor(letters), file, 'ds_enum')
  
  expect_equal(h5_typeof(file, 'ds_uint8'),    'uint8')
  expect_equal(h5_typeof(file, 'ds_uint16'),   'uint16')
  expect_equal(h5_typeof(file, 'ds_uint32'),   'uint32')
  expect_equal(h5_typeof(file, 'ds_uint64'),   'uint64')
  expect_equal(h5_typeof(file, 'ds_int8'),     'int8')
  expect_equal(h5_typeof(file, 'ds_int16'),    'int16')
  expect_equal(h5_typeof(file, 'ds_int32'),    'int32')
  expect_equal(h5_typeof(file, 'ds_int64'),    'int64')
  expect_equal(h5_typeof(file, 'ds_bfloat16'), 'bfloat16')
  expect_equal(h5_typeof(file, 'ds_float16'),  'float16')
  expect_equal(h5_typeof(file, 'ds_float32'),  'float32')
  expect_equal(h5_typeof(file, 'ds_float64'),  'float64')
  expect_equal(h5_typeof(file, 'ds_null'),     'null')
  expect_equal(h5_typeof(file, 'ds_utf8'),     'utf8')
  expect_equal(h5_typeof(file, 'ds_ascii'),    'ascii')
  expect_equal(h5_typeof(file, 'ds_complex'),  'complex')
  expect_true(grepl('compound\\[\\d+\\]', h5_typeof(file, 'ds_compound')))
  expect_equal(h5_typeof(file, 'ds_opaque'),   'opaque')
  expect_equal(h5_typeof(file, 'ds_enum'),     'enum')

  expect_equal(h5_class(file, 'ds_uint8'),    'numeric')
  expect_equal(h5_class(file, 'ds_int8'),     'numeric')
  expect_equal(h5_class(file, 'ds_float16'),  'numeric')
  expect_equal(h5_class(file, 'ds_null'),     'NULL')
  expect_equal(h5_class(file, 'ds_utf8'),     'character')
  expect_equal(h5_class(file, 'ds_ascii'),    'character')
  expect_equal(h5_class(file, 'ds_complex'),  'complex')
  expect_equal(h5_class(file, 'ds_compound'), 'data.frame')
  expect_equal(h5_class(file, 'ds_opaque'),   'raw')
  expect_equal(h5_class(file, 'ds_enum'),     'factor')


  h5_write(1,    file, '/', 'attr_uint8', as = "uint8")
  h5_write(NULL, file, '/', 'attr_null')

  expect_equal(h5_typeof(file, '/', 'attr_uint8'), 'uint8')
  expect_equal(h5_typeof(file, '/', 'attr_null'),  'null')

  expect_equal(h5_class(file, '/', 'attr_uint8'), 'numeric')
  expect_equal(h5_class(file, '/', 'attr_null'),  'NULL')


  expect_stdout(h5_str(file))
})


local({
  #test_that("h5_dim / h5_length", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))
  
  mtx_2d <- as.matrix(mtcars)
  mtx_3d <- array(1:24, dim = c(2, 3, 4))

  # Datasets of different dimensions
  h5_write(numeric(0), file, 'vec_n0')
  h5_write(numeric(1), file, 'vec_n1')
  h5_write(numeric(2), file, 'vec_n2')
  h5_write(I(1),       file, 'scalar')
  h5_write(mtcars,     file, 'mtcars')
  h5_write(mtx_2d,     file, 'mtx_2d')
  h5_write(mtx_3d,     file, 'mtx_3d')
  
  expect_equal(h5_dim(file, 'vec_n0'), 0L)
  expect_equal(h5_dim(file, 'vec_n1'), 1L)
  expect_equal(h5_dim(file, 'vec_n2'), 2L)
  expect_equal(h5_dim(file, 'scalar'), numeric(0))
  expect_equal(h5_dim(file, 'mtcars'), dim(mtcars))
  expect_equal(h5_dim(file, 'mtx_2d'), dim(mtx_2d))
  expect_equal(h5_dim(file, 'mtx_3d'), dim(mtx_3d))

  expect_equal(h5_length(file, 'vec_n0'), 0L)
  expect_equal(h5_length(file, 'vec_n1'), 1L)
  expect_equal(h5_length(file, 'vec_n2'), 2L)
  expect_equal(h5_length(file, 'scalar'), 1L)
  expect_equal(h5_length(file, 'mtcars'), length(mtcars))
  expect_equal(h5_length(file, 'mtx_2d'), length(mtx_2d))
  expect_equal(h5_length(file, 'mtx_3d'), length(mtx_3d))

  # Attributes of different dimensions
  h5_write(numeric(0), file, '/', 'vec_n0')
  h5_write(numeric(1), file, '/', 'vec_n1')
  h5_write(numeric(2), file, '/', 'vec_n2')
  h5_write(I(1),       file, '/', 'scalar')
  h5_write(mtcars,     file, '/', 'mtcars')
  h5_write(mtx_2d,     file, '/', 'mtx_2d')
  h5_write(mtx_3d,     file, '/', 'mtx_3d')
  
  expect_equal(h5_dim(file, '/', 'vec_n0'), 0L)
  expect_equal(h5_dim(file, '/', 'vec_n1'), 1L)
  expect_equal(h5_dim(file, '/', 'vec_n2'), 2L)
  expect_equal(h5_dim(file, '/', 'scalar'), numeric(0))
  expect_equal(h5_dim(file, '/', 'mtcars'), dim(mtcars))
  expect_equal(h5_dim(file, '/', 'mtx_2d'), dim(mtx_2d))
  expect_equal(h5_dim(file, '/', 'mtx_3d'), dim(mtx_3d))

  expect_equal(h5_length(file, '/', 'vec_n0'), 0L)
  expect_equal(h5_length(file, '/', 'vec_n1'), 1L)
  expect_equal(h5_length(file, '/', 'vec_n2'), 2L)
  expect_equal(h5_length(file, '/', 'scalar'), 1L)
  expect_equal(h5_length(file, '/', 'mtcars'), length(mtcars))
  expect_equal(h5_length(file, '/', 'mtx_2d'), length(mtx_2d))
  expect_equal(h5_length(file, '/', 'mtx_3d'), length(mtx_3d))

  expect_equal(h5_length(file, "/"), 7L)
  expect_equal(h5_length(file, "/", "missing_attr"), NA_integer_)
  expect_equal(h5_length(file, "missing_object"), NA_integer_)
  expect_equal(h5_length(file, "missing_object", "missing_attr"), NA_integer_)

  expect_stdout(h5_str(file))
})


local({
  #test_that("h5_exists", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))

  expect_false(h5_exists(file))
  expect_false(h5_exists(file, "/"))
  expect_false(h5_exists(file, "missing_object"))
  expect_false(h5_exists(file, "/", "missing_attr"))
  expect_false(h5_exists(file, "missing_object", "missing_attr"))

  h5_create_file(file)

  expect_true(h5_exists(file))
  expect_true(h5_exists(file, "/"))
  expect_false(h5_exists(file, "missing_object"))
  expect_false(h5_exists(file, "/", "missing_attr"))
  expect_false(h5_exists(file, "missing_object", "missing_attr"))

  h5_write(1, file, "object")
  expect_true(h5_exists(file, "object"))
  expect_false(h5_exists(file, "/", "missing_attr"))
  expect_false(h5_exists(file, "object", "missing_attr"))

  h5_write(1, file, "/", "attr")
  expect_true(h5_exists(file, "/", "attr"))
  expect_false(h5_exists(file, "object", "attr"))

  h5_write(1, file, "object", "attr")
  expect_true(h5_exists(file, "object", "attr"))

  expect_error(h5_exists(tempfile(fileext = ".h5"), assert = TRUE))
  expect_error(h5_exists(file, "missing_object",    assert = TRUE))
  expect_error(h5_exists(file, "/", "missing_attr", assert = TRUE))
})


local({
  #test_that("h5_is_group / h5_is_dataset", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))

  expect_false(h5_is_group(file, "/"))
  expect_false(h5_is_group(file, "/", "missing_attr"))
  expect_false(h5_is_group(file, "missing_object"))
  expect_false(h5_is_group(file, "missing_object", "missing_attr"))
  expect_false(h5_is_dataset(file, "/"))
  expect_false(h5_is_dataset(file, "/", "missing_attr"))
  expect_false(h5_is_dataset(file, "missing_object"))
  expect_false(h5_is_dataset(file, "missing_object", "missing_attr"))

  h5_create_file(file)

  expect_true(h5_is_group(file, "/"))
  expect_false(h5_is_group(file, "/", "missing_attr"))
  expect_false(h5_is_group(file, "missing_object"))
  expect_false(h5_is_group(file, "missing_object", "missing_attr"))
  expect_false(h5_is_dataset(file, "/"))
  expect_false(h5_is_dataset(file, "/", "missing_attr"))
  expect_false(h5_is_dataset(file, "missing_object"))
  expect_false(h5_is_dataset(file, "missing_object", "missing_attr"))

  h5_create_group(file, "group")
  h5_write(1, file, "dataset")

  expect_true(h5_is_group(file, "group"))
  expect_true(h5_is_dataset(file, "dataset"))
  expect_false(h5_is_group(file, "dataset"))
  expect_false(h5_is_dataset(file, "group"))
  expect_false(h5_is_group(file, "group", "missing_attr"))
  expect_false(h5_is_dataset(file, "dataset", "missing_attr"))

  
  h5_write(1, file, "group", "group_attr")
  h5_write(1, file, "dataset", "dataset_attr")

  expect_true(h5_is_dataset(file, "group", "group_attr"))
  expect_true(h5_is_dataset(file, "dataset", "dataset_attr"))
  expect_false(h5_is_group(file, "group", "group_attr"))
  expect_false(h5_is_group(file, "dataset", "dataset_attr"))
})


local({
  #test_that("h5_names / h5_attr_names", {
  file <- tempfile(fileext = ".h5")
  on.exit(unlink(file))

  expect_error(h5_names(file))
  expect_error(h5_names(file, "missing_object"))
  expect_error(h5_names(file, "/", "missing_attr"))
  expect_error(h5_names(file, "missing_object", "missing_attr"))

  expect_error(h5_attr_names(file))
  expect_error(h5_attr_names(file, "missing_object"))
  expect_error(h5_attr_names(file, "/", "missing_attr"))
  expect_error(h5_attr_names(file, "missing_object", "missing_attr"))

  h5_create_file(file)

  expect_equal(h5_names(file), character(0))
  expect_equal(h5_names(file, "/"), character(0))

  expect_error(h5_names(file, "missing_object"))
  expect_error(h5_names(file, "/", "missing_attr"))
  expect_error(h5_names(file, "missing_object", "missing_attr"))

  expect_equal(h5_attr_names(file), character(0))
  expect_equal(h5_attr_names(file, "/"), character(0))

  expect_error(h5_attr_names(file, "missing_object"))

  vec <- setNames(1:3, c("a", "b", "c"))
  h5_write(mtcars, file, "ds_mtcars")
  h5_write(vec,    file, "ds_vec")
  h5_write(mtcars, file, "/", "attr_mtcars")
  h5_write(vec,    file, "/", "attr_vec")
  h5_write(1:10,   file, "group/unnamed")

  expect_equal(h5_names(file), c("ds_mtcars", "ds_vec", "group"))
  expect_equal(h5_names(file, "/"), c("ds_mtcars", "ds_vec", "group"))
  expect_equal(h5_names(file, "ds_mtcars"), names(mtcars))
  expect_equal(h5_names(file, "ds_vec"), names(vec))
  expect_equal(h5_names(file, "group"), c("unnamed"))
  expect_equal(h5_names(file, "group/unnamed"), character(0))
  expect_equal(h5_names(file, "/", "attr_mtcars"), names(mtcars))
  expect_equal(h5_names(file, "/", "attr_vec"), character(0))

  expect_equal(h5_attr_names(file), c("attr_mtcars", "attr_vec"))
  expect_equal(h5_attr_names(file, "/"), c("attr_mtcars", "attr_vec"))
  expect_equal(h5_attr_names(file, "group"), character(0))
  expect_equal(h5_attr_names(file, "group/unnamed"), character(0))
})

#' Inspect HDF5 Dataset Creation Properties
#'
#' Retrieves the Dataset Creation Property List (DCPL) details including storage 
#' layout, chunk dimensions, and a detailed list of all applied filters.
#'
#' @param file The path to the HDF5 file.
#' @param name The full path of the dataset to inspect.
#' @return An object of class `inspect` (a named list) containing:
#'   \item{layout}{A string indicating storage layout (e.g., "chunked", "contiguous").}
#'   \item{chunk_dims}{A numeric vector of chunk dimensions, or `NULL` if not chunked.}
#'   \item{filters}{A list describing each filter applied.}
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' compress <- h5_compression('lz4-9', int_packing = TRUE, checksum = TRUE)
#' h5_write(matrix(5001:5100, 10, 10), file, "packed_mtx", compress = compress)
#' h5_inspect(file, "packed_mtx")
#' 
#' mtx <- matrix(rnorm(1000), 100, 10)
#' h5_write(mtx, file, "float_mtx", compress = 'blosc2-zfp-prec-3')
#' res <- h5_inspect(file, "float_mtx")
#' print(res)
#' 
#' # Print the raw cd_values for blosc2
#' dput(res$filters[[1]]$cd_values)
#' 
#' unlink(file)
h5_inspect <- function(file, name) {
  file <- validate_strings(file, name, must_exist = TRUE)
  
  if (!h5_is_dataset(file, name)) {
    stop("Target must be a dataset to inspect its DCPL.")
  }
  
  res <- .Call("C_h5_inspect", file, name, PACKAGE = "h5lite")
  
  # Append the 'inspect' class while keeping 'list' for standard subsetting
  class(res) <- c("inspect", "list")
  
  return(res)
}


# Helper to format raw bytes nicely
fmt_bytes <- function(b) {
  if (is.na(b) || b == 0) return("0 B")
  units <- c("B", "KB", "MB", "GB", "TB", "PB", "EB")
  idx   <- max(1, min(7, floor(log(b, 1024)) + 1))
  sprintf("%.2f %s", b / (1024^(idx - 1)), units[idx])
}



#' Print method for HDF5 inspect objects
#'
#' @keywords internal
#' @param x An object of class `inspect`.
#' @param ... Further arguments passed to or from other methods.
#' @export
print.inspect <- function(x, ...) {
  
  cat("<HDF5 Dataset Properties>\n")
  
  # --- Row 1: Type & Size ---
  left_1 <- sprintf("  %-8s %s", "Type:", x$type)
  cat(sprintf("%-30s %-8s %s\n", left_1, "Size:", fmt_bytes(x$uncompressed_size)))
  
  # --- Row 2: Layout & Disk ---
  left_2 <- sprintf("  %-8s %s", "Layout:", x$layout)
  cat(sprintf("%-30s %-8s %s\n", left_2, "Disk:", fmt_bytes(x$storage_size)))
  
  # --- Row 3: Chunks & Ratio ---
  chunk_str <- if (is.null(x$chunk_dims)) "N/A" else paste0("[", paste(x$chunk_dims, collapse = " x "), "]")
  left_3 <- sprintf("  %-8s %s", "Chunks:", chunk_str)
  
  ratio <- "N/A"
  if (x$storage_size > 0 && x$uncompressed_size > 0) {
    ratio <- sprintf("%.2fx", x$uncompressed_size / x$storage_size)
  }
  cat(sprintf("%-30s %-8s %s\n", left_3, "Ratio:", ratio))
  
  # --- Row 4: The Pipeline ---
  cat("  Pipeline: ")
  if (length(x$filters) == 0) {
    cat("None (Uncompressed)\n")
  } else {
    filter_names <- vapply(x$filters, function(f) {
      fname <- f$name
      
      # 1. Intercept Blosc (Version 1)
      if (fname == "blosc" && length(f$cd_values) >= 7) {
        shuffle  <- f$cd_values[6] # 0: None, 1: Byte, 2: Bit
        compcode <- f$cd_values[7]
        
        internal <- c()
        if (shuffle == 1) internal <- c(internal, "byteshuffle") # nocov
        if (shuffle == 2) internal <- c(internal, "bitshuffle")
        
        # 2. Intercept Blosc2
      } else if (fname == "blosc2" && length(f$cd_values) >= 7) {
        mask     <- f$cd_values[6] # Bitmask pre-filters
        compcode <- f$cd_values[7]
        
        internal <- c()
        if (bitwAnd(mask, 2) > 0) {
          internal <- c(internal, "bitshuffle")
        } else if (bitwAnd(mask, 1) > 0) {
          internal <- c(internal, "byteshuffle") # nocov
        }
        
        if (bitwAnd(mask, 4) > 0) internal <- c(internal, "delta")
        
        if (bitwAnd(mask, 8) > 0) {
          ndim <- if (length(f$cd_values) >= 8) f$cd_values[8] else 0
          meta_idx <- 8 + ndim + 1
          meta_val <- if (length(f$cd_values) >= meta_idx) f$cd_values[meta_idx] else "?"
          internal <- c(internal, sprintf("trunc(bits=%s)", meta_val))
        }
        
        # 3. Intercept Bitshuffle
      } else if (fname == "bitshuffle" && length(f$cd_values) >= 5) {
        comp <- f$cd_values[5]
        if (comp == 2) {
          return("bitshuffle -> lz4")
        } else if (comp == 3) {
          return("bitshuffle -> zstd")
        }
        return("bitshuffle") # nocov
        
        # 4. Intercept SZIP
      } else if (fname == "szip" && length(f$cd_values) >= 1) {
        mask <- f$cd_values[1]
        if (bitwAnd(mask, 32) > 0) {
          return("szip-nn")
        } else if (bitwAnd(mask, 4) > 0) {
          return("szip-ec")
        }
        return("szip") # nocov
        
      } else {
        # Not a dynamically expanded filter, return normal name
        return(fname)
      }
      
      # Decode the internal compressor code (Applies to Blosc and Blosc2)
      codec <- switch(as.character(compcode),
                      "0"  = "blosclz",
                      "1"  = "lz4",
                      "2"  = "lz4hc", 
                      "3"  = "snappy",
                      "4"  = "gzip",
                      "5"  = "zstd",
                      "32" = "ndlz",
                      "33" = "zfp-acc",
                      "34" = "zfp-prec",
                      "35" = "zfp-rate",
                      "36" = "openhtj2k",
                      "37" = "grok",
                      "38" = "openzl",
                      paste0("codec-", compcode))
      
      internal <- c(internal, codec)
      return(sprintf("%s [%s]", fname, paste(internal, collapse = " -> ")))
      
    }, FUN.VALUE = character(1))
    
    cat(paste(filter_names, collapse = " -> "), "\n")
  }
  
  invisible(x)
}

#' Create an HDF5 File Handle
#' 
#' Creates a file handle that provides a convenient, object-oriented interface
#' for interacting with and navigating a specific HDF5 file.
#'
#' @details
#' This function returns a special `h5` object that wraps the standard `h5lite`
#' functions. The primary benefit is that the `file` argument is pre-filled,
#' allowing for more concise and readable code when performing multiple
#' operations on the same file.
#'
#' For example, instead of writing:
#' ```r
#' h5_write(1:10, file, "dset1")
#' h5_write(2:20, file, "dset2")
#' h5_ls(file)
#' ```
#' You can create a handle and use its methods. Note that the `file` argument
#' is omitted from the method calls:
#' ```r
#' h5 <- h5_open("my_file.h5")
#' h5$write(1:10, "dset1")
#' h5$write(2:20, "dset2")
#' h5$ls()
#' h5$close()
#' ```
#'
#' @section Pass-by-Reference Behavior:
#' Unlike most R objects, the `h5` handle is an **environment**. This means it
#' is passed by reference. If you assign it to another variable (e.g.,
#' `h5_alias <- h5`), both variables point to the *same* handle. Modifying one
#' (e.g., by calling `h5_alias$close()`) will also affect the other.
#'
#' @section Interacting with the HDF5 File:
#' The `h5` object provides several ways to interact with the HDF5 file:
#'
#' \subsection{Standard `h5lite` Functions as Methods}{
#'   Most `h5lite` functions (e.g., `h5_read`, `h5_write`, `h5_ls`) are
#'   available as methods on the `h5` object, without the `h5_` prefix.
#'
#'   For example, `h5$write(data, "dset")` is equivalent to
#'   `h5_write(data, file, "dset")`.
#'
#'   The available methods are: `attr_names`, `cd`, `class`, `close`, 
#'   `create_group`, `delete`, `dim`, `exists`, `is_dataset`, `is_group`, 
#'   `length`, `ls`, `move`, `names`, `pwd`, `read`, `str`, `typeof`, `write`.
#' }
#'
#' \subsection{Navigation (`$cd()`, `$pwd()`)}{
#'   The handle maintains an internal working directory to simplify
#'   path management.
#'   \itemize{
#'     \item{`h5$cd(group)`: Changes the handle's internal working directory.
#'       This is a stateful, pass-by-reference operation. It understands absolute
#'       paths (e.g., `"/new/path"`) and relative navigation (e.g., `"../other"`).
#'       The target group does not need to exist.
#'     }
#'     \item{`h5$pwd()`: Returns the current working directory.}
#'   }
#'   When you call a method like `h5$read("dset")`, the handle automatically
#'   prepends the current working directory to any relative path. If you provide
#'   an absolute path (e.g., `h5$read("/path/to/dset")`), the working directory
#'   is ignored.
#' }
#'
#' \subsection{Closing the Handle (`$close()`)}{
#' The `h5lite` package does not keep files persistently open. Each operation
#' opens, modifies, and closes the file. Therefore, the `h5$close()` method
#' does not perform any action on the HDF5 file itself.
#'
#' Its purpose is to invalidate the handle, preventing any further operations
#' from being called. After `h5$close()` is called, any subsequent method
#' call (e.g., `h5$ls()`) will throw an error.
#' }
#'
#' @param file Path to the HDF5 file. The file will be created if it does not
#'   exist.
#' @return An object of class `h5` with methods for interacting with the file.
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' 
#' # Open the handle
#' h5 <- h5_open(file)
#' 
#' # Write data (note: 'data' is the first argument, 'file' is implicit)
#' h5$write(1:5, "vector")
#' h5$write(matrix(1:9, 3, 3), "matrix")
#' 
#' # Create a group and navigate to it
#' h5$create_group("simulations")
#' h5$cd("simulations")
#' print(h5$pwd()) # "/simulations"
#' 
#' # Write data relative to the current working directory
#' h5$write(rnorm(10), "run1") # Writes to /simulations/run1
#' 
#' # Read data
#' dat <- h5$read("run1")
#' 
#' # List contents of current WD
#' h5$ls()
#' 
#' # Close the handle
#' h5$close()
#' unlink(file)
h5_open <- function (file) {
  
  h5_create_file(file)
  
  env        <- new.env(parent = emptyenv())
  env$.file  <- file
  env$.wd    <- "/"
  class(env) <- "h5"
  
  env$read         = \(name = ".",       attr = NULL, as = "auto")                    { h5_run(env, h5_read)  }
  env$write        = \(data, name = ".", attr = NULL, as = "auto", compress = "gzip") { h5_run(env, h5_write) }
  env$ls           = \(name = ".", recursive = TRUE, full.names = FALSE)              { h5_run(env, h5_ls)    }
  env$attr_names   = \(name = ".")                     { h5_run(env, h5_attr_names)   }
  env$class        = \(name, attr = NULL)              { h5_run(env, h5_class)        }
  env$create_group = \(name)                           { h5_run(env, h5_create_group) }
  env$delete       = \(name, attr = NULL, warn = TRUE) { h5_run(env, h5_delete)       }
  env$dim          = \(name, attr = NULL)              { h5_run(env, h5_dim)          }
  env$exists       = \(name, attr = NULL)              { h5_run(env, h5_exists)       }
  env$inspect      = \(name)                           { h5_run(env, h5_inspect)      }
  env$is_dataset   = \(name)                           { h5_run(env, h5_is_dataset)   }
  env$is_group     = \(name)                           { h5_run(env, h5_is_group)     }
  env$length       = \(name = ".", attr = NULL)        { h5_run(env, h5_length)       }
  env$move         = \(from, to)                       { h5_run(env, h5_move)         }
  env$names        = \(name = ".")                     { h5_run(env, h5_names)        }
  env$str          = \(name = ".", attrs = TRUE)       { h5_run(env, h5_str)          }
  env$typeof       = \(name, attr = NULL)              { h5_run(env, h5_typeof)       }
  
  # Navigation methods
  env$cd = function (group = "/") {
    check_open(env)
    env$.wd <- normalize_path(env$.wd, group)
    return(invisible(NULL))
  }
  env$pwd = function () {
    check_open(env)
    return(env$.wd)
  }
  
  # Control methods
  env$close = function () {
    check_open(env)
    env$.file <- NULL
    env$.wd   <- NULL
    return(invisible(NULL))
  }
  
  return(env)
}

check_open <- function(env) {
  if (is.null(env$.file))
    stop("This h5 file handle has been closed.", call. = FALSE)
}


#' @noRd
#' @keywords internal
h5_run <- function(env, func) {
  
  check_open(env) # Ensure the handle is not closed
  
  # Capture the arguments passed to the calling function (e.g., h5$read)
  args <- as.list(parent.frame())
  
  # Normalize paths for 'name', 'from', and 'to' arguments if they exist
  if ("name" %in% names(args)) { args$name <- normalize_path(env$.wd, args$name) }
  if ("from" %in% names(args)) { args$from <- normalize_path(env$.wd, args$from) }
  if ("to"   %in% names(args)) { args$to   <- normalize_path(env$.wd, args$to)   }
  
  # Call the underlying h5lite function (e.g., h5_read) with the modified arguments
  args$file <- env$.file
  do.call(func, args, envir = parent.frame(n = 2))
}


#' @noRd
#' @keywords internal
normalize_path <- function(wd, path) {
  # An absolute path from the user overrides the working directory
  if (startsWith(path, "/")) {
    return(path)
  }
  
  # If path is absolute, start from root. Otherwise, start from current wd.
  start_dir <- if (startsWith(path, "/")) "/" else wd
  
  full_path <- file.path(start_dir, path, fsep = "/")
  
  # Split into components and process '..' and '.'
  parts <- strsplit(full_path, "/")[[1]]
  new_parts <- character(0)
  for (part in parts) {
    if (part == "" || part == ".") next
    if (part == "..") {
      if (length(new_parts) > 0) new_parts <- new_parts[-length(new_parts)]
    } else {
      new_parts <- c(new_parts, part)
    }
  }
  
  # Reconstruct the path
  if (length(new_parts) == 0) "/" else paste0("/", paste(new_parts, collapse = "/"))
}



#' @export
print.h5 <- function(x, ...) {
  if (is.null(x$.file)) {
    cat("<h5 handle for a closed file>\n")
  } else {
    size <- if (file.exists(x$.file)) file.size(x$.file) else NA
    n_objects <- if (file.exists(x$.file)) length(h5_ls(x$.file, recursive = FALSE)) else NA
    
    cat("<h5 handle>\n")
    cat("  File: ", x$.file, "\n")
    cat("  WD:   ", x$pwd(), "\n")
    if (!is.na(size)) {
      cat("  Size: ", format(structure(size, class = "object_size")), "\n")
    }
    if (!is.na(n_objects)) {
      cat("  Objects (root): ", n_objects, "\n")
    }
  }
  invisible(x)
}

#' @export
str.h5 <- function(object, ...) {
  object$str(...)
}

#' @export
as.character.h5 <- function(x, ...) {
  # Return the file path, or NULL if the handle is closed.
  x$.file
}

#' Define HDF5 Compression and Filter Settings
#' 
#' Constructs a comprehensive filter pipeline configuration to be passed 
#' as the `compress` argument to [h5_write()]. This function allows fine-grained 
#' control over chunking, pre-filters, compression algorithms, and data scaling.
#'
#' @param compress A string specifying the compression algorithm and optional level 
#'   (e.g., `"none"`, `"gzip"`, `"zstd-7"`, `"lz4"`, `"blosc1-lz4-9"`, 
#'   `"blosc2-gzip-3"`, `"blosc2-zstd"`). See the **Valid Compression Strings** 
#'   section below for an exhaustive list of supported formats. Default is `"gzip"`.
#' @param chunk_size An integer specifying the target chunk size in bytes. 
#'   Default is `1048576` (1 MB).
#' @param checksum A logical value indicating whether to apply the Fletcher32 
#'   checksum filter at the end of the pipeline to detect data corruption. Default is `FALSE`.
#' @param int_packing Control the HDF5 Scale-Offset filter for integer datasets. *(Note: 
#'   Incompatible with `szip`, `zfp`, `bshuf`, and Blosc2 pre-filters).*
#'   * `FALSE` (Default): Disabled.
#'   * `TRUE`: Automatically calculates and applies the mathematically optimal minimum 
#'     bit-width for each individual chunk.
#'   * Integer (e.g., `8`): Forces packing into exactly that many bits.
#' @param float_rounding Control the HDF5 Scale-Offset filter for floating-point 
#'   datasets.*(Note: Incompatible with `szip`, `zfp`, `bshuf`, and Blosc2 pre-filters).*
#'   * `NULL` (Default): Disabled.
#'   * Integer (e.g., `3`): The number of base-10 decimal places of detail to preserve 
#'     before truncating and packing the values (e.g., `3.141`). Negative numbers 
#'     round to powers of 10.
#' @param blosc2_delta A logical value. If `TRUE` and a `blosc2` compressor is selected, 
#'   applies the Blosc2 Delta pre-filter before compression. Default is `FALSE`.
#' @param blosc2_truncate An integer. If provided and a `blosc2` compressor is selected, 
#'   applies the Blosc2 Truncate Precision pre-filter to floating-point data, preserving 
#'   exactly the specified number of uncompressed bits. Default is `NULL`.
#' 
#' @section Valid Compression Strings:
#' The `compress` argument accepts a highly specific string syntax to define both 
#' the codec and its operational level.
#' 
#' ### Native / Core Codecs
#' 
#' * `"none"`: No compression.
#' * `"gzip-[level]"`: Levels `1` to `9`. Default is `5`. (e.g., `"gzip"` or `"gzip-9"`).
#' * `"zstd-[level]"`: Levels `1` to `22`. Default is `3`. (e.g., `"zstd"` or `"zstd-7"`).
#' * `"lz4-[level]"`: Levels `0` to `12`. Default is `0`. Level `0` is standard LZ4. Levels `1+` trigger LZ4-HC.
#' 
#' ### Bitshuffle Pre-filter
#' Forces the native Bitshuffle pre-filter before compression.
#' 
#' * `"bshuf-lz4"`: Bitshuffle + LZ4.
#' * `"bshuf-zstd-[level]"`: Bitshuffle + Zstd (Levels `1` to `22`).
#' 
#' ### Blosc Meta-compressors
#' Blosc applies its own highly optimized bitshuffling and multi-threading.
#' 
#' * **Blosc2 (Recommended):** `"blosc2"` (blosclz), `"blosc2-lz4-[level]"`, `"blosc2-zstd-[level]"`, `"blosc2-gzip-[level]"`, `"blosc2-ndlz"`
#' * **Blosc1 (Legacy):**  `"blosc1"` (blosclz), `"blosc1-lz4-[level]"`, `"blosc1-zstd-[level]"`, `"blosc1-gzip-[level]"`, `"blosc1-snappy"`
#' 
#' ### ZFP (Lossy Floating-Point Compression)
#' ZFP can be run standalone (for integers and floats) or inside Blosc2 (floats only). Unlike `[level]`, `[tolerance]` and `[bits]` are required.
#' 
#' * **Accuracy Mode** (Absolute error tolerance): `"zfp-acc-[tolerance]"` or `"blosc2-zfp-acc-[tolerance]"` (e.g., `"zfp-acc-0.001"`).
#' * **Precision Mode** (Bits of precision): `"zfp-prec-[bits]"` or `"blosc2-zfp-prec-[bits]"` (e.g., `"zfp-prec-16"`).
#' * **Rate Mode** (Bits of storage per value): `"zfp-rate-[bits]"` or `"blosc2-zfp-rate-[bits]"` (e.g., `"zfp-rate-8"`).
#' * **Reversible Mode** (Standalone Lossless): `"zfp-rev"`.
#' 
#' ### Legacy Codecs
#' 
#' * `"szip-nn"`, `"szip-ec"`: SZIP Nearest Neighbor or Entropy Coding.
#' * `"bzip2-[level]"`: Levels `1` to `9`. Default is `9`. (e.g., `"bzip2-4"`).
#' * `"lzf"`, `"snappy"`: Fast, unconfigurable legacy compressors.
#' 
#' @section Automatic Shuffling:
#' 
#' To maximize compression ratios without requiring users to manually manage 
#' complex pipeline interactions, `h5_compression` automatically configures the 
#' optimal shuffling pre-filter based on the following strict hierarchy:
#' 
#' **1. Blosc's Internal Bitshuffle (Preferred)**
#' If a Blosc meta-compressor is selected (e.g., `"blosc2-zstd"`), the pipeline 
#' automatically enables Blosc's highly optimized, internal bitshuffle routine. 
#' This achieves peak compression performance without requiring the standalone 
#' Bitshuffle plugin to be installed.
#' 
#' **2. Explicit Bitshuffle Plugin**
#' If a standard codec is explicitly prefixed with `bshuf-` (e.g., `"bshuf-lz4"`), 
#' the pipeline delegates to the standalone Bitshuffle plugin.
#' 
#' **3. Native HDF5 Byte Shuffle (Fallback)**
#' If a standard compressor is selected (e.g., `"zstd-5"` or `"gzip"`), 
#' the pipeline safely falls back to the core HDF5 library's native byte shuffle 
#' filter. This guarantees improved compression while maintaining universal 
#' compatibility.
#' 
#' **4. Strict Mutual Exclusions (When Shuffling is Disabled)**
#' To prevent data corruption or wasted CPU cycles, all shuffling is 
#' **forcefully disabled** in the following scenarios:
#' 
#' * **Scale-Offset Active:** If `int_packing` or `float_rounding` is applied, 
#'   shuffling is disabled because scale-offset destroys the byte-alignment 
#'   that shuffling relies on.
#' * **ZFP & SZIP:** These algorithms perform mathematical compression directly 
#'   on numerical values and will corrupt if the bitstream is rearranged beforehand.
#' * **1-Byte Data:** Characters, booleans, and 8-bit integers cannot be meaningfully 
#'   shuffled, so the step is skipped.
#' 
#' @return An S3 object of class `compress` containing the parsed pipeline parameters.
#' @seealso [h5_write()], `vignette('compression')`
#' @export
#' @examples
#' # 1. Simple fast compression (Zstd level 7)
#' h5_compression("zstd-7")
#' 
#' # 2. Optimal integer packing (Scale-Offset)
#' h5_compression("gzip-9", int_packing = TRUE)
#' 
#' # 3. Complex Blosc2 Pipeline (Delta + Zstd)
#' h5_compression("blosc2-zstd-5", blosc2_delta = TRUE)
#' 
#' # 4. Lossy ZFP compression (Tolerance of 0.05)
#' h5_compression("zfp-acc-0.05")
#' 
#' # Pass the compress object directly to h5_write
#' file <- tempfile(fileext = ".h5")
#' cmp  <- h5_compression("gzip-9", checksum = TRUE)
#' h5_write(combn(1:10, 3), file, "sets", compress = cmp)
#' 
#' print(cmp)
#' 
#' h5_inspect(file, "sets")
#' 
#' # Clean up
#' unlink(file)
#' 
h5_compression <- function(
    compress = "gzip", chunk_size = 1024 * 1024, checksum = FALSE, 
    int_packing = FALSE, float_rounding = NULL,
    blosc2_delta = FALSE, blosc2_truncate = NULL ) {
  
  if (inherits(compress, 'compress'))           return(compress)
  if (is.numeric(compress) && compress %in% 1:9) compress <- paste0("gzip-", compress)
  if (!is.character(compress))                   compress <- "none"
  compress <- tolower(trimws(compress))
  
  # 1. Handle polymorphic int_packing
  if      (identical(int_packing, FALSE)) { int_packing <- NA_integer_ }
  else if (identical(int_packing, TRUE))  { int_packing <- 0L          }
  else if (!(is.numeric(int_packing) && isTRUE(int_packing %% 1 == 0)))
    stop("`int_packing` must be TRUE/FALSE or an integer.", call. = FALSE)

  assert_scalar_character(compress)
  assert_scalar_logical(checksum, blosc2_delta)
  assert_scalar_integer(chunk_size)
  assert_scalar_integer(float_rounding, blosc2_truncate, .null_ok = TRUE)

  # Check string format against known valid patterns
  if (!any(
      grepl('^(gzip|zstd|lz4|bshuf-zstd|bzip2)(\\-[0-9]+)?$',              compress),
      grepl('^(none|zfp-rev|bshuf-lz4|szip-ec|szip-nn|lzf|snappy)$',       compress),
      grepl('^blosc1(\\-(snappy|(lz4|gzip|zstd)(\\-[0-9]+)?))?$',          compress),
      grepl('^blosc2(\\-(ndlz|(lz4|gzip|zstd)(\\-[0-9]+)?))?$',            compress),
      grepl('^(blosc2\\-)?zfp\\-((prec|rate)\\-[0-9]+|acc\\-[0-9\\.]+)$',  compress) )) {
    stop("Invalid `compress` string: '", compress, "'.", call. = FALSE)
  }
  
  blosc <- switch(substr(compress, 1, 6), 'blosc1' = 1L, 'blosc2' = 2L, 0L)

  # Extract the base codec name
  codecs <- c(
    'gzip', 'none', 'bshuf-lz4', 'bshuf-zstd', 'lz4', 'zstd', 
    'bzip2', 'lzf', 'zfp-rev', 'zfp-prec', 'zfp-acc', 'zfp-rate', 
    'szip-ec', 'szip-nn', 'snappy', 'ndlz', 'blosclz')
  for (codec in codecs)
    if (isTRUE(grepl(codec, compress, fixed = TRUE))) break
  
  # Parse the level, applying sensible defaults
  level_str <- rev(strsplit(compress, "-")[[1]])[[1]]
  level <- suppressWarnings(as.numeric(level_str))
  
  if (is.na(level)) {
    level <- switch(codec, 'gzip' = 5, 'zstd' = 3, 'bshuf-zstd' = 3, 'bzip2' = 9, 'lz4' = 0, 0)
  } else {
    rng <- switch(codec, 
      'gzip' = c(1, 9), 'bzip2' = c(1, 9), 'lz4' = c(0, 12), 
      'zstd' = c(1, 22), 'bshuf-zstd' = c(1, 22), NULL)
    
    if (!is.null(rng) && (!isTRUE(level %% 1 == 0) || level < rng[[1]] || level > rng[[2]])) {
      stop(compress, " level must be an integer between ", rng[[1]]," and ", rng[[2]], ".", call. = FALSE)
    }
  }

  if (is.null(float_rounding))  float_rounding  <- NA
  if (is.null(blosc2_truncate)) blosc2_truncate <- NA

  for (var in c('int_packing', 'float_rounding'))
    if (!is.na(get(var))) {
      if (grepl('(szip|zfp|bshuf)', codec)) stop("`", var, "` is not compatible with `", codec, "`.",     call. = FALSE)
      if (!is.na(blosc2_truncate))          stop("`", var, "` is not compatible with `blosc2_truncate`.", call. = FALSE)
      if (isTRUE(blosc2_delta))             stop("`", var, "` is not compatible with `blosc2_delta`.",    call. = FALSE)
    }

  structure(
    .Data = compress,
    class = 'compress',
    codec          = as.character(codec),
    level          = as.double(level),
    blosc          = as.integer(blosc), 
    int_packing    = as.integer(int_packing),
    float_rounding = as.integer(float_rounding),
    chunk_size     = as.integer(chunk_size), 
    checksum       = as.logical(checksum),
    b2_delta       = as.logical(blosc2_delta), 
    b2_trunc       = as.integer(blosc2_truncate)
  )
}

#' @keywords internal
#' @export
print.compress <- function(x, ...) {
  
  codec <- attr(x, "codec")
  lvl   <- attr(x, "level")
  blosc <- attr(x, "blosc")
  ip    <- attr(x, "int_packing")
  fr    <- attr(x, "float_rounding")
  
  # Construct the codec, size, and checksum strings
  codec_str <- codec
  if (blosc > 0) codec_str <- paste0("blosc", blosc, "-", codec_str)
  if (grepl('(gzip|zstd|lz4|bzip2|zfp)', codec)) codec_str <- paste0(codec_str, "-", lvl)
  codec_str <- sub('lz4-0', 'lz4', codec_str)
  
  size_str     <- fmt_bytes(attr(x, "chunk_size"))
  checksum_str <- if (attr(x, "checksum")) "Fletcher32" else "None"
  
  # Determine Shuffle Filter Logic
  is_scaled <- !is.na(ip) || !is.na(fr)
  
  if      (codec == "none")        { shuffle_str <- "None"                            }
  else if (is_scaled)              { shuffle_str <- "None (Disabled by Scale-Offset)" }
  else if (grepl("^szip",  codec)) { shuffle_str <- "None (Incompatible with szip)"   }
  else if (grepl("^zfp",   codec)) { shuffle_str <- "None (Incompatible with zfp)"    }
  else if (grepl("^bshuf", codec)) { shuffle_str <- "Bitshuffle (Standalone Plugin)"  }
  else if (blosc > 0)              { shuffle_str <- "Bitshuffle (Blosc Internal)"     }
  else                             { shuffle_str <- "Byte Shuffle (Native HDF5)"      }
  
  # Print Configuration
  cat("<HDF5 Compression Configuration>\n")
  cat(sprintf("  %-16s %s\n", "Codec:",      codec_str))
  cat(sprintf("  %-16s %s\n", "Shuffle:",    shuffle_str))
  cat(sprintf("  %-16s %s\n", "Chunk Size:", size_str))
  cat(sprintf("  %-16s %s\n", "Checksum:",   checksum_str))
  
  # Optional: Scale-Offset Modifiers
  if (!is.na(ip)) {
    val <- if (ip == 0) "Optimal (Auto)" else paste(ip, "bits")
    cat(sprintf("  %-16s %s\n", "Int Packing:", val))
  }
  
  if (!is.na(fr)) {
    cat(sprintf("  %-16s %s decimal places\n", "Float Rounding:", fr))
  }
  
  # Optional: Blosc2 Modifiers
  if (blosc == 2) {
    b2_d <- attr(x, "b2_delta")
    b2_t <- attr(x, "b2_trunc")
    if (isTRUE(b2_d)) cat(sprintf("  %-16s %s\n", "Blosc2 Delta:", "TRUE"))
    if (!is.na(b2_t)) cat(sprintf("  %-16s %s bits\n", "Blosc2 Trunc:", b2_t))
  }
  
  invisible(x)
}

# ==============================================================================
# 1. R-Level: h5_compression() Argument Parsing & Defaults
# ==============================================================================
local({
  #test_that("h5_compression parses inputs and applies defaults", {
  
  # Base object checks
  expect_inherits(h5_compression(), "compress")
  
  # Identity checks (inherits 'compress')
  fs <- h5_compression("zstd-5")
  expect_equal(h5_compression(fs), fs)
  
  # Numeric & Null conversions
  expect_equal(as.character(h5_compression(5)), "gzip-5")
  expect_equal(as.character(h5_compression(NULL)), "none")
  
  # Codec defaults
  expect_equal(attr(h5_compression("zstd"), "level"), 3)
  expect_equal(attr(h5_compression("gzip"), "level"), 5)
  expect_equal(attr(h5_compression("lz4"), "level"), 0)
  expect_equal(attr(h5_compression("bzip2"), "level"), 9)
  
  # int_packing logic
  expect_equal(attr(h5_compression(int_packing = TRUE), "int_packing"), 0L)
  expect_equal(attr(h5_compression(int_packing = 8), "int_packing"), 8L)
  expect_true(is.na(attr(h5_compression(int_packing = FALSE), "int_packing")))
  
  # float_rounding logic
  expect_equal(attr(h5_compression(float_rounding = 3), "float_rounding"), 3L)
  expect_true(is.na(attr(h5_compression(), "float_rounding")))
})


# ==============================================================================
# 2. R-Level: h5_compression() Error Triggers (stop() cases)
# ==============================================================================
local({
  #test_that("h5_compression triggers all boundary and type errors", {
  
  # Type mismatches
  expect_error(h5_compression(chunk_size = 1.23), "must be a scalar integer")
  expect_error(h5_compression(int_packing = "yes"), "must be TRUE/FALSE or an integer")
  expect_error(h5_compression(int_packing = 1.5), "must be TRUE/FALSE or an integer")
  expect_error(h5_compression("invalid-codec"), "Invalid `compress` string")
  
  # Level boundaries
  expect_error(h5_compression("gzip-10"), "level must be an integer between")
  expect_error(h5_compression("zstd-25"), "level must be an integer between")
  expect_error(h5_compression("lz4-15"), "level must be an integer between")
  expect_error(h5_compression("bzip2-0"), "level must be an integer between")
  
  # Mutually exclusive filters (Scale-Offset Conflicts)
  expect_error(h5_compression("szip-nn", int_packing = TRUE), "not compatible with `szip-nn`")
  expect_error(h5_compression("zfp-rate-8", float_rounding = 3), "not compatible with `zfp-rate`")
  expect_error(h5_compression("bshuf-lz4", int_packing = TRUE), "not compatible with `bshuf-lz4`")
  
  # Blosc modifier conflicts
  expect_error(
    h5_compression("blosc2-zstd", float_rounding = 2, blosc2_truncate = 16), 
    "not compatible with `blosc2_truncate`"
  )
  expect_error(
    h5_compression("blosc2-zstd", float_rounding = 2, blosc2_delta = TRUE), 
    "not compatible with `blosc2_delta`"
  )
})


# ==============================================================================
# 3. C-Level: apply_compression() Valid Codec & Pipeline Applications
# ==============================================================================
local({
  #test_that("apply_compression succeeds for all standard codecs", {
  f <- tempfile(fileext = ".h5")
  on.exit(unlink(f))
  
  v_int <- 1:100
  
  # Base HDF5 Algorithms (triggers native H5Pset_shuffle due to type_size > 1)
  expect_silent(h5_write(v_int, f, "gzip", compress = h5_compression("gzip-9")))
  expect_silent(h5_write(v_int, f, "zstd", compress = h5_compression("zstd-22"))) 
  expect_silent(h5_write(v_int, f, "lz4_fast", compress = h5_compression("lz4-0")))
  expect_silent(h5_write(v_int, f, "lz4_hc", compress = h5_compression("lz4-9")))
  
  # SZIP
  expect_silent(h5_write(v_int, f, "szip_nn", compress = h5_compression("szip-nn")))
  expect_silent(h5_write(v_int, f, "szip_ec", compress = h5_compression("szip-ec")))
  
  # Standalone Bitshuffle
  expect_silent(h5_write(v_int, f, "bshuf_lz4", compress = h5_compression("bshuf-lz4")))
  expect_silent(h5_write(v_int, f, "bshuf_zstd", compress = h5_compression("bshuf-zstd-5")))
})

local({
  #test_that("apply_compression succeeds for Blosc architectures", {
  f <- tempfile(fileext = ".h5")
  on.exit(unlink(f))
  
  v_int <- 1:100
  
  # Legacy Blosc
  expect_silent(h5_write(v_int, f, "blosc1", compress = h5_compression("blosc1-lz4-5")))
  
  # Blosc2 inner codecs
  expect_silent(h5_write(v_int, f, "b2_zstd", compress = h5_compression("blosc2-zstd-5")))
  expect_silent(h5_write(v_int, f, "b2_lz4", compress = h5_compression("blosc2-lz4-9")))
  expect_silent(h5_write(v_int, f, "b2_gzip", compress = h5_compression("blosc2-gzip-5")))
  expect_silent(h5_write(v_int, f, "b2_ndlz", compress = h5_compression("blosc2-ndlz")))
  
  # Blosc2 Modifiers
  expect_silent(h5_write(v_int, f, "b2_delta", 
                         compress = h5_compression("blosc2-zstd-5", blosc2_delta = TRUE)))
  
  expect_silent(h5_write(rnorm(100), f, "b2_trunc", 
                         compress = h5_compression("blosc2-zstd-5", blosc2_truncate = 16)))
})

local({
  #test_that("apply_compression succeeds for all ZFP modes", {
  f <- tempfile(fileext = ".h5")
  on.exit(unlink(f))
  
  v_float <- 1:100 + rnorm(100) / 10
  
  # Standalone ZFP
  expect_silent(h5_write(v_float, f, "zfp_rev", compress = h5_compression("zfp-rev")))
  expect_silent(h5_write(v_float, f, "zfp_rate", compress = h5_compression("zfp-rate-16")))
  expect_silent(h5_write(v_float, f, "zfp_prec", compress = h5_compression("zfp-prec-16")))
  expect_silent(h5_write(v_float, f, "zfp_acc", compress = h5_compression("zfp-acc-0.001")))
  
  # Blosc2 ZFP
  expect_silent(h5_write(v_float, f, "b2_zfp_rate", compress = h5_compression("blosc2-zfp-rate-16")))
  expect_silent(h5_write(v_float, f, "b2_zfp_prec", compress = h5_compression("blosc2-zfp-prec-16")))
  expect_silent(h5_write(v_float, f, "b2_zfp_acc", compress = h5_compression("blosc2-zfp-acc-0.001")))
})


# ==============================================================================
# 4. C-Level: Defensive Fallbacks & Edge Cases
# ==============================================================================
local({
  #test_that("apply_compression safely degrades un-chunkable or invalid data", {
  f <- tempfile(fileext = ".h5")
  on.exit(unlink(f))
  
  # 1. Rank 0 (Scalar Bypass)
  expect_silent(h5_write(I(5), f, "scalar", compress = h5_compression("gzip-5")))
  
  # 2. Checksum Application
  expect_silent(h5_write(1:100, f, "checksum", 
                         compress = h5_compression("none", checksum = TRUE)))
  
  # 3. Scale-Offset Application (Directly valid)
  expect_silent(h5_write(1:100, f, "so_int", 
                         compress = h5_compression("gzip-5", int_packing = TRUE)))
  expect_silent(h5_write(rnorm(100), f, "so_float", 
                         compress = h5_compression("gzip-5", float_rounding = 3)))
  
  # 4. SZIP Constraints (No R-level hacking required for these)
  fs_szip <- h5_compression("szip-nn")
  # Non-numeric string -> falls back to gzip
  expect_silent(h5_write(letters[1:10], f, "szip_char", compress = fs_szip)) 
  # Dim < 32 but >= 2 -> chunk adjusted to 14
  expect_silent(h5_write(1:15, f, "szip_15", compress = fs_szip)) 
  # Dim = 1 -> chunk adjusted to gzip fallback
  expect_silent(h5_write(1, f, "szip_1", compress = fs_szip)) 
  
  # 5. ZFP Constraints (No R-level hacking required)
  fs_zfp <- h5_compression("zfp-rev")
  fs_b2_zfp <- h5_compression("blosc2-zfp-rate-8")
  
  # Standalone not numeric -> gzip
  expect_silent(h5_write(letters[1:10], f, "zfp_char", compress = fs_zfp)) 
  # Standalone size != 4 or 8 (RAW is 1 byte) -> gzip
  expect_silent(h5_write(as.raw(1:10), f, "zfp_raw", compress = fs_zfp)) 
  # Blosc-ZFP applied to integers -> gzip fallback
  expect_silent(h5_write(1:100, f, "b2_zfp_int", compress = fs_b2_zfp)) 
})

local({
  #test_that("apply_compression defensive C-level modifier fallbacks trigger", {
  f <- tempfile(fileext = ".h5")
  on.exit(unlink(f))
  
  # To test the C-level defensive downgrades (like preventing scaled data from 
  # hitting SZIP/ZFP), we must intentionally hack an R S3 object to bypass the 
  # R-level `stop()` checks we validated in Test Block 2.
  
  # Hack 1: Force szip + Scale-Offset to trigger the `is_scaled` fallback to GZIP
  fs_hack_szip <- h5_compression("szip-ec")
  attr(fs_hack_szip, "float_rounding") <- 2L
  expect_silent(h5_write(rnorm(100), f, "hack_szip", compress = fs_hack_szip))
  
  # Hack 2: Force zfp + Scale-Offset to trigger the `is_scaled` fallback to GZIP
  fs_hack_zfp <- h5_compression("zfp-rate-16")
  attr(fs_hack_zfp, "float_rounding") <- 2L
  expect_silent(h5_write(rnorm(100), f, "hack_zfp", compress = fs_hack_zfp))
})

# ==============================================================================
# 5. R-Level: print.compress() Code Coverage
# ==============================================================================
local({
  #test_that("print.compress evaluates base configuration and unit math", {
  
  # Base output and 0-byte (B) edge case + Checksum
  c_base <- h5_compression("none", chunk_size = 500, checksum = TRUE)
  expect_stdout(print(c_base), "<HDF5 Compression Configuration>")
  expect_stdout(print(c_base), "Codec:\\s+none")
  expect_stdout(print(c_base), "Shuffle:\\s+None")
  expect_stdout(print(c_base), "Chunk Size:\\s+500.00 B")
  expect_stdout(print(c_base), "Checksum:\\s+Fletcher32")
  
  # KB size conversion
  c_kb <- h5_compression("gzip-5", chunk_size = 2048)
  expect_stdout(print(c_kb), "Chunk Size:\\s+2.00 KB")
  
  # MB size conversion (default 1048576)
  c_mb <- h5_compression("lz4-0")
  expect_stdout(print(c_mb), "Chunk Size:\\s+1.00 MB")
  
  # GB size conversion
  c_gb <- h5_compression("zstd-5", chunk_size = 1024^3)
  expect_stdout(print(c_gb), "Chunk Size:\\s+1.00 GB")
  
  # Force NA level to test `if (!is.na(lvl))` bypass
  c_na <- c_base
  attr(c_na, "level") <- NA
  expect_stdout(print(c_na), "Codec:\\s+none\n")
})

local({
  #test_that("print.compress evaluates shuffle exclusion hierarchy", {
  
  # 1. is_scaled (Highest priority override)
  c_scaled <- h5_compression("gzip-5", int_packing = TRUE)
  expect_stdout(print(c_scaled), "Shuffle:\\s+None \\(Disabled by Scale-Offset\\)")
  
  # 2. SZIP
  c_szip <- h5_compression("szip-nn")
  expect_stdout(print(c_szip), "Shuffle:\\s+None \\(Incompatible with szip\\)")
  
  # 3. ZFP
  c_zfp <- h5_compression("zfp-rate-16")
  expect_stdout(print(c_zfp), "Shuffle:\\s+None \\(Incompatible with zfp\\)")
  
  # 4. Standalone Bitshuffle
  c_bshuf <- h5_compression("bshuf-lz4")
  expect_stdout(print(c_bshuf), "Shuffle:\\s+Bitshuffle \\(Standalone Plugin\\)")
  
  # 5. Blosc Internal Bitshuffle
  c_blosc <- h5_compression("blosc1-lz4-5")
  expect_stdout(print(c_blosc), "Shuffle:\\s+Bitshuffle \\(Blosc Internal\\)")
  
  # 6. Native Fallback
  c_native <- h5_compression("gzip-9")
  expect_stdout(print(c_native), "Shuffle:\\s+Byte Shuffle \\(Native HDF5\\)")
})

local({
  #test_that("print.compress evaluates modifiers correctly", {
  
  # Int Packing: Optimal / Auto (0L)
  c_ip_auto <- h5_compression("gzip-5", int_packing = TRUE)
  expect_stdout(print(c_ip_auto), "Int Packing:\\s+Optimal \\(Auto\\)")
  
  # Int Packing: Manual Bits
  c_ip_man <- h5_compression("gzip-5", int_packing = 8)
  expect_stdout(print(c_ip_man), "Int Packing:\\s+8 bits")
  
  # Float Rounding
  c_fr <- h5_compression("gzip-5", float_rounding = 3)
  expect_stdout(print(c_fr), "Float Rounding:\\s+3 decimal places")
  
  # Blosc2 Modifiers
  c_b2 <- h5_compression("blosc2-zstd-5", blosc2_delta = TRUE, blosc2_truncate = 16)
  expect_stdout(print(c_b2), "Codec:\\s+blosc2-zstd-5")
  expect_stdout(print(c_b2), "Blosc2 Delta:\\s+TRUE")
  expect_stdout(print(c_b2), "Blosc2 Trunc:\\s+16 bits")
  
  # Blosc2 without modifiers (covers the nested if bypasses)
  c_b2_bare <- h5_compression("blosc2-zstd-5")
  # Assert it prints normally without throwing errors for missing attributes
  expect_stdout(print(c_b2_bare), "Codec:\\s+blosc2-zstd-5")
})

if (requireNamespace("tinytest", quietly = TRUE)) {
  
  library(h5lite)
  
  if (dir.exists("tinytests")) { test_dir <- "tinytests"       }
  else                         { test_dir <- "tests/tinytests" }
  
  results <- tinytest::run_test_dir(test_dir)
  passed  <- as.logical(results)
  
  print(results)
  
  if (!all(passed)) {
    stop("One or more tinytest tests failed.")
  }
}

#' List HDF5 Objects
#' 
#' Lists the names of objects (datasets and groups) within an HDF5 file or group.
#'
#' @param file The path to the HDF5 file.
#' @param name The group path to start listing from. Defaults to the root group (`/`).
#' @param recursive If `TRUE` (default), lists all objects found recursively
#'   under `name`. If `FALSE`, lists only the immediate children.
#' @param full.names If `TRUE`, the full paths from the file's root are
#'   returned. If `FALSE` (the default), names are relative to `name`.
#' @param scales If `TRUE`, also returns datasets that are dimensions scales for
#'   other datasets.
#' 
#' @return A character vector of object names. If `name` is `/` (the default),
#'   the paths are relative to the root of the file. If `name` is another group,
#'   the paths are relative to that group (unless `full.names = TRUE`).
#' 
#' @seealso [h5_attr_names()], [h5_str()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' h5_create_group(file, "foo/bar")
#' h5_write(1:5, file, "foo/data")
#' 
#' # List everything recursively
#' h5_ls(file)
#' 
#' # List only top-level objects
#' h5_ls(file, recursive = FALSE)
#' 
#' # List relative to a sub-group
#' h5_ls(file, "foo")
#' 
#' unlink(file)
h5_ls <- function(file, name = "/", recursive = TRUE, full.names = FALSE, scales = FALSE) {

  file <- validate_strings(file, name, must_exist = TRUE)
  assert_scalar_logical(recursive, full.names, scales)
  
  # Call the C function that performs a recursive or non-recursive listing.
  .Call("C_h5_ls", file, name, recursive, full.names, scales, PACKAGE = "h5lite")
}

#' Display the Structure of an HDF5 Object
#'
#' Recursively prints a summary of an HDF5 group or dataset, similar to
#' the structure of `h5ls -r`. It displays the nested structure, object types,
#' dimensions, and attributes.
#'
#' @details
#' This function provides a quick and convenient way to inspect the contents of
#' an HDF5 file. It performs a recursive traversal of the file from the C-level
#' and prints a formatted summary to the R console.
#'
#' This function **does not read any data** into R. It only inspects the
#' metadata (names, types, dimensions) of the objects in the file, making it
#' fast and memory-safe for arbitrarily large files.
#'
#' @param file The path to the HDF5 file.
#' @param name The name of the group or dataset to display. Defaults to the root
#'   group "/".
#' @param attrs Set to `FALSE` to hide attributes. The default (`TRUE`) shows
#'   attributes prefixed with `@`.
#' @param members Set to `FALSE` to hide compound dataset members. The default
#'   (`TRUE`) shows members prefixed with `$`.
#' @param markup Set to `FALSE` to remove colors and italics from the output.
#' @return This function is called for its side-effect of printing to the
#'   console and returns \code{NULL} invisibly.
#' @seealso [h5_ls()], [h5_attr_names()]
#' @export
#' @examples
#' file <- tempfile(fileext = ".h5")
#' h5_write(list(x = 1:10, y = matrix(1:9, 3, 3)), file, "group")
#' h5_write("metadata", file, "group", attr = "info")
#' 
#' # Print structure
#' h5_str(file)
#' 
#' unlink(file)
h5_str <- function(file, name = "/", attrs = TRUE, members = TRUE, markup = interactive()) {
  
  file <- validate_strings(file, name, must_exist = TRUE)
  assert_scalar_logical(attrs, members, markup)
  
  # Call the C function that recursively visits objects and prints a summary.
  .Call("C_h5_str", file, name, attrs, members, markup, PACKAGE = "h5lite")
  invisible(NULL)
}

---
title: "Partial Reading"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Partial Reading}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

When working with large datasets, loading an entire HDF5 object into R's memory isn't always feasible—or necessary. `h5lite` provides a highly efficient "partial reading" feature using the `start` and `count` parameters in `h5_read()`. 

This vignette explains why partial reading is vastly more efficient than R's standard indexing for large data, and how to use the "smart" `start` parameter across different data structures.

## Why Partial Reading Matters

If you are working with small datasets, partial reading isn't strictly necessary. By default, `h5lite` chunks data in 1 MB blocks. For objects smaller than this, reading the whole dataset into memory and subsetting it in R is perfectly fine.

However, when dealing with datasets that span gigabytes and exceed your system's RAM, partial reading becomes essential.

### The HDF5 Storage Model: Chunking and Compression

To understand why `start` and `count` are designed the way they are, it helps to understand how HDF5 stores data. 



Unlike a standard CSV, HDF5 datasets are divided into "chunks" which are compressed individually. When you want to read a specific piece of data, the HDF5 library must locate the chunk containing that data, decompress the *entire* chunk into memory, and then extract your requested values.

If you request a contiguous block of data, HDF5 only needs to decompress a handful of chunks. This is incredibly fast. 

However, if you try to use typical random-access indexing—for example, trying to extract a single column from a massive, row-oriented HDF5 matrix—the library has to decompress almost every single chunk in the dataset just to piece together that one column. To fetch a single column, it is often faster to read the entire dataset into R first and then subset it.

### Designing for Partial Reading

If you are the one designing and writing the HDF5 file, you should actively consider optimizing your data storage for partial reading. Well-designed HDF5 files lay out large datasets in such a way that users can extract useful subsets while only decompressing a minimal number of internal chunks. For instance, if you anticipate that users will primarily extract data row-by-row, your data should be oriented so that rows are kept contiguous.

The "smart" `start` parameter is purposefully designed to work seamlessly with datasets that are arranged optimally in this way, ensuring that the most efficient access patterns are also the easiest to type.

### Memory Efficiency of `start` and `count`

Another massive benefit of partial reading is the memory footprint of the request itself. 

In standard R, if you want to extract the first ten million elements of a vector, you might write `vec[1:10000000]`. Behind the scenes, R expands `1:10000000` into an actual vector of ten million 32-bit integers. That index vector alone consumes nearly 40 MB of RAM just to be passed as an argument!

In `h5lite`, fetching those same ten million elements looks like this: `h5_read(file, "vec", start = 1, count = 10000000)`. Those two arguments are passed as simple numeric values, consuming just 16 bytes.

---

## The "Smart" `start` Parameter

The `start` parameter is designed to relieve you from doing complex index math. Assuming your HDF5 file is well-designed and stores data in the most logical way it will be retrieved, **90% of the time you only need to provide a single integer to `start`**. 

When you provide a single integer, `start` automatically applies itself to the most meaningful dimension of the dataset:

* **1D Vector:** `start` specifies the **element**.
* **2D Matrix:** `start` specifies the **row**.
* **2D Data Frame:** `start` specifies the **row**.
* **3D Array:** `start` specifies the **2D matrix**.

The `count` parameter is an optional single integer that simply says, "Starting from `start`, how many of these structural units do you want to read?" 

### Single-Value Examples

Here is how this intuitive behavior looks in practice across different shapes of data when fetching a block of units:

```{r single-value}
library(h5lite)
file <- tempfile(fileext = ".h5")

# --- 1. Vectors (Element-level targeting) ---
h5_write(seq(10, 100, by = 10), file, "my_vector")

# Start at the 4th element, read 3 elements
h5_read(file, "my_vector", start = 4, count = 3)

# --- 2. Matrices (Row-level targeting) ---
mat <- matrix(1:50, nrow = 10, ncol = 5)
h5_write(mat, file, "my_matrix")

# Start at row 5, read 3 complete rows (automatically spans all columns)
h5_read(file, "my_matrix", start = 5, count = 3)

# --- 3. Data Frames (Row-level targeting) ---
h5_write(mtcars, file, "my_mtcars")

# Start at row 10, read 5 complete rows
h5_read(file, "my_mtcars", start = 10, count = 5)

# --- 4. 3D Arrays (Matrix-level targeting) ---
arr <- array(1:24, dim = c(2, 3, 4)) 
h5_write(arr, file, "my_array")

# Start at the 2nd matrix, read 2 complete matrices
h5_read(file, "my_array", start = 2, count = 2)
```

### Dimension Simplification (Exact vs. Range Indexing)

`h5lite` mimics R's native subsetting behavior when it comes to preserving or dropping dimensions. This behavior is controlled entirely by whether you include the `count` argument.

**Exact Indexing (Omitting `count`)** If you provide `start` but omit `count`, `h5lite` assumes you are requesting an exact point index. It will read 1 unit and **drop** the targeted dimension to simplify the resulting data structure.

```{r exact-index}
# Read exactly row 5 of the matrix. 
# The row dimension is dropped, returning a 1D vector.
row_vec <- h5_read(file, "my_matrix", start = 5)
row_vec

class(row_vec)
```

**Range Indexing (Providing `count`)** If you explicitly provide `count` (even if `count = 1`), `h5lite` assumes you are reading a range. The dataset's original dimensions are **preserved**. This is incredibly useful when programming dynamically and you need to guarantee that your matrix remains a matrix, even if your batch loop happens to fetch only a single row.

```{r range-index}
# Read row 5, but signal a range request by setting count = 1.
# The original geometry is preserved, returning a 1x5 matrix.
row_mat <- h5_read(file, "my_matrix", start = 5, count = 1)
row_mat

class(row_mat)
```

### Drilling Down: Multi-Value `start` and N-Dimensional Arrays

While the single-value form covers most use cases, `start` is flexible enough to target lower-rank dimensions for unusual or highly specific extractions. 

If you need to extract a specific contiguous block *inside* a matrix or array, you can pass a vector of integers to `start`. When you do this, the `count` dropping rules apply to the **last** dimension you specify, while all preceding dimensions are treated as exact point indices and dropped unconditionally.

To make this intuitive, `start` maps its values to the dataset's dimensions in a specific priority order, targeting the "outermost" structural blocks first, and the specific rows/columns last. For any N-dimensional array, the mapping order is:

* **Priority Order:** `Dimension N, Dimension N-1, ..., Dimension 3, Dimension 1 (Rows), Dimension 2 (Cols)`

For a 3D array, this means the first value targets the matrix, the second targets the row, and the third targets the column.

```{r multi-value}
# Matrix: Start at row 5, column 2, and read 3 elements along that row.
# The row is an exact point index (dropped). The columns are a range (preserved).
# Returns a 1D vector of length 3.
h5_read(file, "my_matrix", start = c(5, 2), count = 3)

# Matrix: Extract exactly row 5, column 2. 
# Because count is omitted, the final dimension is also dropped.
# Returns an unnamed scalar value.
h5_read(file, "my_matrix", start = c(5, 2))

# 3D Array: Target matrix 2, row 1.
# The matrix and row are exact point indices (dropped). 
# Returns a 1D vector containing the columns of that specific row.
h5_read(file, "my_array", start = c(2, 1))
```

*(Note: Data frames are a special case. Because HDF5 stores data frames as 1-dimensional lists of compound records, they do not have columns in the same structural way a matrix does. Therefore, `start` for a data frame must always be a single integer targeting the row. To get specific columns, read the rows you need first, then subset the columns in R).*

```{r cleanup, include=FALSE}
unlink(file)
```

---
title: "Data Types & Compression"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Data Types & Compression}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

`h5lite` is designed to seamlessly map R's diverse data structures to HDF5's portable format. This vignette explains the supported R data types, how `h5lite` writes them to HDF5, and how you can precisely control data types and compression when needed.

```{r setup}
library(h5lite)
file <- tempfile(fileext = ".h5")
```

## Supported Data Types

`h5lite` supports reading and writing a wide range of R data types. The table below lists the default mapping when writing to HDF5.

| R Data Type    | HDF5 Equivalent  | Description                                    |
| :------------- | :--------------- | :--------------------------------------------- |
| **Numeric**    | *variable*       | Selects optimal type: `uint8`, `float32`, etc. |
| **Logical**    | `H5T_STD_U8LE`   | Stored as 0 (FALSE) or 1 (TRUE) (`uint8`).     |
| **Character**  | `H5T_STRING`     | Variable or fixed-length UTF-8 strings.        |
| **Complex**    | `H5T_COMPLEX`    | Native HDF5 2.0+ complex numbers.              |
| **Raw**        | `H5T_OPAQUE`     | Raw bytes / binary data.                       |
| **Factor**     | `H5T_ENUM`       | Integer indices with label mapping.            |
| **integer64**  | `H5T_STD_I64LE`  | 64-bit signed integers via `bit64` package.    |
| **POSIXt**     | `H5T_STRING`     | ISO 8601 string (`YYYY-MM-DDTHH:MM:SSZ`).      |
| **List**       | `H5O_TYPE_GROUP` | Recursive container structure.                 |
| **Data Frame** | `H5T_COMPOUND`   | Table of mixed types.                          |
| **NULL**       | `H5S_NULL`       | Creates a placeholder.                         |

## Dimensions: Scalars, Vectors, and Arrays

Atomic data types (Integer, integer64, Double, Logical, Character, Complex, Raw, and POSIXt) can be written to HDF5 as scalars, 1D vectors, or N-dimensional arrays.

* **Scalars:** To write a single value as a true HDF5 scalar (0 dimensions), you must wrap the value in `I()`.
* **Vectors:** Standard R vectors are written as 1D arrays (Simple Dataspace with rank 1).
* **Arrays/Matrices:** R objects with `dim` attributes are written as N-dimensional datasets, preserving their shape.

```{r}
# 1. Scalar (0 dims)
h5_write(I(42), file, "structure/scalar")

# 2. Vector (1 dim)
h5_write(c(1, 2, 3), file, "structure/vector")

# 3. Matrix (2 dims)
h5_write(matrix(1:9, 3, 3), file, "structure/matrix")
```

*For more complex dimensional structures, refer to `vignette('matrices')`.*

## Numeric Data

R uses 32-bit integers and 64-bit doubles. When writing with `as = "auto"`, `h5lite` analyzes the range of your data to select the most compact HDF5 type.

* **Default:** Selects optimal type based on range of values.
* **With NA:** `float64` (`H5T_IEEE_F64LE`)
* **Fractional Values:** Double-precision vectors with fractional values default to `float64`.
* **Coercion:** You can override this using `int[8|16|32|64]`, `uint[8|16|32|64]`, `float[16|32|64]`, or `bfloat16`.

```{r}
# Integers between 0 and 255 (uint8)
h5_write(c(1L, 2L, 3L), file, "integers/small")

# Integers with NA -> float64
h5_write(c(1L, NA, 3L), file, "integers/with_na")

# Force larger type (int16)
h5_write(1:100, file, "integers/short", as = "int16")
```

## 64-bit Integers (`integer64`)

* **Default:** `int64` (`H5T_STD_I64LE`)
* **Coercion:** none

R does not natively support 64-bit integers, but `h5lite` supports reading and writing them via the `bit64` package.

```{r}
if (requireNamespace("bit64", quietly = TRUE)) {
  val <- bit64::as.integer64(c("9223372036854775807", "-9223372036854775807"))
  h5_write(val, file, "integers/int64")
}
```

## Double (Numeric) Data

R's default numeric type is double-precision.

* **Default:** `float64` (`H5T_IEEE_F64LE`)
* **Coercion:** `int[8|16|32|64]`, `uint[8|16|32|64]`, `float[16|32|64]`, or `bfloat16`

```{r}
data <- rnorm(10)

# Default (float64)
h5_write(data, file, "doubles/default")

# Single Precision (float32) - Saves 50% space
h5_write(data, file, "doubles/float32", as = "float32")
```

## Logical Data

* **Default:** `uint8` (`H5T_STD_U8LE`)
* **With NA:** `float64` (`H5T_IEEE_F64LE`)
* **Coercion:** `int[8|16|32|64]`, `uint[8|16|32|64]`, `float[16|32|64]`, or `bfloat16`

```{r}
bools <- sample(c(TRUE, FALSE), 1000, replace = TRUE)

h5_write(bools, file, "logicals/packed")
```

## Character Data

HDF5 supports two methods for storing strings. By default (`as = "auto"`), `h5lite` chooses the best approach:

* **Variable-Length:** Used if the vector contains `NA` or if string lengths are highly inconsistent.
* **Fixed-Length:** Used for short, consistent strings without `NA` to allow for compression.

### **Variable-Length:**

Explicitly requested with `as = "utf8"` or `as = "ascii"`.

* Compressible: **NO**
* Handles `NA`: **YES**

```r
# UTF-8 variable length
h5_write(c("apple", "banana", NA), file, "strings/var_utf8")

# ASCII variable length
h5_write(c("A", "B", "C"), file, "strings/var_ascii", as = "ascii")
```

### **Fixed-Length:**

Use `as = "ascii[10]"` / `as = "utf8[10]"` (explicit size=10) or `as = "ascii[]"` / `as = "utf8[]"` (auto-detect max length).

* Compressible: **YES**
* Handles `NA`: **NO**

```{r}
# UTF-8 auto-detected fixed length
h5_write(c("apple", "banana"), file, "strings/fixed_utf8")

# ASCII fixed length (1 byte)
h5_write(c("A", "B", "C"), file, "strings/fixed_ascii", as = "ascii[1]")
```

> **Technical Note:** `h5lite` uses `H5T_C_S1` for all strings, and `H5T_STR_NULLTERM` for all fixed length strings.


## Dates and Times (`POSIXt`)

R date-time objects (`POSIXct` / `POSIXlt`) are stored as **Strings** in ISO 8601 format (`YYYY-MM-DDTHH:MM:SSZ`). This ensures maximum portability with other languages and HDF5 tools that do not share R's specific epoch-based integer representation.

```{r}
now <- Sys.time()
h5_write(now, file, "datetime/iso8601")
```

## Complex Data

R complex numbers are written using the new complex floating-point type introduced in HDF5 2.0.0 (`H5T_COMPLEX_IEEE_F64LE`).

**Compatibility Warning:** This data type for complex numbers is a feature specific to HDF5 version 2.0+. Datasets written with this type generally cannot be read by HDF5 readers built against older versions of the library (e.g., HDF5 1.10 or 1.12). Ensure that any downstream tools or libraries used to read these files are updated to support HDF5 2.0 standards.

```{r}
comp <- c(1+2i, 3+4i)
h5_write(comp, file, "complex_data")
```

## Raw Data

Raw vectors (bytes) are stored as HDF5 `OPAQUE` types. This is ideal for storing binary blobs, images, or serialized objects where you need to preserve the exact byte sequence without interpretation.

```{r}
raw_vec <- as.raw(c(0x01, 0xFF, 0x1A))
h5_write(raw_vec, file, "binary_blob")
```

## Factors

R Factors are stored as HDF5 `ENUM` types. This maps the integer codes to the factor levels (labels) efficiently within the file header, ensuring the labels are preserved without duplicating string data for every element.

```{r}
fac <- factor(c("low", "high", "medium", "low"))
h5_write(fac, file, "categorical")
```

## Lists

R lists are mapped to HDF5 **Groups**. Since lists are recursive containers, `h5lite` walks the list and creates a dataset (or subgroup) for every element found. You can use `as = c("element_name" = "skip")` to exclude specific items.

```{r}
my_list <- list(data = 1:100, meta = list(valid = TRUE))
h5_write(my_list, file, "types/list")
```

## Data Frames

Data Frames are stored as HDF5 **Compound** types (tables). This ensures that rows are kept together in memory. You can use the `as` argument to specify the type of individual columns.

*For a comprehensive guide, see `vignette('data-frames')`.*

```{r}
df <- data.frame(
  id = 1:5,
  score = c(10.5, 20.2, 15.0, 9.8, 30.1)
)

# 1. 'id' coerced to uint16
# 2. 'score' coerced to float32
h5_write(df, file, "types/dataframe", as = c(
  "id"    = "uint16",
  "score" = "float32"
))
```

## NULL

The `NULL` object in R is mapped to a dataset with a **NULL Dataspace** (`H5S_NULL`). This creates a dataset that exists in the file structure but contains no data elements and consumes no storage space.

```{r}
h5_write(NULL, file, "placeholders/empty_slot")
```

## Compression

HDF5 supports transparent data compression using the zlib (gzip) and szip algorithms. You can control the compression behavior using the `compress` argument.

* **`"gzip-5"`** (default): Standard zlib compression at level 5. Levels `"gzip-1"` through `"gzip-9"` are also supported. Safe and universally compatible.
* **`"szip-nn"`**: Szip with Nearest Neighbor coding. Best for continuous, correlated, or floating-point data (e.g., time series or smooth gradients).
* **`"szip-ec"`**: Szip with Entropy Coding. Best for uncorrelated, discrete, or categorical integer data.
* **`"none"`**: Disables compression entirely.

```{r}
# Maximum zlib compression
h5_write(rnorm(1000), file, "data/max", compress = "gzip-9")

# Szip Entropy Coding for discrete integer data
h5_write(sample(1:5, 1000, replace = TRUE), file, "data/szip", compress = "szip-ec")
```

### The Shuffle Filter

When `gzip` compression is enabled, `h5lite` automatically applies the HDF5 **Byte Shuffle Filter** before the data is compressed. The Shuffle Filter does not compress data itself; rather, it rearranges the byte stream to make it more compressible by zlib.

It works by separating the bytes of each value by their significance. For example, in a 4-byte integer array:

1.  All the 1st bytes (least significant) are grouped together.
2.  All the 2nd bytes are grouped together.
3.  And so on.

**Why this helps:**

* **Integers:** Small integers often have many zero-padding bytes. The shuffle filter groups these zeros into long runs, which zlib compresses extremely efficiently. This allows `int32` data to compress nearly as well as `int8` data if the values are small.
* **Doubles:** Floating point numbers often share the same exponent bytes if they are in a similar range. The shuffle filter groups these identical exponent bytes, creating repetitive patterns that zlib can compress.

```{r, include = FALSE}
unlink(file)
```

---
title: "Atomic Vectors"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Atomic Vectors}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

Atomic vectors are the fundamental data structure in R. They include **numeric** (integer and double), **logical**, **character**, **complex**, and **raw** vectors. This vignette explains how `h5lite` maps these R types to HDF5 datasets and provides guidance on controlling storage types and compression.

```{r setup}
library(h5lite)
file <- tempfile(fileext = ".h5")
```

## Basic Usage

Writing a vector to HDF5 is straightforward using `h5_write()`. The package automatically creates the necessary dataset and handles dimensions.

```{r}
# Write a numeric vector
vec <- c(1.5, 2.3, 4.2, 5.1)
h5_write(vec, file, "data/numeric_vector")

# Read it back
res <- h5_read(file, "data/numeric_vector")
print(res)
```

## Scalars vs. 1D Arrays

In R, a "scalar" is simply a vector of length 1. However, HDF5 distinguishes between a **Scalar Dataspace** (a single value with no dimensions) and a **Simple Dataspace** (an array) with dimensions `[1]`.

By default, `h5lite` treats length-1 vectors as 1D arrays to maintain consistency with R's vector behavior. To write a true HDF5 scalar, you must wrap the value in `I()`.

```{r}
# 1. Default: 1D Array (Length 1)
h5_write(42, file, "structure/array_1d")

# 2. Explicit Scalar: Wrapped in I()
h5_write(I(42), file, "structure/scalar")

h5_str(file, "structure")
```

*Note: When reading data back into R, both storage formats appear as standard R vectors of length 1.*

## Numeric and Logical Data

### Automatic Type Selection

`h5lite` attempts to map R types to the most efficient HDF5 equivalents automatically (`as = "auto"`).

1.  **Numeric:** `h5lite` analyzes the range of your data and picks the smallest fitting HDF5 type (e.g., `uint8`, `int16`, `int32`, `float64`).
2.  **Logicals:** `h5lite` maps these to `uint8` (0 or 1) in HDF5 to save space.

### Handling Missing Values (NA)

A key challenge in HDF5 is that standard integer and boolean types do not have a native representation for `NA` (missing values).

To ensure data safety, `h5lite` performs the following check:

* If an integer or logical vector contains `NA`, it is **automatically promoted to `float64`**.
* The `NA` values are stored as an `NaN` variant in the file.
* When read back, `h5_read()` restores them as `numeric` vectors with `NA`.

```{r}
# Integer vector with NO missing values -> Automatic optimal type (uint8)
h5_write(c(1L, 2L, 3L), file, "safe/ints")
h5_typeof(file, "safe/ints")

# Integer vector WITH missing values -> Promoted to float64
h5_write(c(1L, NA, 3L), file, "safe/ints_na")
h5_typeof(file, "safe/ints_na")
```

### Forcing Specific Types

If you know your data range fits into a smaller type (e.g., `int8`, `uint16`), you can use the `as` argument to force a specific storage type.

*Warning: If you force an integer type on data containing `NA` or values outside the integer type's range then `h5lite` will throw an error.*

```{r}
# Store small integers as 8-bit signed integers
h5_write(c(10, -5, 100), file, "small_ints", as = "int8")

# Store logicals as 8-bit unsigned integers
h5_write(c(TRUE, FALSE), file, "bools", as = "uint8")
```

## Character Vectors (Strings)

HDF5 supports two primary methods for storing strings: **Variable-Length** and **Fixed-Length**.

### Automatic Type Selection

By default (`as = "auto"`), `h5lite` chooses the most efficient string representation:

* If the vector contains `NA`, it uses **Variable-Length UTF-8** (which natively supports missing values).
* If there are no missing values and the strings are relatively short and consistent in length, it uses **Fixed-Length UTF-8** to allow for compression and faster access.

### Variable-Length

You can explicitly request variable-length storage using `as = "utf8"` or `as = "ascii"`.

* **Pros:** Most flexible; exact memory usage per string; supports `NA` (stored as NULL pointers).
* **Cons:** Cannot be compressed using standard HDF5 filters; slower to read/write for extreme dataset sizes.

```{r}
# Variable length strings (handles NA)
h5_write(c("apple", "banana", NA), file, "strings/var")
```

### Fixed-Length

You can force fixed-length storage using the syntax `[n]`, where `n` is the number of bytes.

* **Pros:** Fast; allows compression.
* **Cons:** Truncates strings longer than `n`; pads shorter strings; **does not support `NA`**.

```{r}
# Fixed length strings (10 bytes per string)
h5_write(c("A", "B", "C"), file, "strings/fixed", as = "ascii[10]")

# Auto-detect max length (converts to fixed length based on longest string)
h5_write(c("short", "longer", "longest"), file, "strings/auto_fixed", as = "ascii[]")
```

## Compression

Compression in HDF5 requires the dataset to be "chunked". `h5lite` handles chunking parameters automatically when you enable compression.

You can configure compression using the `compress` argument:

* `"gzip-5"` (default): Standard zlib compression at level 5. Levels `"gzip-1"` through `"gzip-9"` are also supported. Safe and universally compatible.
* `"szip-nn"`: Szip with Nearest Neighbor coding. Best for continuous, correlated, or floating-point data (e.g., time series or smooth gradients).
* `"szip-ec"`: Szip with Entropy Coding. Best for uncorrelated, discrete, or categorical integer data.
* `"none"`: Disables compression entirely.

```{r}
# Write a large vector with max zlib compression
x <- rep(rnorm(100), 100)
h5_write(x, file, "compressed_data", compress = "gzip-9")

# Write a smooth, correlated dataset using szip Nearest Neighbor
smooth_data <- sin(seq(0, 10, length.out = 1000))
h5_write(smooth_data, file, "szip_data", compress = "szip-nn")
```

## 64-bit Integers

R does not natively support 64-bit integers, but the `bit64` package provides an `integer64` class. `h5lite` supports reading and writing these types directly to HDF5 `int64`.

```{r}
if (requireNamespace("bit64", quietly = TRUE)) {
  val <- bit64::as.integer64(c("9223372036854775807", "-9223372036854775807"))
  
  h5_write(val, file, "huge_ints")
  h5_typeof(file, "huge_ints")
  
  in_val <- h5_read(file, "huge_ints", as = "bit64")
  print(class(in_val))
}
```

```{r, include=FALSE}
unlink(file)
```

---
title: "Matrices and Arrays"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Matrices and Arrays}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

HDF5 is an excellent format for storing large, multi-dimensional numerical arrays. `h5lite` simplifies the process of reading and writing matrices and arrays by handling the complex memory layout differences between R and HDF5 automatically.

This vignette covers writing matrices, preserving dimension names (`dimnames`), and understanding how `h5lite` manages dimension ordering.

```{r setup}
library(h5lite)
file <- tempfile(fileext = ".h5")
```

## Writing Matrices

In R, matrices are simply 2-dimensional arrays. You can write them directly using `h5_write()`. `h5lite` preserves the dimensions exactly as they appear in R.

```{r}
# Create a 3x4 matrix
mat <- matrix(1:12, nrow = 3, ncol = 4)

# Write to file
h5_write(mat, file, "linear_algebra/mat_a")

# Read back
mat_in <- h5_read(file, "linear_algebra/mat_a")

# Verify
all.equal(mat, mat_in)
```

## Writing N-Dimensional Arrays

The same logic applies to arrays with 3 or more dimensions.

```{r}
# Create a 3D array (e.g., spatial data over time: x, y, time)
vol <- array(runif(24), dim = c(4, 3, 2))

h5_write(vol, file, "spatial/volume")

# Check dimensions without reading the full data
h5_dim(file, "spatial/volume")
```

## Dimension Names (dimnames)

R objects often carry metadata in the form of `dimnames` (row names, column names, etc.). HDF5 does not have a native "row name" concept for numerical arrays, but it supports **Dimension Scales**.

`h5lite` automatically converts R `dimnames` into HDF5 Dimension Scales. This allows your row and column names to survive the round-trip to disk and back.

```{r}
# Create a matrix with row and column names
data <- matrix(rnorm(6), nrow = 2)
rownames(data) <- c("Sample_A", "Sample_B")
colnames(data) <- c("Gene_1", "Gene_2", "Gene_3")

h5_write(data, file, "genetics/expression")

# Read back
data_in <- h5_read(file, "genetics/expression")

print(data_in)
```

> **Technical Note:** In the HDF5 file, the names are stored as separate datasets (e.g., `_rownames`, `_colnames`) and linked to the main dataset using HDF5 Dimension Scale attributes.

## Dimension Ordering (Row-Major vs. Column-Major)

One of the most confusing aspects of HDF5 for R users is dimension ordering.

* **R** is **Column-Major**: The first dimension varies fastest.
* **HDF5** (and C/C++/Python) is **Row-Major**: The last dimension varies fastest.

### How h5lite handles it

To ensure that a `3x4` matrix in R looks like a `3x4` dataset in HDF5 tools (like `h5dump` or `HDFView`), `h5lite` **rearranges** the data during read/write operations.

1.  **Writing:** `h5lite` converts R's column-major memory layout to HDF5's row-major layout.
2.  **Reading:** `h5lite` converts the data back to column-major for R.

This ensures that **indexing is preserved**. `x[2, 1]` in R refers to the exact same value after reading it back from HDF5.

### Interoperability with Python

Because `h5lite` writes the data in C-order (Row-Major) to match the HDF5 specification, files created with `h5lite` are perfectly readable by Python (`h5py` or `pandas`).

* **R:** Shape is `(3, 4)`
* **Python:** Shape is `(3, 4)`

*Note: Some other R packages create HDF5 files by swapping the dimensions (writing a 3x4 matrix as 4x3) to avoid the cost of transposing data. `h5lite` prioritizes correctness and interoperability over raw write speed.*

## Compression and Chunking

Matrices and arrays benefit significantly from compression. When you enable compression, `h5lite` automatically "chunks" the dataset (breaks it into smaller tiles).

```{r}
# Large matrix of zeros (highly compressible)
sparse_mat <- matrix(0, nrow = 1000, ncol = 1000)
sparse_mat[1:10, 1:10] <- 1

# Write with default compression (zlib level 5)
h5_write(sparse_mat, file, "compressed/matrix")

# Write with high compression (zlib level 9)
h5_write(sparse_mat, file, "compressed/matrix_max", compress = "gzip-9")
```

## Partial I/O

`h5lite` is designed for simplicity and currently reads/writes full datasets at once. It does **not** support partial I/O (hyperslabs), such as reading only rows 1-10 of a 1,000,000 row matrix.

If you need to read specific subsets of data that are too large to fit in memory, you should consider using the `rhdf5` or `hdf5r` packages.

```{r, include=FALSE}
unlink(file)
```

---
title: "Get Started with h5lite"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Get Started with h5lite}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

The `h5lite` package provides a simplified, user-friendly interface for interacting with HDF5 files in R. While HDF5 is a complex hierarchical data format, `h5lite` is designed to feel familiar to R users by mapping HDF5 concepts directly to R's native data structures.

## HDF5 for R Users

If you are new to HDF5, the easiest way to understand it is through R analogues. HDF5 files function like a file system within a single file, containing "groups" (folders) and "datasets" (files).

| R Concept           | HDF5 Concept         | Description                                                       |
| :------------------ | :------------------- | :---------------------------------------------------------------- |
| **List**            | **Group**            | A container that holds other objects (datasets or other groups).  |
| **Vector / Matrix** | **Dataset**          | A multidimensional array of data (numeric, character, etc.).      |
| **Data Frame**      | **Compound Dataset** | A table where each column can have a different data type.         |
| **Attribute**       | **Attribute**        | Metadata attached to a specific object (e.g., units, timestamps). |

## Basic Usage

The package uses two primary functions: `h5_write()` to save data and `h5_read()` to load it.

### Writing Numeric Data

You can write standard R arrays and vectors directly to an HDF5 file. `h5lite` automatically handles dimensions.

```{r}
library(h5lite)
file <- tempfile(fileext = ".h5")

# 1. 1D Array (Vector)
vec <- c(1.5, 2.3, 4.2)
h5_write(vec, file, "examples/vector")

# 2. 2D Array (Matrix)
mat <- matrix(1:9, nrow = 3, ncol = 3)
h5_write(mat, file, "examples/matrix")

# 3. 3D Array
arr <- array(1:24, dim = c(4, 3, 2))
h5_write(arr, file, "examples/array_3d")

# 4. Scalar
# By default, R treats length-1 vectors as arrays. 
# Wrap in I() to write a true HDF5 scalar.
val <- I(42)
h5_write(val, file, "examples/scalar")
```

*Note: While `h5lite` supports preserving row and column names for matrices and vectors, these examples omit them for simplicity. See `vignette('attributes-in-depth')` for details on how dimension names are stored.*

### Reading Data

Data is read back into its native R format.

```{r}
x <- h5_read(file, "examples/matrix")
print(x)
```

## Data Type Mapping

`h5lite` automatically selects the appropriate HDF5 data type based on the content of your R objects.

| R Data Type    | HDF5 Equivalent  | Description                                   |
| :------------- | :--------------- | :-------------------------------------------- |
| **Numeric**    | *variable*       | Selects optimal type: `int8`, `float32`, etc. |
| **Logical**    | `H5T_STD_U8LE`   | Stored as 0 (FALSE) or 1 (TRUE) (`uint8`).    |
| **Character**  | `H5T_STRING`     | Variable or fixed-length UTF-8 strings.       |
| **Complex**    | `H5T_COMPLEX`    | Native HDF5 2.0+ complex numbers.             |
| **Raw**        | `H5T_OPAQUE`     | Raw bytes / binary data.                      |
| **Factor**     | `H5T_ENUM`       | Integer indices with label mapping.           |
| **integer64**  | `H5T_STD_I64LE`  | 64-bit signed integers via `bit64` package.   |
| **POSIXt**     | `H5T_STRING`     | ISO 8601 string (`YYYY-MM-DDTHH:MM:SSZ`).     |
| **List**       | `H5O_TYPE_GROUP` | Recursive container structure.                |
| **Data Frame** | `H5T_COMPOUND`   | Table of mixed types.                         |
| **NULL**       | `H5S_NULL`       | Creates a placeholder.                        |

You can use the `as` argument to explicitly set the HDF5 data type for numeric, logical, and character vectors. See `vignette('data-types')` for details.

## Complex Data Structures

### Lists as Groups
R lists are naturally hierarchical, making them perfect for creating HDF5 groups.

```{r}
my_list <- list(
  config = list(id = 1L, status = "active"),
  data   = runif(10)
)
# Creates a group "/experiment" containing "config" (group) and "data" (dataset)
h5_write(my_list, file, "experiment") 
```

### Data Frames as Compound Datasets
Data frames are written as native HDF5 compound datasets, allowing efficient storage of tabular data with mixed types.

```{r}
df <- data.frame(
  id = 1:5,
  val = c(1.1, 2.2, 3.3, 4.4, 5.5)
)
h5_write(df, file, "study_data")
```

For more details on these structures, including how to handle factors and nested lists, refer to `vignette('data-frames')`.

## Attributes

Attributes are small pieces of metadata attached to objects. `h5lite` writes R attributes (like `units` or `description`) as HDF5 attributes.

```{r}
# Write a dataset
h5_write(1:10, file, "measurements")

# Attach an attribute to it
h5_write("meters", file, "measurements", attr = "units")
```

See `vignette('attributes-in-depth')` for information on reading specific attributes and how special R attributes like `dimnames` are handled.

## File Inspection

You can inspect the contents of an HDF5 file without reading the data into memory using `h5_ls()` and `h5_str()`.

```{r}
# List contents
h5_ls(file)

# Print structure tree (like R's str())
h5_str(file)
```

For advanced file operations, including moving, deleting, and verifying objects, refer to `vignette('organization')`.

---
title: "Attributes In-Depth"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Attributes In-Depth}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

In HDF5, **attributes** are small pieces of metadata attached to groups or datasets. They are best used to store descriptive information: units, timestamps, descriptions, or experimental parameters—separately from the main data array.

This vignette covers how to write, read, and manage these attributes using `h5lite`, as well as important limitations regarding their structure.

```{r setup}
library(h5lite)
file <- tempfile(fileext = ".h5")
```

## Writing Attributes

There are two ways to write attributes in `h5lite`: explicitly (targeting an object) or implicitly (saving R attributes).

### 1. Explicit Writing

You can write an attribute to any existing group or dataset using the `attr` argument in `h5_write()`. This is useful for adding metadata after the data has been saved.

```{r}
# First, write a dataset
h5_write(1:10, file, "measurements/temperature")

# Now, attach attributes to it
h5_write(I("Celsius"),    file, "measurements/temperature", attr = "units")
h5_write(I("2023-10-27"), file, "measurements/temperature", attr = "date")
h5_write(I(0.1),          file, "measurements/temperature", attr = "precision")
```

*Note: If the attribute already exists, it will be overwritten.*

### 2. Implicit Writing (R Attributes)

`h5lite` automatically preserves custom R attributes attached to your objects. When you write an R object, any attributes (except for standard internal ones like `dim`, `names`, or `class`) are written as HDF5 attributes.

```{r}
# Create a vector with custom R attributes
data <- rnorm(5)
attr(data, "description") <- I("Randomized control group")
attr(data, "valid")       <- I(TRUE)

# Write the object
h5_write(data, file, "experiment/control")

# Check the file - the attributes are there
h5_attr_names(file, "experiment/control")

h5_str(file)
```

## Reading Attributes

### 1. Accessing Specific Attributes

If you only need a specific piece of metadata without reading the full dataset, you can use `h5_read(..., attr = "name")`.

```{r}
# Read just the 'units' attribute
units <- h5_read(file, "measurements/temperature", attr = "units")
print(units)
```

### 2. Reading with the Dataset

When you read a dataset, `h5lite` automatically reads all attached attributes and re-attaches them to the resulting R object.

```{r}
# Read the full dataset
temps <- h5_read(file, "measurements/temperature")

# The attributes are available in R
attributes(temps)

str(temps)
```

## Managing Attributes

### Listing Attributes

Use `h5_attr_names()` to list the names of all attributes attached to a specific object.

```{r}
h5_attr_names(file, "measurements/temperature")
```

### Deleting Attributes

You can remove a specific attribute using `h5_delete()`.

```{r}
# Delete the 'precision' attribute
h5_delete(file, "measurements/temperature", attr = "precision")

# Verify removal
h5_attr_names(file, "measurements/temperature")
```

## Important Limitations

While attributes are powerful for storing metadata, they are fundamentally simpler structures than HDF5 Datasets. HDF5 enforces specific constraints that affect how `h5lite` can store complex R objects as attributes.

### 1. No Dimension Scales (Loss of Names)

HDF5 **Dimension Scales** (the mechanism `h5lite` uses to store `names`, `dimnames`, and `row.names`) can only be attached to **Datasets**. They cannot be attached to attributes.

This means if you write a named vector, matrix, or array as an attribute, **the names will be lost**.

```{r}
# A vector with names
named_vec <- c(a = 1, b = 2, c = 3)

# Write as a standard Dataset -> Names are preserved
h5_write(named_vec, file, "my_dataset")
h5_names(file, "my_dataset")

# Write as an Attribute -> Names are LOST
h5_write(named_vec, file, "measurements/temperature", attr = "meta_vec")
h5_names(file, "measurements/temperature", attr = "meta_vec")
```

**Exception: Data Frames**
There is one major exception: `data.frame` objects.

Because HDF5 stores data frames as **Compound Types**, the column names are baked into the type definition itself, not stored as side-loaded metadata. Therefore, **column names are preserved** even when writing a data frame as an attribute. However, `row.names` (which rely on dimension scales) will still be lost.

```{r}
# A data frame with metadata
df <- data.frame(
  id = 1:3, 
  status = c("ok", "fail", "ok")
)

# Write as attribute
h5_write(df, file, "measurements/temperature", attr = "log")

# Column names survive!
h5_names(file, "measurements/temperature", attr = "log")
```

### 2. No Attributes on Attributes (Nesting)

In HDF5, you cannot attach attributes to other attributes. This hierarchy is strictly one level deep: Groups/Datasets can have attributes, but attributes cannot.

Consequently, you cannot treat an attribute as a "Group" or folder to store other items. If you need a hierarchical structure for your metadata, you should create a Group (e.g., `/metadata`) and store your metadata as Datasets inside it, rather than attaching them as attributes to another object.

## Controlling Attribute Types

Attributes in HDF5 are typed just like datasets. `h5lite` allows you to control the storage type of attributes using the `as` argument in `h5_write()` or `h5_read()`.

To target an attribute specifically, prefix the name with `@` in the `as` vector.

### Customizing Storage Type

```{r}
# Write the temperature data again, but use a fixed length string for 'description'
h5_write(data, file, "experiment/control", as = c("@description" = "ascii[]"))

# Store an attribute as a `uint8` instead of the default `int32`
h5_write(I(42), file, "measurements/temperature", "sensor_id", as = "uint8")
```

### Customizing Read Type

You can also coerce attributes when reading them.

```{r}
# Force the 'valid' attribute to be read as logical, even if stored as integer
meta <- h5_read(file, "experiment/control", attr = "valid", as = "logical")
```

## Special Note: Dimensions

You might notice that standard R attributes like `dim` are not visible in `h5_attr_names()`.

This is because `h5lite` handles structural attributes implicitly. The dimensions of the attribute data itself are stored in the HDF5 Dataspace, not as a separate attribute. `h5lite` automatically restores the `dim` attribute on the R object when reading, ensuring matrices and arrays retain their shape.

```{r, include=FALSE}
unlink(file)
```

---
title: "Object-Oriented Interface"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Object-Oriented Interface}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

The `h5lite` package provides a convenient object-oriented interface for interacting with HDF5 files. While the standard functions (e.g., `h5_read`, `h5_write`) are excellent for stateless, atomic operations, the object-oriented handle is often more natural when performing multiple operations on a single file or when navigating deep group structures.

## Basic Usage

To use this interface, create a file handle using `h5_open()`. This returns an `h5` object (specifically, an environment) that maintains a reference to your file.

```{r setup}
library(h5lite)

# Create a temporary file for this example
file <- tempfile(fileext = ".h5")

# Open the handle
h5 <- h5_open(file)

# The print method shows the file path and current internal working directory
print(h5)
```

Once the handle is open, you can access standard `h5lite` functions as methods of this object. The primary benefit is that you no longer need to pass the `file` argument; it is handled implicitly.

```{r write_read}
# Write data using the handle
h5$write(1:10, "dataset1")
h5$write(matrix(1:9, 3, 3), "matrix_data")

# List contents
h5$ls()

# Read data back
my_data <- h5$read("dataset1")
print(my_data)
```

Almost all standard functions map directly to methods by dropping the `h5_` prefix:

* `h5_write()` $\rightarrow$ `h5$write()`
* `h5_read()` $\rightarrow$ `h5$read()`
* `h5_ls()` $\rightarrow$ `h5$ls()`
* `h5_attr_names()` $\rightarrow$ `h5$attr_names()`
* And so on (see **Method Reference** below).

## Navigation and Working Directories

A unique feature of the `h5` handle is the ability to maintain a stateful "Working Directory" (WD) inside the HDF5 file. This functions similarly to `setwd()` in R or `cd` in a shell.

By default, the WD is the root (`/`). You can change it using `$cd()` and check it using `$pwd()`.

```{r navigation}
# Create a group structure
h5$create_group("simulations")

# Navigate into the group
h5$cd("simulations")
h5$pwd()

# Write data using relative paths (writes to /simulations/run1)
h5$write(rnorm(10), "run1")

# Verify the location
h5$ls()
```

Absolute paths (paths starting with `/`) always override the current working directory:

```{r absolute_paths}
# Writes to the root, ignoring the fact that we are in /simulations
h5$write(100, "/root_dataset")
```

You can also use `..` to navigate up the hierarchy:

```{r relative_nav}
h5$cd("..")
h5$pwd() # Now back at "/"
```

## Reference Semantics

The object returned by `h5_open` is an **R environment**. This means it follows "pass-by-reference" semantics, unlike most R objects which are "pass-by-value".

If you assign the handle to a new variable, you are creating a new reference to the *same* handle. Modifying one will modify the other.

```{r ref_semantics}
h5_alias <- h5

# Change directory in the alias
h5_alias$cd("simulations")

# The original handle is also updated
h5$pwd()
```

## Closing the Handle

When you are finished, you can close the handle using `$close()`. 

**Note:** `h5lite` does not keep the actual HDF5 file lock open persistently (it opens and closes the file for every read/write operation to ensure safety). Therefore, `$close()` is a logical operation: it clears the internal file path and working directory from the handle, preventing further methods from being called on it.

```{r close}
h5$close()

# Further attempts to use the handle will result in an error:
# h5$ls() 
# Error: This h5 file handle has been closed.
```

## Method Reference

The following methods are available on the `h5` object:

| Method                    | Description                                   |
| :------------------------ | :-------------------------------------------- |
| **I/O**                   |                                               |
| `$read(name, ...)`        | Read a dataset or attribute.                  |
| `$write(data, name, ...)` | Write a dataset or attribute.                 |
| **Navigation**            |                                               |
| `$cd(group)`              | Change the internal working directory.        |
| `$pwd()`                  | Print the current internal working directory. |
| **Structure**             |                                               |
| `$ls(name, ...)`          | List contents of a group.                     |
| `$create_group(name)`     | Create a new group.                           |
| `$delete(name)`           | Delete a group or dataset.                    |
| `$move(from, to)`         | Move/Rename an object.                        |
| `$names(name)`            | Get names of objects in a group.              |
| **Inspection**            |                                               |
| `$exists(name)`           | Check if an object exists.                    |
| `$dim(name)`              | Get dimensions of a dataset.                  |
| `$typeof(name)`           | Get the HDF5 type of an object.               |
| `$class(name)`            | Get the class (dataset/group) of an object.   |
| `$is_dataset(name)`       | Boolean check for dataset.                    |
| `$is_group(name)`         | Boolean check for group.                      |
| `$attr_names(name)`       | List attribute names.                         |
| `$str(name)`              | Display structure of an object.               |

---
title: "Data Frames"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Data Frames}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

Data frames are the workhorse of data analysis in R. In HDF5, data frames are stored as **Compound Datasets**. This allows different columns to have different data types (e.g., integer, float, string) within the same dataset, much like a SQL table.

This vignette explains how `h5lite` handles data frames, including row names, factors, and missing values.

```{r setup}
library(h5lite)
file <- tempfile(fileext = ".h5")
```

## Basic Usage

Writing a data frame is as simple as writing any other object. `h5lite` automatically maps each column to its appropriate HDF5 type.

```{r}
# Create a standard data frame
df <- data.frame(
  id = 1:5,
  group = c("A", "A", "B", "B", "C"),
  score = c(10.5, 9.2, 8.4, 7.1, 6.0),
  passed = c(TRUE, TRUE, TRUE, FALSE, FALSE),
  stringsAsFactors = FALSE
)

# Write to HDF5
h5_write(df, file, "study_data/results")

# Fetch the column names
h5_names(file, "study_data/results")

# Read back
df_in <- h5_read(file, "study_data/results")

head(df_in)
```

## Customizing Column Types

You can use the `as` argument to control the storage type for specific columns. This is passed as a named vector where the names correspond to the column names.

This is particularly useful for optimizing storage (e.g., saving space by storing small integers as `int8` or single characters as `ascii[1]`).

```{r}
df_small <- data.frame(
  id   = 1:10,
  code = rep("A", 10)
)

# Force 'id' to be uint16 and 'code' to be an ascii string
h5_write(df_small, file, "custom_df", 
         as = c(id = "uint16", code = "ascii[]"))
```

## Row Names

Standard HDF5 Compound Datasets do not have a concept of "row names". However, `h5lite` preserves them using **Dimension Scales**.

When you write a data frame with row names, `h5lite` creates a separate dataset (usually named `_rownames`) and links it to the main table. When reading, `h5lite` automatically restores these as the `row.names` of the data frame.

```{r}
mtcars_subset <- head(mtcars, 3)

h5_write(mtcars_subset, file, "cars")

h5_str(file)

# Read back
result <- h5_read(file, "cars")
print(row.names(result))
```


```{r, include=FALSE}
unlink(file)
```

---
title: "Organization"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Organization}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

HDF5 files are best understood as "file systems within a file." Just as your computer has folders and files, an HDF5 file has **Groups** (folders) and **Datasets** (files). This hierarchical structure allows you to organize complex experimental data, metadata, and configuration settings into a single, self-describing package.

This vignette explains how to create, manage, and modify this structure using `h5lite`.

```{r setup}
library(h5lite)
file <- tempfile(fileext = ".h5")
```

## The Hierarchical Model

HDF5 uses POSIX-style paths (like Linux or macOS) to identify objects. The root of the file is `/`.

* `/` : The Root Group
* `/experiment_1` : A Group (folder)
* `/experiment_1/data` : A Dataset (file) inside the group

## Creating Groups

### Implicit Creation (Recommended)

In most cases, you do not need to create groups manually. When you write a dataset to a path like `"data/experiment/run1"`, `h5lite` automatically creates the parent groups `"data"` and `"data/experiment"` if they do not exist.

### Explicit Creation

If you need to create an empty group structure (perhaps to add attributes to it), you can use `h5_create_group()`. This function works like `mkdir -p`: it creates all necessary parent groups.

```{r}
# Create a deep hierarchy
h5_create_group(file, "project_A/simulation/run_01")

# Verify
h5_str(file)
```

## Using Lists as Groups

The most powerful way to organize data in `h5lite` is by mapping R **lists** to HDF5 **groups**.

When you pass a named list to `h5_write()`, `h5lite` recursively writes the list structure to the file.

* **Named Lists** become **Groups**.
* **Atomic Vectors/Matrices** inside the list become **Datasets**.

This allows you to organize your entire data structure in R and save it to disk in one command.

```{r}
# Define a complex structure in R
experiment_data <- list(
  metadata = list(
    id         = I(101),
    technician = I("Dr. Smith"),
    timestamp  = I("2023-10-27")
  ),
  measurements = list(
    raw         = runif(10),
    calibration = c(0.1, 0.9)
  ),
  status = I("complete")
)

# Write the entire structure to a group named "exp_101"
h5_write(experiment_data, file, "exp_101")
```

## Inspecting Structure

You can visualize the organization of your file using `h5_ls()` and `h5_str()`.

* `h5_ls()`: Returns a character vector of names. Useful for programmatic checks.
* `h5_str()`: Prints a tree diagram. Useful for interactive exploration.

```{r}
# List all objects recursively
h5_ls(file)

# Visualize the tree
h5_str(file)
```

## Moving and Renaming

Data organization often changes. You can rename objects or move them to different groups using `h5_move()`.

This operation is metadata-only, meaning it is extremely fast even for large datasets, as the data itself is not rewritten.

```{r}
# Rename 'exp_101' to 'archive_101'
h5_move(file, "exp_101", "archive_101")

# Move 'project_A' inside 'archive_101'
h5_move(file, "project_A", "archive_101/project_A")

h5_ls(file)
```

## Deleting Objects

You can remove groups or datasets using `h5_delete()`.

* Deleting a dataset removes the data.
* Deleting a group removes the group **and all of its children** (recursively).
* The file size does not change, but the freed space can be reused.

```{r}
# Delete the entire archive group
h5_delete(file, "archive_101")

# The file is now empty (except for the root)
h5_ls(file)
```

```{r, include=FALSE}
unlink(file)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/organize.r
\name{h5_delete}
\alias{h5_delete}
\title{Delete an HDF5 Object or Attribute}
\usage{
h5_delete(file, name, attr = NULL, warn = TRUE)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{The full path of the object to delete (e.g., \code{"/data/dset"} or \code{"/groups/g1"}).}

\item{attr}{The name of the attribute to delete.
\itemize{
\item If \code{NULL} (the default), the object specified by \code{name} is deleted.
\item If a string is provided, the attribute named \code{attr} is removed from the object \code{name}.
}}

\item{warn}{Emit a warning if the name/attr does not exist. Default: \code{TRUE}}
}
\value{
Invisibly returns \code{NULL}. This function is called for its side effects.
}
\description{
Deletes an object (dataset or group) or an attribute from an HDF5 file.
If the object or attribute does not exist, a warning is issued and the function returns
successfully (no error is raised).
}
\examples{
file <- tempfile(fileext = ".h5")
h5_create_file(file)

# Create some data and attributes
h5_write(matrix(1:10, 2, 5), file, "matrix")
h5_write("A note", file, "matrix", attr = "note")

# Review the file structure
h5_str(file)

# Delete the attribute
h5_delete(file, "matrix", attr = "note")

# Review the file structure
h5_str(file)

# Delete the dataset
h5_delete(file, "matrix")

# Review the file structure
h5_str(file)

# Cleaning up
unlink(file)
}
\seealso{
\code{\link[=h5_create_group]{h5_create_group()}}, \code{\link[=h5_move]{h5_move()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inspect.r
\name{print.inspect}
\alias{print.inspect}
\title{Print method for HDF5 inspect objects}
\usage{
\method{print}{inspect}(x, ...)
}
\arguments{
\item{x}{An object of class \code{inspect}.}

\item{...}{Further arguments passed to or from other methods.}
}
\description{
Print method for HDF5 inspect objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.r
\name{h5_length}
\alias{h5_length}
\title{Get the Total Length of an HDF5 Object or Attribute}
\usage{
h5_length(file, name, attr = NULL)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{The full path of the object (group or dataset).}

\item{attr}{The name of an attribute to check. If provided, the length of the attribute is returned.}
}
\value{
An numeric scalar representing the total length (number of elements).
}
\description{
Behaves like \code{length()} for R objects.
\itemize{
\item For \strong{Compound Datasets} (data.frames), this is the number of columns.
\item For \strong{Datasets} and \strong{Attributes}, this is the product of all dimensions (total number of elements).
\item For \strong{Groups}, this is the number of objects directly contained in the group.
\item Scalar datasets or attributes return 1.
}
}
\examples{
file <- tempfile(fileext = ".h5")

h5_write(1:100, file, "my_vec")
h5_length(file, "my_vec") # 100

h5_write(mtcars, file, "my_df")
h5_length(file, "my_df") # 11 (ncol(mtcars))

h5_write(as.matrix(mtcars), file, "my_mtx")
h5_length(file, "my_mtx") # 352 (prod(dim(mtcars)))

h5_length(file, "/") # 3

unlink(file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ls.r
\name{h5_str}
\alias{h5_str}
\title{Display the Structure of an HDF5 Object}
\usage{
h5_str(file, name = "/", attrs = TRUE, members = TRUE, markup = interactive())
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{The name of the group or dataset to display. Defaults to the root
group "/".}

\item{attrs}{Set to \code{FALSE} to hide attributes. The default (\code{TRUE}) shows
attributes prefixed with \code{@}.}

\item{members}{Set to \code{FALSE} to hide compound dataset members. The default
(\code{TRUE}) shows members prefixed with \code{$}.}

\item{markup}{Set to \code{FALSE} to remove colors and italics from the output.}
}
\value{
This function is called for its side-effect of printing to the
console and returns \code{NULL} invisibly.
}
\description{
Recursively prints a summary of an HDF5 group or dataset, similar to
the structure of \code{h5ls -r}. It displays the nested structure, object types,
dimensions, and attributes.
}
\details{
This function provides a quick and convenient way to inspect the contents of
an HDF5 file. It performs a recursive traversal of the file from the C-level
and prints a formatted summary to the R console.

This function \strong{does not read any data} into R. It only inspects the
metadata (names, types, dimensions) of the objects in the file, making it
fast and memory-safe for arbitrarily large files.
}
\examples{
file <- tempfile(fileext = ".h5")
h5_write(list(x = 1:10, y = matrix(1:9, 3, 3)), file, "group")
h5_write("metadata", file, "group", attr = "info")

# Print structure
h5_str(file)

unlink(file)
}
\seealso{
\code{\link[=h5_ls]{h5_ls()}}, \code{\link[=h5_attr_names]{h5_attr_names()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.r
\name{h5_typeof}
\alias{h5_typeof}
\title{Get HDF5 Storage Type of an Object or Attribute}
\usage{
h5_typeof(file, name, attr = NULL)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{Name of the dataset or object.}

\item{attr}{The name of an attribute to check. If \code{NULL} (default), the function
returns the type of the object itself.}
}
\value{
A character string representing the HDF5 storage type
(e.g., "float32", "uint32", "ascii[10]", "compound[2]").
}
\description{
Returns the low-level HDF5 storage type of a dataset or an attribute
(e.g., "int8", "float64", "utf8", "ascii[10]"). This allows inspecting the
file storage type before reading the data into R.
}
\examples{
file <- tempfile(fileext = ".h5")

h5_write(1L, file, "int32_val", as = "int32")
h5_typeof(file, "int32_val") # "int32"

h5_write(mtcars, file, "mtcars")
h5_typeof(file, "mtcars") # "compound[11]"

h5_write(c("a", "b", "c"), file, "strings")
h5_typeof(file, "strings") # "utf8[1]"

unlink(file)
}
\seealso{
\code{\link[=h5_class]{h5_class()}}, \code{\link[=h5_exists]{h5_exists()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h5lite-package.r
\docType{package}
\name{h5lite-package}
\alias{h5lite}
\alias{h5lite-package}
\title{h5lite: A Simple and Lightweight HDF5 Interface}
\description{
The \code{h5lite} package provides a simple, lightweight, and user-friendly
interface for reading and writing HDF5 files. It is designed for R users
who want to save and load common R objects (vectors, matrices, arrays,
factors, and data.frames) to an HDF5 file without needing to understand
the low-level details of the HDF5 C API.
}
\section{Key Features}{

\itemize{
\item \strong{Simple API:} Use familiar functions like \code{\link[=h5_read]{h5_read()}} and \code{\link[=h5_write]{h5_write()}}.
\item \strong{Automatic Handling:} Dimensions, data types, and group creation are handled automatically.
\item \strong{Safe by Default:} Auto-selects a safe R data type for numeric data to prevent overflow.
\item \strong{Easy Installation:} The required HDF5 library is bundled with the package.
}
}

\seealso{
Useful links:
\itemize{
\item \url{https://cmmr.github.io/h5lite/}
\item Report bugs at \url{https://github.com/cmmr/h5lite/issues}
}

Key functions: \code{\link[=h5_read]{h5_read()}}, \code{\link[=h5_write]{h5_write()}}, \code{\link[=h5_ls]{h5_ls()}}, \code{\link[=h5_str]{h5_str()}}
}
\author{
\strong{Maintainer}: Daniel P. Smith \email{dansmith01@gmail.com} (\href{https://orcid.org/0000-0002-2479-2044}{ORCID})

Other contributors:
\itemize{
  \item Alkek Center for Metagenomics and Microbiome Research [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.r
\name{h5_class}
\alias{h5_class}
\title{Get R Class of an HDF5 Object or Attribute}
\usage{
h5_class(file, name, attr = NULL)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{The full path of the object (group or dataset) to check.}

\item{attr}{The name of an attribute to check. If \code{NULL} (default), the function
checks the class of the object itself.}
}
\value{
A character string representing the R class (e.g., \code{"integer"}, \code{"numeric"},
\code{"complex"}, \code{"character"}, \code{"factor"}, \code{"raw"}, \code{"list"}, \code{"NULL"}).
Returns \code{NA_character_} for HDF5 types that \code{h5lite} cannot read.
}
\description{
Inspects an HDF5 object (or an attribute attached to it) and returns the R class
that \code{h5_read()} would produce.
}
\details{
This function determines the resulting R class by inspecting the storage metadata.

\itemize{
\item \strong{Group} \eqn{\rightarrow} \code{"list"}
\item \strong{Integer} \eqn{\rightarrow} \code{"numeric"}
\item \strong{Floating Point} \eqn{\rightarrow} \code{"numeric"}
\item \strong{String} \eqn{\rightarrow} \code{"character"}
\item \strong{Complex} \eqn{\rightarrow} \code{"complex"}
\item \strong{Enum} \eqn{\rightarrow} \code{"factor"}
\item \strong{Opaque} \eqn{\rightarrow} \code{"raw"}
\item \strong{Compound} \eqn{\rightarrow} \code{"data.frame"}
\item \strong{Null} \eqn{\rightarrow} \code{"NULL"}
}
}
\examples{
file <- tempfile(fileext = ".h5")

h5_write(1:10, file, "my_ints", as = "int32")
h5_class(file, "my_ints") # "numeric"

h5_write(mtcars, file, "mtcars")
h5_class(file, "mtcars") # "data.frame"

h5_write(c("a", "b", "c"), file, "strings")
h5_class(file, "strings") # "character"

h5_write(c(1, 2, 3), file, "my_floats", as = "float64")
h5_class(file, "my_floats") # "numeric"

unlink(file)
}
\seealso{
\code{\link[=h5_typeof]{h5_typeof()}}, \code{\link[=h5_read]{h5_read()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inspect.r
\name{h5_inspect}
\alias{h5_inspect}
\title{Inspect HDF5 Dataset Creation Properties}
\usage{
h5_inspect(file, name)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{The full path of the dataset to inspect.}
}
\value{
An object of class \code{inspect} (a named list) containing:
\item{layout}{A string indicating storage layout (e.g., "chunked", "contiguous").}
\item{chunk_dims}{A numeric vector of chunk dimensions, or \code{NULL} if not chunked.}
\item{filters}{A list describing each filter applied.}
}
\description{
Retrieves the Dataset Creation Property List (DCPL) details including storage
layout, chunk dimensions, and a detailed list of all applied filters.
}
\examples{
file <- tempfile(fileext = ".h5")

compress <- h5_compression('lz4-9', int_packing = TRUE, checksum = TRUE)
h5_write(matrix(5001:5100, 10, 10), file, "packed_mtx", compress = compress)
h5_inspect(file, "packed_mtx")

mtx <- matrix(rnorm(1000), 100, 10)
h5_write(mtx, file, "float_mtx", compress = 'blosc2-zfp-prec-3')
res <- h5_inspect(file, "float_mtx")
print(res)

# Print the raw cd_values for blosc2
dput(res$filters[[1]]$cd_values)

unlink(file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ls.r
\name{h5_ls}
\alias{h5_ls}
\title{List HDF5 Objects}
\usage{
h5_ls(file, name = "/", recursive = TRUE, full.names = FALSE, scales = FALSE)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{The group path to start listing from. Defaults to the root group (\code{/}).}

\item{recursive}{If \code{TRUE} (default), lists all objects found recursively
under \code{name}. If \code{FALSE}, lists only the immediate children.}

\item{full.names}{If \code{TRUE}, the full paths from the file's root are
returned. If \code{FALSE} (the default), names are relative to \code{name}.}

\item{scales}{If \code{TRUE}, also returns datasets that are dimensions scales for
other datasets.}
}
\value{
A character vector of object names. If \code{name} is \code{/} (the default),
the paths are relative to the root of the file. If \code{name} is another group,
the paths are relative to that group (unless \code{full.names = TRUE}).
}
\description{
Lists the names of objects (datasets and groups) within an HDF5 file or group.
}
\examples{
file <- tempfile(fileext = ".h5")
h5_create_group(file, "foo/bar")
h5_write(1:5, file, "foo/data")

# List everything recursively
h5_ls(file)

# List only top-level objects
h5_ls(file, recursive = FALSE)

# List relative to a sub-group
h5_ls(file, "foo")

unlink(file)
}
\seealso{
\code{\link[=h5_attr_names]{h5_attr_names()}}, \code{\link[=h5_str]{h5_str()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/open.r
\name{h5_open}
\alias{h5_open}
\title{Create an HDF5 File Handle}
\usage{
h5_open(file)
}
\arguments{
\item{file}{Path to the HDF5 file. The file will be created if it does not
exist.}
}
\value{
An object of class \code{h5} with methods for interacting with the file.
}
\description{
Creates a file handle that provides a convenient, object-oriented interface
for interacting with and navigating a specific HDF5 file.
}
\details{
This function returns a special \code{h5} object that wraps the standard \code{h5lite}
functions. The primary benefit is that the \code{file} argument is pre-filled,
allowing for more concise and readable code when performing multiple
operations on the same file.

For example, instead of writing:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{h5_write(1:10, file, "dset1")
h5_write(2:20, file, "dset2")
h5_ls(file)
}\if{html}{\out{</div>}}

You can create a handle and use its methods. Note that the \code{file} argument
is omitted from the method calls:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{h5 <- h5_open("my_file.h5")
h5$write(1:10, "dset1")
h5$write(2:20, "dset2")
h5$ls()
h5$close()
}\if{html}{\out{</div>}}
}
\section{Pass-by-Reference Behavior}{

Unlike most R objects, the \code{h5} handle is an \strong{environment}. This means it
is passed by reference. If you assign it to another variable (e.g.,
\code{h5_alias <- h5}), both variables point to the \emph{same} handle. Modifying one
(e.g., by calling \code{h5_alias$close()}) will also affect the other.
}

\section{Interacting with the HDF5 File}{

The \code{h5} object provides several ways to interact with the HDF5 file:

\subsection{Standard \code{h5lite} Functions as Methods}{
Most \code{h5lite} functions (e.g., \code{h5_read}, \code{h5_write}, \code{h5_ls}) are
available as methods on the \code{h5} object, without the \code{h5_} prefix.

For example, \code{h5$write(data, "dset")} is equivalent to
\code{h5_write(data, file, "dset")}.

The available methods are: \code{attr_names}, \code{cd}, \code{class}, \code{close},
\code{create_group}, \code{delete}, \code{dim}, \code{exists}, \code{is_dataset}, \code{is_group},
\code{length}, \code{ls}, \code{move}, \code{names}, \code{pwd}, \code{read}, \code{str}, \code{typeof}, \code{write}.
}

\subsection{Navigation (\verb{$cd()}, \verb{$pwd()})}{
The handle maintains an internal working directory to simplify
path management.
\itemize{
\item{\code{h5$cd(group)}: Changes the handle's internal working directory.
This is a stateful, pass-by-reference operation. It understands absolute
paths (e.g., \code{"/new/path"}) and relative navigation (e.g., \code{"../other"}).
The target group does not need to exist.
}
\item{\code{h5$pwd()}: Returns the current working directory.}
}
When you call a method like \code{h5$read("dset")}, the handle automatically
prepends the current working directory to any relative path. If you provide
an absolute path (e.g., \code{h5$read("/path/to/dset")}), the working directory
is ignored.
}

\subsection{Closing the Handle (\verb{$close()})}{
The \code{h5lite} package does not keep files persistently open. Each operation
opens, modifies, and closes the file. Therefore, the \code{h5$close()} method
does not perform any action on the HDF5 file itself.

Its purpose is to invalidate the handle, preventing any further operations
from being called. After \code{h5$close()} is called, any subsequent method
call (e.g., \code{h5$ls()}) will throw an error.
}
}

\examples{
file <- tempfile(fileext = ".h5")

# Open the handle
h5 <- h5_open(file)

# Write data (note: 'data' is the first argument, 'file' is implicit)
h5$write(1:5, "vector")
h5$write(matrix(1:9, 3, 3), "matrix")

# Create a group and navigate to it
h5$create_group("simulations")
h5$cd("simulations")
print(h5$pwd()) # "/simulations"

# Write data relative to the current working directory
h5$write(rnorm(10), "run1") # Writes to /simulations/run1

# Read data
dat <- h5$read("run1")

# List contents of current WD
h5$ls()

# Close the handle
h5$close()
unlink(file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/organize.r
\name{h5_create_group}
\alias{h5_create_group}
\title{Create an HDF5 Group}
\usage{
h5_create_group(file, name)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{The full path of the group to create (e.g., "/g1/g2").}
}
\value{
Invisibly returns \code{NULL}. This function is called for its side effects.
}
\description{
Explicitly creates a new group (or nested groups) in an HDF5 file.
This is useful for creating an empty group structure.
}
\examples{
file <- tempfile(fileext = ".h5")
h5_create_file(file)

# Create a nested group structure
h5_create_group(file, "/data/experiment/run1")
h5_ls(file)

unlink(file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.r
\name{h5_dim}
\alias{h5_dim}
\title{Get Dimensions of an HDF5 Object or Attribute}
\usage{
h5_dim(file, name, attr = NULL)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{Name of the dataset or object.}

\item{attr}{The name of an attribute to check. If \code{NULL} (default), the function
returns the dimensions of the object itself.}
}
\value{
An numeric vector of dimensions, or \code{numeric(0)} for scalars.
}
\description{
Returns the dimensions of a dataset or an attribute as an integer vector.
These dimensions match the R-style (column-major) interpretation.
}
\examples{
file <- tempfile(fileext = ".h5")

h5_write(matrix(1:10, 2, 5), file, "matrix")
h5_dim(file, "matrix") # 2 5

h5_write(mtcars, file, "mtcars")
h5_dim(file, "mtcars") # 32 11

h5_write(I(TRUE), file, "my_bool")
h5_dim(file, "my_bool") # numeric(0)

h5_write(1:10, file, "my_ints")
h5_dim(file, "my_ints") # 10

unlink(file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.r
\name{h5_names}
\alias{h5_names}
\title{Get Names of an HDF5 Object}
\usage{
h5_names(file, name = "/", attr = NULL)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{The full path of the object.}

\item{attr}{The name of an attribute. If provided, returns the names associated with the attribute
(e.g., field names if the attribute is a compound type). (Default: \code{NULL})}
}
\value{
A character vector of names, or \code{NULL} if the object has no names.
}
\description{
Returns the names of the object.
\itemize{
\item For \strong{Groups}, it returns the names of the objects contained in the group (similar to \code{ls()}).
\item For \strong{Compound Datasets} (data.frames), it returns the column names.
\item For other \strong{Datasets}, it looks for a dimension scale and returns it if found.
}
}
\examples{
file <- tempfile(fileext = ".h5")

h5_write(data.frame(x=1, y=2), file, "df")
h5_names(file, "df") # "x" "y"

x <- 1:5
names(x) <- letters[1:5]
h5_write(x, file, "x")
h5_names(file, "x") # "a" "b" "c" "d" "e"

h5_write(mtcars[,c("mpg", "hp")], file, "dset")
h5_names(file, "dset") # "mpg" "hp"

unlink(file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.r
\name{h5_exists}
\alias{h5_exists}
\title{Check if an HDF5 File, Object, or Attribute Exists}
\usage{
h5_exists(file, name = "/", attr = NULL, assert = FALSE)
}
\arguments{
\item{file}{Path to the file.}

\item{name}{The full path of the object to check (e.g., \code{"/data/matrix"}).
Defaults to \code{"/"} to test file validity.}

\item{attr}{The name of an attribute to check. If provided, the function tests
for the existence of this attribute on \code{name}.}

\item{assert}{Logical. If \code{TRUE} and the target does not exist, the function
will stop with an informative error message instead of returning \code{FALSE}.
Defaults to \code{FALSE}.}
}
\value{
A logical value: \code{TRUE} if the target exists and is valid, \code{FALSE} otherwise.
}
\description{
Safely checks if a file, an object within a file, or an attribute on an object exists.
}
\details{
This function provides a robust, error-free way to test for existence.

\itemize{
\item \strong{Testing for a File:} If \code{name} is \code{/} and \code{attr} is \code{NULL},
the function checks if \code{file} is a valid, readable HDF5 file.

\item \strong{Testing for an Object:} If \code{name} is a path (e.g., \verb{/data/matrix})
and \code{attr} is \code{NULL}, the function checks if the specific object exists.

\item \strong{Testing for an Attribute:} If \code{attr} is provided, the function checks
if that attribute exists on the specified object \code{name}.
}
}
\examples{
file <- tempfile(fileext = ".h5")

h5_exists(file) # FALSE

h5_create_file(file)
h5_exists(file) # TRUE

h5_exists(file, "missing_object") # FALSE

h5_write(1:10, file, "my_ints")
h5_exists(file, "my_ints") # TRUE

h5_exists(file, "my_ints", "missing_attr") # FALSE

h5_write(1:10, file, "my_ints", attr = "my_attr")
h5_exists(file, "my_ints", "my_attr") # TRUE

unlink(file)
}
\seealso{
\code{\link[=h5_is_group]{h5_is_group()}}, \code{\link[=h5_is_dataset]{h5_is_dataset()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.r
\name{h5_read}
\alias{h5_read}
\title{Read an HDF5 Object or Attribute}
\usage{
h5_read(file, name = "/", attr = NULL, as = "auto", start = NULL, count = NULL)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{The full path of the dataset or group to read (e.g., \code{"/data/matrix"}).}

\item{attr}{The name of an attribute to read.
\itemize{
\item If \code{NULL} (default), the function reads the object specified by \code{name} (and attaches its attributes to the result).
\item If provided (string), the function reads \emph{only} the specified attribute from \code{name}.
}}

\item{as}{The target R data type.
\itemize{
\item \strong{Global:} \code{"auto"} (default), \code{"integer"}, \code{"double"}, \code{"logical"}, \code{"bit64"}, \code{"null"}.
\item \strong{Specific:} A named vector mapping names or type classes to R types (see Section "Type Conversion").
}}

\item{start}{A numeric vector specifying the 1-based coordinate(s) for a partial read.
Most often, this is a \strong{single value} targeting the most logical structural unit
(e.g., the row of a matrix, or the 2D matrix of a 3D array).
If \code{NULL} (default), the entire dataset is read.}

\item{count}{A single numeric value specifying the number of elements or units to read.
If \code{NULL} (default) and \code{start} is provided, \code{h5lite} reads exactly 1 unit and
simplifies the resulting dimensions (see Section "Dimension Simplification").}
}
\value{
An R object corresponding to the HDF5 object or attribute.
Returns \code{NULL} if the object is skipped via \code{as = "null"}.
}
\description{
Reads a dataset, a group, or a specific attribute from an HDF5 file into an R object.
Supports partial reading (hyperslabs) to load specific subsets of data without
loading the entire object into memory.
}
\note{
The \code{@} prefix is \strong{only} used to configure attached attributes when reading a dataset (\code{attr = NULL}).
If you are reading a specific attribute directly (e.g., \code{h5_read(..., attr = "id")}), do \strong{not} use
the \code{@} prefix in the \code{as} argument.

Partial reading (\code{start}/\code{count}) is currently only supported for datasets, not attributes.
}
\section{Partial Reading (Hyperslabs)}{

You can read specific subsets of an n-dimensional dataset by utilizing the \code{start}
and \code{count} arguments.

\strong{The "Smart" \code{start} Parameter}

\code{start} is designed to be intuitive. Most of the time, you only need to provide a single value.
This single value automatically targets the most meaningful dimension of the dataset:
\itemize{
\item \strong{1D Vector:} \code{start} specifies the \strong{element}.
\item \strong{2D Matrix / Data Frame:} \code{start} specifies the \strong{row}.
\item \strong{3D Array:} \code{start} specifies the \strong{2D matrix}.
}

The \code{count} parameter is a \strong{single value} that determines how many of those units
to read sequentially. For example, \code{start = 5} and \code{count = 3} on a matrix will read 3 complete
rows starting at row 5 (automatically spanning all columns).

\strong{Multi-Value \code{start} and N-Dimensional Arrays}

If you need to extract a specific block \emph{inside} a structural unit, you can provide a vector of
values to \code{start}. To make indexing intuitive across higher-order arrays, \code{start} maps
its values to dimensions in the following priority order, targeting the outermost blocks first
and specific rows/columns last:
\itemize{
\item \verb{N, N-1, ..., 3, 1 (Rows), 2 (Cols)}
}

For example, on a 3D array, \code{start = c(2, 5)} targets the 2nd matrix, and the 5th row.
The \code{count} argument always applies to the \strong{last} dimension specified in \code{start}.

\strong{Dimension Simplification (Dropping)}

\code{h5lite} mimics R's native subsetting behavior regarding dimension preservation:
\itemize{
\item \strong{Exact Indexing (\code{count = NULL}):} If you provide \code{start} but omit \code{count}, \code{h5lite}
assumes you are targeting an exact point index. It will read 1 unit and \strong{drop} the
targeted dimension. (e.g., reading a specific row of a matrix will return a 1D vector).
\item \strong{Range Indexing (\code{count} provided):} If you explicitly provide \code{count} (even \code{count = 1}),
\code{h5lite} assumes you are reading a range. The dataset's original structural geometry is
\strong{preserved}. (e.g., reading \verb{start = 5, count = 1} on a matrix will return a 1xN matrix).
}
}

\section{Type Conversion (\code{as})}{

You can control how HDF5 data is converted to R types using the \code{as} argument.

\strong{1. Mapping by Name:}
\itemize{
\item \code{as = c("data_col" = "integer")}: Reads the dataset/column named "data_col" as an integer.
\item \code{as = c("@validated" = "logical")}: When reading a dataset, this forces the attached attribute "validated" to be read as logical.
}

\strong{2. Mapping by HDF5 Type Class:}
You can target specific HDF5 data types using keys prefixed with a dot (\code{.}).
Supported classes include:
\itemize{
\item \strong{Integer:} \code{.int}, \code{.int8}, \code{.int16}, \code{.int32}, \code{.int64}
\item \strong{Unsigned:} \code{.uint}, \code{.uint8}, \code{.uint16}, \code{.uint32}, \code{.uint64}
\item \strong{Floating Point:} \code{.float}, \code{.float16}, \code{.float32}, \code{.float64}
}
Example: \code{as = c(.uint8 = "logical", .int = "bit64")}

\strong{3. Precedence & Attribute Config:}
\itemize{
\item \strong{Attributes vs Datasets:} Attribute type mappings take precedence over dataset mappings.
If you specify \code{as = c(.uint = "logical", "@.uint" = "integer")}, unsigned integer datasets
will be read as \code{logical}, but unsigned integer \emph{attributes} will be read as \code{integer}.
\item \strong{Specific vs Generic:} Specific keys (e.g., \code{.uint32}) take precedence over generic keys (e.g., \code{.uint}),
which take precedence over the global default (\code{.}).
}
}

\examples{
file <- tempfile(fileext = ".h5")

# --- Setup: Write Test Data ---
h5_write(c(10L, 20L, 30L, 40L, 50L), file, "ints")

m <- matrix(1:50, nrow = 10, ncol = 5, dimnames = list(paste0("r", 1:10), paste0("c", 1:5)))
h5_write(m, file, "matrix_data")

arr <- array(1:24, dim = c(2, 3, 4))
h5_write(arr, file, "array_data")

# --- Standard Reading ---
# Read the entire dataset
x <- h5_read(file, "ints")

# --- Type Conversion ---
# Force integer dataset to be read as numeric (double)
x_dbl <- h5_read(file, "ints", as = "double")
class(x_dbl)

# --- Partial Reading: Single-Value 'start' ---
# Vector: Start at 2nd element, read 3 elements
h5_read(file, "ints", start = 2, count = 3)

# Matrix: Start at row 5, read 3 complete rows (returns 3x5 matrix)
h5_read(file, "matrix_data", start = 5, count = 3)

# 3D Array: Start at 2nd matrix, read 2 complete matrices (returns 2x3x2 array)
h5_read(file, "array_data", start = 2, count = 2)

# --- Partial Reading: Dimension Simplification ---
# Omit 'count' to extract an exact point index and drop the targeted dimension

# Matrix: Extract exactly row 5 (drops row dimension, returns a 1D vector)
h5_read(file, "matrix_data", start = 5)

# Matrix: Extract row 5, but preserve matrix structure (returns 1x5 matrix)
h5_read(file, "matrix_data", start = 5, count = 1)

# --- Partial Reading: Multi-Value 'start' ---
# Matrix: Extract exactly row 5, column 2 (drops both dims, returns a scalar)
h5_read(file, "matrix_data", start = c(5, 2))

# 3D Array: Target matrix 2, row 1. (drops matrix and row dims, returns 1D vector of cols)
h5_read(file, "array_data", start = c(2, 1))

unlink(file)
}
\seealso{
\code{\link[=h5_write]{h5_write()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.r
\name{h5_write}
\alias{h5_write}
\title{Write an R Object to HDF5}
\usage{
h5_write(data, file, name, attr = NULL, as = "auto", compress = "gzip")
}
\arguments{
\item{data}{The R object to write. Supported: \code{numeric}, \code{complex},
\code{logical}, \code{character}, \code{factor}, \code{raw}, \code{matrix}, \code{data.frame}, \code{integer64},
\code{POSIXt}, \code{NULL}, and nested \code{list}s.}

\item{file}{The path to the HDF5 file.}

\item{name}{The name of the dataset or group to write (e.g., "/data/matrix").}

\item{attr}{The name of an attribute to write.
\itemize{
\item If \code{NULL} (default), \code{data} is written as a dataset or group at the path \code{name}.
\item If provided (string), \code{data} is written as an attribute named \code{attr} attached to the object \code{name}.
}}

\item{as}{The target HDF5 data type. Defaults to \code{"auto"}.
See the \strong{Data Type Selection} section for a full list of valid options
(including \code{"int64"}, \code{"bfloat16"}, \code{"utf8[n]"}, etc.) and how to map
sub-components of \code{data}.}

\item{compress}{Compression configuration. Default is \code{"gzip"}. Pass a basic string
to specify the algorithm and level (e.g., \code{"none"}, \code{"gzip"}, \code{"zstd-7"}, \code{"lz4"},
\code{"blosc1-lz4-9"}, \code{"blosc2-gzip-3"}, \code{"blosc2-zstd"}), or pass a \code{compress} object
created by \code{\link[=h5_compression]{h5_compression()}} for advanced pipeline control (including scale-offset
algorithms, Fletcher32 checksums, or Blosc2 pre-filters). See \code{\link[=h5_compression]{h5_compression()}}
for the complete list of available codecs and options.}
}
\value{
Invisibly returns \code{file}. This function is called for its side effects.
}
\description{
Writes an R object to an HDF5 file, creating the file if it does not exist.
This function acts as a unified writer for datasets, groups (lists), and attributes.
}
\section{Writing Scalars}{


By default, \code{h5_write} saves single-element vectors as 1-dimensional arrays.
To write a true HDF5 scalar, wrap the value in \code{I()} to treat it "as-is."
\subsection{Examples}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{h5_write(I(5), file, "x") # Creates a scalar dataset
h5_write(5, file, "x")    # Creates a 1D array of length 1
}\if{html}{\out{</div>}}
}
}

\section{Data Type Selection (\code{as} Argument)}{


By default, \code{as = "auto"} will automatically select the most appropriate
data type for the given object. For numeric types, this will be the smallest
type that can represent all values in the vector. For character types,
\code{h5lite} will use a ragged vs rectangular heuristic, favoring small file
size over fast I/O. For R data types not mentioned below, see
\code{vignette("data-types")} for information on their fixed mappings to HDF5
data types.
\subsection{Numeric and Logical Vectors}{

When writing a numeric or logical vector, you can specify one of the
following storage types for it:
\itemize{
\item \strong{Floating Point:} \code{"float16"}, \code{"float32"}, \code{"float64"}, \code{"bfloat16"}
\item \strong{Signed Integer:} \code{"int8"}, \code{"int16"}, \code{"int32"}, \code{"int64"}
\item \strong{Unsigned Integer:} \code{"uint8"}, \code{"uint16"}, \code{"uint32"}, \code{"uint64"}
}

\strong{NOTE:} \code{NA} values must be stored as \code{float64}. \code{NaN}, \code{Inf}, and \code{-Inf}
must be stored as a floating point type.
\subsection{Examples}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{h5_write(1:100, file, "big_ints", as = "int64")
h5_write(TRUE,  file, "my_bool",  as = "float32")
}\if{html}{\out{</div>}}
}

}

\subsection{Character Vectors}{

You can control whether character vectors are stored as variable or fixed
length strings, and whether to use UTF-8 or ASCII encoding.
\itemize{
\item \strong{Variable Length Strings:} \code{"utf8"}, \code{"ascii"}
\item \strong{Fixed Length Strings:}
\itemize{
\item \code{"utf8[]"} or \code{"ascii[]"} (length is set to the longest string)
\item \code{"utf8[n]"} or \code{"ascii[n]"} (where \code{n} is the length in bytes)
}
}

\strong{NOTE:} Variable-length strings allow for \code{NA} values but cannot be
compressed on disk. Fixed-length strings allow for compression but do not
support \code{NA}.
\subsection{Examples}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{h5_write(letters[1:5],    file, "len10_strs", as = "utf8[10]")
h5_write(c('X', 'Y', NA), file, "var_chars",  as = "ascii")
}\if{html}{\out{</div>}}
}

}

\subsection{Lists, Data Frames, and Attributes}{

Provide a named vector to apply type mappings to sub-components of \code{data}.
Set \code{"skip"} as the type to skip a specific component.
\itemize{
\item \strong{Specific Name:} \code{"col_name" = "type"} (e.g., \code{c(score = "float32")})
\item \strong{Specific Attribute:} \code{"@attr_name" = "type"}
\item \strong{Class-based:} \code{".integer" = "type"}, \code{".numeric" = "type"}
\item \strong{Class-based Attribute:} \code{"@.character" = "type"}, \code{"@.logical" = "type"}
\item \strong{Global Fallback:} \code{"." = "type"}
\item \strong{Global Attribute Fallback:} \code{"@." = "type"}
}
\subsection{Examples}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{# To strip attributes when writing:
h5_write(data, file, 'no_attrs_obj', as = c('@.' = "skip"))

# To only save the `hp` and `wt` columns:
h5_write(mtcars, file, 'my_df', as = c('hp' = "auto", 'wt' = "float32", '.' = "skip"))
}\if{html}{\out{</div>}}
}

}
}

\section{Dimension Scales}{

\code{h5lite} automatically writes \code{names}, \code{row.names}, and \code{dimnames} as
HDF5 dimension scales. Named vectors will generate an \verb{<name>_names}
dataset. A data.frame with row names will generate an \verb{<name>_rownames}
dataset (column names are saved internally in the original dataset).
Matrices will generate \verb{<name>_rownames} and \verb{<name>_colnames} datasets.
Arrays will generate \verb{<name>_dimscale_1}, \verb{<name>_dimscale_2}, etc.
Special HDF5 metadata attributes link the dimension scales to the dataset.
The dimension scales can be relocated with \code{h5_move()} without breaking the
link.
}

\examples{
file <- tempfile(fileext = ".h5")

# 1. Writing Basic Datasets
h5_write(1:10, file, "data/integers")
h5_write(rnorm(10), file, "data/floats")
h5_write(letters[1:5], file, "data/chars")

# 2. Writing Attributes
# Write an object first
h5_write(1:10, file, "data/vector")
# Attach an attribute to it using the 'attr' parameter
h5_write(I("My Description"), file, "data/vector", attr = "description")
h5_write(I(100), file, "data/vector", attr = "scale_factor")

# 3. Controlling Data Types
# Store values as 32-bit signed integers
h5_write(1:5, file, "small_ints", as = "int32")

# 4. Writing Complex Structures (Lists/Groups)
my_list <- list(
  meta    = list(id = 1, name = "Experiment A"),
  results = matrix(runif(9), 3, 3),
  valid   = I(TRUE)
)
h5_write(my_list, file, "experiment_1", as = c(id = "uint16"))

# 5. Writing Data Frames (Compound Datasets)
df <- data.frame(
  id    = 1:5,
  score = c(10.5, 9.2, 8.4, 7.1, 6.0),
  grade = factor(c("A", "A", "B", "C", "D"))
)
h5_write(df, file, "records/scores", as = c(grade = "ascii[1]"))

# 6. Fixed-Length Strings
h5_write(c("A", "B"), file, "fixed_str", as = "ascii[10]")

# 7. Review the file structure
h5_str(file)

# 8. Clean up
unlink(file)
}
\seealso{
\code{\link[=h5_read]{h5_read()}}, \code{\link[=h5_compression]{h5_compression()}}, \code{vignette('compression')}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compress.r
\name{h5_compression}
\alias{h5_compression}
\title{Define HDF5 Compression and Filter Settings}
\usage{
h5_compression(
  compress = "gzip",
  chunk_size = 1024 * 1024,
  checksum = FALSE,
  int_packing = FALSE,
  float_rounding = NULL,
  blosc2_delta = FALSE,
  blosc2_truncate = NULL
)
}
\arguments{
\item{compress}{A string specifying the compression algorithm and optional level
(e.g., \code{"none"}, \code{"gzip"}, \code{"zstd-7"}, \code{"lz4"}, \code{"blosc1-lz4-9"},
\code{"blosc2-gzip-3"}, \code{"blosc2-zstd"}). See the \strong{Valid Compression Strings}
section below for an exhaustive list of supported formats. Default is \code{"gzip"}.}

\item{chunk_size}{An integer specifying the target chunk size in bytes.
Default is \code{1048576} (1 MB).}

\item{checksum}{A logical value indicating whether to apply the Fletcher32
checksum filter at the end of the pipeline to detect data corruption. Default is \code{FALSE}.}

\item{int_packing}{Control the HDF5 Scale-Offset filter for integer datasets. \emph{(Note:
Incompatible with \code{szip}, \code{zfp}, \code{bshuf}, and Blosc2 pre-filters).}
\itemize{
\item \code{FALSE} (Default): Disabled.
\item \code{TRUE}: Automatically calculates and applies the mathematically optimal minimum
bit-width for each individual chunk.
\item Integer (e.g., \code{8}): Forces packing into exactly that many bits.
}}

\item{float_rounding}{Control the HDF5 Scale-Offset filter for floating-point
datasets.\emph{(Note: Incompatible with \code{szip}, \code{zfp}, \code{bshuf}, and Blosc2 pre-filters).}
\itemize{
\item \code{NULL} (Default): Disabled.
\item Integer (e.g., \code{3}): The number of base-10 decimal places of detail to preserve
before truncating and packing the values (e.g., \code{3.141}). Negative numbers
round to powers of 10.
}}

\item{blosc2_delta}{A logical value. If \code{TRUE} and a \code{blosc2} compressor is selected,
applies the Blosc2 Delta pre-filter before compression. Default is \code{FALSE}.}

\item{blosc2_truncate}{An integer. If provided and a \code{blosc2} compressor is selected,
applies the Blosc2 Truncate Precision pre-filter to floating-point data, preserving
exactly the specified number of uncompressed bits. Default is \code{NULL}.}
}
\value{
An S3 object of class \code{compress} containing the parsed pipeline parameters.
}
\description{
Constructs a comprehensive filter pipeline configuration to be passed
as the \code{compress} argument to \code{\link[=h5_write]{h5_write()}}. This function allows fine-grained
control over chunking, pre-filters, compression algorithms, and data scaling.
}
\section{Valid Compression Strings}{

The \code{compress} argument accepts a highly specific string syntax to define both
the codec and its operational level.
\subsection{Native / Core Codecs}{
\itemize{
\item \code{"none"}: No compression.
\item \code{"gzip-[level]"}: Levels \code{1} to \code{9}. Default is \code{5}. (e.g., \code{"gzip"} or \code{"gzip-9"}).
\item \code{"zstd-[level]"}: Levels \code{1} to \code{22}. Default is \code{3}. (e.g., \code{"zstd"} or \code{"zstd-7"}).
\item \code{"lz4-[level]"}: Levels \code{0} to \code{12}. Default is \code{0}. Level \code{0} is standard LZ4. Levels \verb{1+} trigger LZ4-HC.
}
}

\subsection{Bitshuffle Pre-filter}{

Forces the native Bitshuffle pre-filter before compression.
\itemize{
\item \code{"bshuf-lz4"}: Bitshuffle + LZ4.
\item \code{"bshuf-zstd-[level]"}: Bitshuffle + Zstd (Levels \code{1} to \code{22}).
}
}

\subsection{Blosc Meta-compressors}{

Blosc applies its own highly optimized bitshuffling and multi-threading.
\itemize{
\item \strong{Blosc2 (Recommended):} \code{"blosc2"} (blosclz), \code{"blosc2-lz4-[level]"}, \code{"blosc2-zstd-[level]"}, \code{"blosc2-gzip-[level]"}, \code{"blosc2-ndlz"}
\item \strong{Blosc1 (Legacy):}  \code{"blosc1"} (blosclz), \code{"blosc1-lz4-[level]"}, \code{"blosc1-zstd-[level]"}, \code{"blosc1-gzip-[level]"}, \code{"blosc1-snappy"}
}
}

\subsection{ZFP (Lossy Floating-Point Compression)}{

ZFP can be run standalone (for integers and floats) or inside Blosc2 (floats only). Unlike \verb{[level]}, \verb{[tolerance]} and \verb{[bits]} are required.
\itemize{
\item \strong{Accuracy Mode} (Absolute error tolerance): \code{"zfp-acc-[tolerance]"} or \code{"blosc2-zfp-acc-[tolerance]"} (e.g., \code{"zfp-acc-0.001"}).
\item \strong{Precision Mode} (Bits of precision): \code{"zfp-prec-[bits]"} or \code{"blosc2-zfp-prec-[bits]"} (e.g., \code{"zfp-prec-16"}).
\item \strong{Rate Mode} (Bits of storage per value): \code{"zfp-rate-[bits]"} or \code{"blosc2-zfp-rate-[bits]"} (e.g., \code{"zfp-rate-8"}).
\item \strong{Reversible Mode} (Standalone Lossless): \code{"zfp-rev"}.
}
}

\subsection{Legacy Codecs}{
\itemize{
\item \code{"szip-nn"}, \code{"szip-ec"}: SZIP Nearest Neighbor or Entropy Coding.
\item \code{"bzip2-[level]"}: Levels \code{1} to \code{9}. Default is \code{9}. (e.g., \code{"bzip2-4"}).
\item \code{"lzf"}, \code{"snappy"}: Fast, unconfigurable legacy compressors.
}
}
}

\section{Automatic Shuffling}{


To maximize compression ratios without requiring users to manually manage
complex pipeline interactions, \code{h5_compression} automatically configures the
optimal shuffling pre-filter based on the following strict hierarchy:

\strong{1. Blosc's Internal Bitshuffle (Preferred)}
If a Blosc meta-compressor is selected (e.g., \code{"blosc2-zstd"}), the pipeline
automatically enables Blosc's highly optimized, internal bitshuffle routine.
This achieves peak compression performance without requiring the standalone
Bitshuffle plugin to be installed.

\strong{2. Explicit Bitshuffle Plugin}
If a standard codec is explicitly prefixed with \verb{bshuf-} (e.g., \code{"bshuf-lz4"}),
the pipeline delegates to the standalone Bitshuffle plugin.

\strong{3. Native HDF5 Byte Shuffle (Fallback)}
If a standard compressor is selected (e.g., \code{"zstd-5"} or \code{"gzip"}),
the pipeline safely falls back to the core HDF5 library's native byte shuffle
filter. This guarantees improved compression while maintaining universal
compatibility.

\strong{4. Strict Mutual Exclusions (When Shuffling is Disabled)}
To prevent data corruption or wasted CPU cycles, all shuffling is
\strong{forcefully disabled} in the following scenarios:
\itemize{
\item \strong{Scale-Offset Active:} If \code{int_packing} or \code{float_rounding} is applied,
shuffling is disabled because scale-offset destroys the byte-alignment
that shuffling relies on.
\item \strong{ZFP & SZIP:} These algorithms perform mathematical compression directly
on numerical values and will corrupt if the bitstream is rearranged beforehand.
\item \strong{1-Byte Data:} Characters, booleans, and 8-bit integers cannot be meaningfully
shuffled, so the step is skipped.
}
}

\examples{
# 1. Simple fast compression (Zstd level 7)
h5_compression("zstd-7")

# 2. Optimal integer packing (Scale-Offset)
h5_compression("gzip-9", int_packing = TRUE)

# 3. Complex Blosc2 Pipeline (Delta + Zstd)
h5_compression("blosc2-zstd-5", blosc2_delta = TRUE)

# 4. Lossy ZFP compression (Tolerance of 0.05)
h5_compression("zfp-acc-0.05")

# Pass the compress object directly to h5_write
file <- tempfile(fileext = ".h5")
cmp  <- h5_compression("gzip-9", checksum = TRUE)
h5_write(combn(1:10, 3), file, "sets", compress = cmp)

print(cmp)

h5_inspect(file, "sets")

# Clean up
unlink(file)

}
\seealso{
\code{\link[=h5_write]{h5_write()}}, \code{vignette('compression')}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.r
\name{h5_attr_names}
\alias{h5_attr_names}
\title{List HDF5 Attributes}
\usage{
h5_attr_names(file, name = "/")
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{The path to the object (dataset or group) to query.
Use \code{/} for the file's root attributes.}
}
\value{
A character vector of attribute names.
}
\description{
Lists the names of attributes attached to a specific HDF5 object.
}
\examples{
file <- tempfile(fileext = ".h5")

h5_write(1:10,          file, "data")
h5_write(I("meters"),   file, "data", attr = "unit")
h5_write(I(Sys.time()), file, "data", attr = "timestamp")

h5_attr_names(file, "data") # "unit" "timestamp"

unlink(file)
}
\seealso{
\code{\link[=h5_ls]{h5_ls()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/organize.r
\name{h5_move}
\alias{h5_move}
\title{Move or Rename an HDF5 Object}
\usage{
h5_move(file, from, to)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{from}{The current (source) path of the object (e.g., \code{"/group/data"}).}

\item{to}{The new (destination) path for the object (e.g., \code{"/group/data_new"}).}
}
\value{
This function is called for its side-effect and returns \code{NULL}
invisibly.
}
\description{
Moves or renames an object (dataset, group, etc.) within an HDF5 file.
}
\details{
This function provides an efficient, low-level wrapper for the HDF5
library's \code{H5Lmove} function. It is a metadata-only operation, meaning the
data itself is not read or rewritten. This makes it extremely fast, even
for very large datasets.

You can use this function to either rename an object within the same group
(e.g., \code{"data/old"} to \code{"data/new"}) or to move an object to a
different group (e.g., \code{"data/old"} to \code{"archive/old"}). The destination
parent group will be automatically created if it does not exist.
}
\examples{
file <- tempfile(fileext = ".h5")
h5_write(1:10, file, "group/dataset")

# Review the file structure
h5_str(file)

# Rename within the same group
h5_move(file, "group/dataset", "group/renamed")

# Review the file structure
h5_str(file)

# Move to a new group (creates parent automatically)
h5_move(file, "group/renamed", "archive/dataset")

# Review the file structure
h5_str(file)

unlink(file)
}
\seealso{
\code{\link[=h5_create_group]{h5_create_group()}}, \code{\link[=h5_delete]{h5_delete()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.r
\name{h5_is_dataset}
\alias{h5_is_dataset}
\title{Check if an HDF5 Object is a Dataset}
\usage{
h5_is_dataset(file, name, attr = NULL)
}
\arguments{
\item{file}{The path to the HDF5 file.}

\item{name}{The full path of the object to check.}

\item{attr}{The name of an attribute. If provided, the function returns \code{TRUE} if the attribute exists,
as all attributes are considered datasets in HDF5 context. (Default: \code{NULL})}
}
\value{
A logical value: \code{TRUE} if the object exists and is a dataset,
\code{FALSE} otherwise (if it is a group, or does not exist).
}
\description{
Checks if the object at a given path is a dataset.
}
\examples{
file <- tempfile(fileext = ".h5")

h5_write(1, file, "dset")
h5_is_dataset(file, "dset") # TRUE

h5_create_group(file, "grp")
h5_is_dataset(file, "grp") # FALSE

h5_write(1, file, "grp", attr = "my_attr")
h5_is_dataset(file, "grp", "my_attr") # TRUE

unlink(file)
}
\seealso{
\code{\link[=h5_is_group]{h5_is_group()}}, \code{\link[=h5_exists]{h5_exists()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/organize.r
\name{h5_create_file}
\alias{h5_create_file}
\title{Create an HDF5 File}
\usage{
h5_create_file(file)
}
\arguments{
\item{file}{Path to the HDF5 file to be created.}
}
\value{
Invisibly returns \code{NULL}. This function is called for its side
effects.
}
\description{
Explicitly creates a new, empty HDF5 file.
}
\details{
This function is a simple wrapper around \code{h5_create_group(file, "/")}.
Its main purpose is to allow for explicit file creation in code.

Note that calling this function is almost always \strong{unnecessary}, as all
\code{h5lite} writing functions (like \code{\link[=h5_write]{h5_write()}} or
\code{\link[=h5_create_group]{h5_create_group()}}) will automatically create
the file if it does not exist.

It is provided as a convenience for users who prefer to explicitly create
a file before writing data to it.
}
\section{File Handling}{

\itemize{
\item If \code{file} does not exist, it will be created as a new, empty HDF5 file.
\item If \code{file} already exists and is a valid HDF5 file, this function does
nothing and returns successfully.
\item If \code{file} exists but is \strong{not} a valid HDF5 file (e.g., a text file),
an error will be thrown and the file will not be modified.
}
}

\examples{
file <- tempfile(fileext = ".h5")

# Explicitly create the file
h5_create_file(file)

if (file.exists(file)) {
  message("File created successfully.")
}

unlink(file)
}
\seealso{
\code{\link[=h5_create_group]{h5_create_group()}}, \code{\link[=h5_write]{h5_write()}}
}
